@@ -2,15 +2,16 @@
 
   bignum.c -
 
-  $Author: usa $
-  $Date: 2002/02/26 12:36:26 $
+  $Author: matz $
+  $Date: 2003/07/24 18:33:39 $
   created at: Fri Jun 10 00:48:55 JST 1994
 
-  Copyright (C) 1993-2000 Yukihiro Matsumoto
+  Copyright (C) 1993-2003 Yukihiro Matsumoto
 
 **********************************************************************/
 
 #include "ruby.h"
+
 #include <math.h>
 #include <ctype.h>
 
@@ -20,27 +21,19 @@ VALUE rb_cBignum;
 #define USHORT _USHORT
 #endif
 
-#if SIZEOF_INT*2 <= SIZEOF_LONG_LONG
-typedef unsigned int BDIGIT;
-typedef unsigned long long BDIGIT_DBL;
-typedef long long BDIGIT_DBL_SIGNED;
-#elif SIZEOF_INT*2 <= SIZEOF___INT64
-typedef unsigned int BDIGIT;
-typedef unsigned __int64 BDIGIT_DBL;
-typedef __int64 BDIGIT_DBL_SIGNED;
-#else
-typedef unsigned short BDIGIT;
-typedef unsigned long BDIGIT_DBL;
-typedef long BDIGIT_DBL_SIGNED;
-#endif
-
 #define BDIGITS(x) ((BDIGIT*)RBIGNUM(x)->digits)
-#define BITSPERDIG (sizeof(BDIGIT)*CHAR_BIT)
+#define BITSPERDIG (SIZEOF_BDIGITS*CHAR_BIT)
 #define BIGRAD ((BDIGIT_DBL)1 << BITSPERDIG)
-#define DIGSPERLONG ((unsigned int)(sizeof(long)/sizeof(BDIGIT)))
+#define DIGSPERLONG ((unsigned int)(SIZEOF_LONG/SIZEOF_BDIGITS))
+#if HAVE_LONG_LONG
+# define DIGSPERLL ((unsigned int)(SIZEOF_LONG_LONG/SIZEOF_BDIGITS))
+#endif
 #define BIGUP(x) ((BDIGIT_DBL)(x) << BITSPERDIG)
 #define BIGDN(x) RSHIFT(x,BITSPERDIG)
 #define BIGLO(x) ((BDIGIT)((x) & (BIGRAD-1)))
+#define BDIGMAX ((BDIGIT)-1)
+
+#define BIGZEROP(x) (RBIGNUM(x)->len == 0 || (RBIGNUM(x)->len == 1 && BDIGITS(x)[0] == 0))
 
 static VALUE
 bignew_1(klass, len, sign)
@@ -86,14 +79,10 @@ get2comp(x, carry)		/* get 2's complement */
 	num = BIGDN(num);
     } while (i < RBIGNUM(x)->len);
     if (!carry) return;
-    if (ds[0] == 1 || ds[0] == 0) {
-	if (RBIGNUM(x)->len == 1) return;
-	for (i=1; i<RBIGNUM(x)->len; i++) {
-	    if (ds[i] != 0) return;
-	}
-	REALLOC_N(RBIGNUM(x)->digits, BDIGIT, RBIGNUM(x)->len++);
+    if ((ds[RBIGNUM(x)->len-1] & (1<<(BITSPERDIG-1))) == 0) {
+	REALLOC_N(RBIGNUM(x)->digits, BDIGIT, ++RBIGNUM(x)->len);
 	ds = BDIGITS(x);
-	ds[RBIGNUM(x)->len-1] = 1;
+	ds[RBIGNUM(x)->len-1] = ~0;
     }
 }
 
@@ -115,16 +104,16 @@ bignorm(x)
 	while (len-- && !ds[len]) ;
 	RBIGNUM(x)->len = ++len;
 
-	if (len*sizeof(BDIGIT) <= sizeof(VALUE)) {
+	if (len*SIZEOF_BDIGITS <= sizeof(VALUE)) {
 	    long num = 0;
 	    while (len--) {
 		num = BIGUP(num) + ds[len];
 	    }
 	    if (num >= 0) {
 		if (RBIGNUM(x)->sign) {
-		    if (POSFIXABLE(num)) return INT2FIX(num);
+		    if (POSFIXABLE(num)) return LONG2FIX(num);
 		}
-		else if (NEGFIXABLE(-(long)num)) return INT2FIX(-(long)num);
+		else if (NEGFIXABLE(-(long)num)) return LONG2FIX(-(long)num);
 	    }
 	}
     }
@@ -147,7 +136,6 @@ rb_uint2big(n)
     BDIGIT *digits;
     VALUE big;
 
-    i = 0;
     big = bignew(DIGSPERLONG, 1);
     digits = BDIGITS(big);
     while (i < DIGSPERLONG) {
@@ -156,7 +144,7 @@ rb_uint2big(n)
     }
 
     i = DIGSPERLONG;
-    while (i-- && !digits[i]) ;
+    while (--i && !digits[i]) ;
     RBIGNUM(big)->len = i+1;
     return big;
 }
@@ -183,7 +171,7 @@ VALUE
 rb_uint2inum(n)
     unsigned long n;
 {
-    if (POSFIXABLE(n)) return INT2FIX(n);
+    if (POSFIXABLE(n)) return LONG2FIX(n);
     return rb_uint2big(n);
 }
 
@@ -191,26 +179,159 @@ VALUE
 rb_int2inum(n)
     long n;
 {
-    if (FIXABLE(n)) return INT2FIX(n);
+    if (FIXABLE(n)) return LONG2FIX(n);
     return rb_int2big(n);
 }
 
+#ifdef HAVE_LONG_LONG
+
+void
+rb_quad_pack(buf, val)
+    char *buf;
+    VALUE val;
+{
+    LONG_LONG q;
+
+    val = rb_to_int(val);
+    if (FIXNUM_P(val)) {
+	q = FIX2LONG(val);
+    }
+    else {
+	long len = RBIGNUM(val)->len;
+	BDIGIT *ds;
+
+	ds = BDIGITS(val);
+	q = 0;
+	while (len--) {
+	    q = BIGUP(q);
+	    q += ds[len];
+	}
+	if (!RBIGNUM(val)->sign) q = -q;
+    }
+    memcpy(buf, (char*)&q, SIZEOF_LONG_LONG);
+}
+
 VALUE
-rb_cstr2inum(str, base)
+rb_quad_unpack(buf, sign)
+    const char *buf;
+    int sign;
+{
+    unsigned LONG_LONG q;
+    long neg = 0;
+    long i;
+    BDIGIT *digits;
+    VALUE big;
+
+    memcpy(&q, buf, SIZEOF_LONG_LONG);
+    if (sign) {
+	if (FIXABLE((LONG_LONG)q)) return LONG2FIX((LONG_LONG)q);
+	if ((LONG_LONG)q < 0) {
+	    q = -(LONG_LONG)q;
+	    neg = 1;
+	}
+    }
+    else {
+	if (POSFIXABLE(q)) return LONG2FIX(q);
+    }
+
+    i = 0;
+    big = bignew(DIGSPERLL, 1);
+    digits = BDIGITS(big);
+    while (i < DIGSPERLL) {
+	digits[i++] = BIGLO(q);
+	q = BIGDN(q);
+    }
+
+    i = DIGSPERLL;
+    while (i-- && !digits[i]) ;
+    RBIGNUM(big)->len = i+1;
+
+    if (neg) {
+	RBIGNUM(big)->sign = 0;
+    }
+    return bignorm(big);
+}
+
+#else
+
+#define QUAD_SIZE 8
+
+void
+rb_quad_pack(buf, val)
+    char *buf;
+    VALUE val;
+{
+    long len;
+
+    memset(buf, 0, QUAD_SIZE);
+    val = rb_to_int(val);
+    if (FIXNUM_P(val)) {
+	val = rb_int2big(FIX2LONG(val));
+    }
+    len = RBIGNUM(val)->len * SIZEOF_BDIGITS;
+    if (len > QUAD_SIZE) len = QUAD_SIZE;
+    memcpy(buf, (char*)BDIGITS(val), len);
+    if (!RBIGNUM(val)->sign) {
+	len = QUAD_SIZE;
+	while (len--) {
+	    *buf = ~*buf;
+	    buf++;
+	}
+    }
+}
+
+#define BNEG(b) (RSHIFT(((BDIGIT*)b)[QUAD_SIZE/SIZEOF_BDIGITS-1],BITSPERDIG-1) != 0)
+
+VALUE
+rb_quad_unpack(buf, sign)
+    const char *buf;
+    int sign;
+{
+    VALUE big = bignew(QUAD_SIZE/SIZEOF_BDIGITS, 1);
+
+    memcpy((char*)BDIGITS(big), buf, QUAD_SIZE);
+    if (sign && BNEG(buf)) {
+	long len = QUAD_SIZE;
+	char *tmp = (char*)BDIGITS(big);
+
+	RBIGNUM(big)->sign = 0;
+	while (len--) {
+	    *tmp = ~*tmp;
+	    tmp++;
+	}
+    }
+
+    return bignorm(big);
+}
+
+#endif
+
+VALUE
+rb_cstr_to_inum(str, base, badcheck)
     const char *str;
     int base;
+    int badcheck;
 {
     const char *s = str;
     char *end;
-    int badcheck = (base==0)?1:0;
-    char sign = 1, c;
+    char sign = 1, nondigit = 0;
+    int c;
     BDIGIT_DBL num;
     long len, blen = 1;
     long i;
     VALUE z;
     BDIGIT *zds;
 
-    while (*str && ISSPACE(*str)) str++;
+    if (!str) {
+	if (badcheck) goto bad;
+	return INT2FIX(0);
+    }
+    if (badcheck) {
+	while (ISSPACE(*str)) str++;
+    }
+    else {
+	while (ISSPACE(*str) || *str == '_') str++;
+    }
 
     if (str[0] == '+') {
 	str++;
@@ -223,33 +344,73 @@ rb_cstr2inum(str, base)
 	if (badcheck) goto bad;
 	return INT2FIX(0);
     }
-    if (base == 0) {
+    if (base <= 0) {
 	if (str[0] == '0') {
-	    if (str[1] == 'x' || str[1] == 'X') {
+	    switch (str[1]) {
+	      case 'x': case 'X':
 		base = 16;
-	    }
-	    else if (str[1] == 'b' || str[1] == 'B') {
+		break;
+	      case 'b': case 'B':
 		base = 2;
-	    }
-	    else {
+		break;
+	      case 'o': case 'O':
+		base = 8;
+		break;
+	      case 'd': case 'D':
+		base = 10;
+		break;
+	      default:
 		base = 8;
 	    }
 	}
+	else if (base < -1) {
+	    base = -base;
+	}
 	else {
 	    base = 10;
 	}
     }
-    if (base == 8) {
+    switch (base) {
+      case 2:
+	len = 1;
+	if (str[0] == '0' && (str[1] == 'b'||str[1] == 'B')) {
+	    str += 2;
+	}
+	break;
+      case 3:
+	len = 2;
+	break;
+      case 8:
+	if (str[0] == '0' && (str[1] == 'o'||str[1] == 'O')) {
+	    str += 2;
+	}
+      case 4: case 5: case 6: case 7:
 	len = 3;
-    }
-    else {			/* base == 10, 2 or 16 */
-	if (base == 16 && str[0] == '0' && (str[1] == 'x'||str[1] == 'X')) {
-	    str += 2;
-	}
-	else if (base == 2 && str[0] == '0' && (str[1] == 'b'||str[1] == 'B')) {
+	break;
+      case 10:
+	if (str[0] == '0' && (str[1] == 'd'||str[1] == 'D')) {
 	    str += 2;
 	}
+      case 9: case 11: case 12: case 13: case 14: case 15:
 	len = 4;
+	break;
+      case 16:
+	len = 4;
+	if (str[0] == '0' && (str[1] == 'x'||str[1] == 'X')) {
+	    str += 2;
+	}
+	break;
+      default:
+	if (base < 2 || 36 < base) {
+	    rb_raise(rb_eArgError, "illegal radix %d", base);
+	}
+	if (base <= 32) {
+	    len = 5;
+	}
+	else {
+	    len = 6;
+	}
+	break;
     }
     if (*str == '0') {		/* squeeze preceeding 0s */
 	while (*++str == '0');
@@ -264,23 +425,20 @@ rb_cstr2inum(str, base)
 	if (badcheck) {
 	    if (end == str) goto bad; /* no number */
 	    while (*end && ISSPACE(*end)) end++;
-	    if (*end) {		      /* trailing garbage */
-	      bad:
-		rb_raise(rb_eArgError, "invalid value for Integer: \"%s\"", s);
-	    }
+	    if (*end) goto bad;	      /* trailing garbage */
 	}
 
 	if (POSFIXABLE(val)) {
-	    if (sign) return INT2FIX(val);
+	    if (sign) return LONG2FIX(val);
 	    else {
 		long result = -(long)val;
-		return INT2FIX(result);
+		return LONG2FIX(result);
 	    }
 	}
 	else {
 	    VALUE big = rb_uint2big(val);
 	    RBIGNUM(big)->sign = sign;
-	    return big;
+	    return bignorm(big);
 	}
     }
   bigparse:
@@ -291,33 +449,27 @@ rb_cstr2inum(str, base)
     zds = BDIGITS(z);
     for (i=len;i--;) zds[i]=0;
     while (c = *str++) {
-	switch (c) {
-	  case '8': case '9':
-	    if (base == 8) {
-		c = base;
-		break;
+	if (c == '_') {
+	    if (badcheck) {
+		if (nondigit) goto bad;
+		nondigit = c;
 	    }
-	  case '0': case '1': case '2': case '3': case '4':
-	  case '5': case '6': case '7': 
-	    c = c - '0';
-	    break;
-	  case 'a': case 'b': case 'c':
-	  case 'd': case 'e': case 'f':
-	    if (base != 16) c = base;
-	    else c = c - 'a' + 10;
-	    break;
-	  case 'A': case 'B': case 'C':
-	  case 'D': case 'E': case 'F':
-	    if (base != 16) c = base;
-	    else c = c - 'A' + 10;
-	    break;
-	  case '_':
 	    continue;
-	  default:
-	    c = base;
+	}
+	else if (isdigit(c)) {
+	    c -= '0';
+	}
+	else if (islower(c)) {
+	    c -= 'a' - 10;
+	}
+	else if (isupper(c)) {
+	    c -= 'A' - 10;
+	}
+	else {
 	    break;
 	}
 	if (c >= base) break;
+	nondigit = 0;
 	i = 0;
 	num = c;
 	for (;;) {
@@ -337,35 +489,121 @@ rb_cstr2inum(str, base)
 	str--;
 	if (s+1 < str && str[-1] == '_') goto bad;
 	while (*str && ISSPACE(*str)) str++;
-	if (*str) goto bad;
+	if (*str) {
+	  bad:
+	    rb_invalid_str(s, "Integer");
+	}
     }
 
     return bignorm(z);
 }
 
+VALUE
+rb_str_to_inum(str, base, badcheck)
+    VALUE str;
+    int base;
+    int badcheck;
+{
+    char *s;
+    long len;
+
+    StringValue(str);
+    if (badcheck) {
+	s = StringValueCStr(str);
+    }
+    else {
+	s = RSTRING(str)->ptr;
+    }
+    if (s) {
+	len = RSTRING(str)->len;
+	if (s[len]) {		/* no sentinel somehow */
+	    char *p = ALLOCA_N(char, len+1);
+
+	    MEMCPY(p, s, char, len);
+	    p[len] = '\0';
+	    s = p;
+	}
+    }
+    return rb_cstr_to_inum(s, base, badcheck); 
+}
+
+#if HAVE_LONG_LONG
+
+VALUE
+rb_ull2big(n)
+    unsigned LONG_LONG n;
+{
+    BDIGIT_DBL num = n;
+    long i = 0;
+    BDIGIT *digits;
+    VALUE big;
+
+    big = bignew(DIGSPERLL, 1);
+    digits = BDIGITS(big);
+    while (i < DIGSPERLL) {
+	digits[i++] = BIGLO(num);
+	num = BIGDN(num);
+    }
+
+    i = DIGSPERLL;
+    while (i-- && !digits[i]) ;
+    RBIGNUM(big)->len = i+1;
+    return big;
+}
+
+VALUE
+rb_ll2big(n)
+    LONG_LONG n;
+{
+    long neg = 0;
+    VALUE big;
+
+    if (n < 0) {
+	n = -n;
+	neg = 1;
+    }
+    big = rb_ull2big(n);
+    if (neg) {
+	RBIGNUM(big)->sign = 0;
+    }
+    return big;
+}
+
+VALUE
+rb_ull2inum(n)
+    unsigned LONG_LONG n;
+{
+    if (POSFIXABLE(n)) return LONG2FIX(n);
+    return rb_ull2big(n);
+}
+
+VALUE
+rb_ll2inum(n)
+    LONG_LONG n;
+{
+    if (FIXABLE(n)) return LONG2FIX(n);
+    return rb_ll2big(n);
+}
+
+#endif  /* HAVE_LONG_LONG */
+ 
+VALUE
+rb_cstr2inum(str, base)
+    const char *str;
+    int base;
+{
+    return rb_cstr_to_inum(str, base, base==0);
+}
+
 VALUE
 rb_str2inum(str, base)
     VALUE str;
     int base;
 {
-    char *s;
-    int len;
-
-    s = rb_str2cstr(str, &len);
-    if (s[len]) {		/* no sentinel somehow */
-	char *p = ALLOCA_N(char, len+1);
-
-	MEMCPY(p, s, char, len);
-	p[len] = '\0';
-	s = p;
-    }
-    if (len != strlen(s)) {
-	rb_raise(rb_eArgError, "string for Integer contains null byte");
-    }
-    return rb_cstr2inum(s, base); 
+    return rb_str_to_inum(str, base, base==0);
 }
 
-static char hexmap[] = "0123456789abcdef";
+const char ruby_digitmap[] = "0123456789abcdefghijklmnopqrstuvwxyz";
 VALUE
 rb_big2str(x, base)
     VALUE x;
@@ -381,28 +619,42 @@ rb_big2str(x, base)
 	return rb_fix2str(x, base);
     }
     i = RBIGNUM(x)->len;
-    if (i == 0) return rb_str_new2("0");
-    if (base == 10) {
-	j = (sizeof(BDIGIT)/sizeof(char)*CHAR_BIT*i*241L)/800+2;
-	hbase = 10000;
+    if (BIGZEROP(x)) {
+	return rb_str_new2("0");
     }
-    else if (base == 16) {
-	j = (sizeof(BDIGIT)/sizeof(char)*CHAR_BIT*i)/4+2;
-	hbase = 0x10000;
-    }
-    else if (base == 8) {
-	j = (sizeof(BDIGIT)/sizeof(char)*CHAR_BIT*i)+2;
-	hbase = 010000;
-    }
-    else if (base == 2) {
-	j = (sizeof(BDIGIT)*CHAR_BIT*i)+2;
-	hbase = 020;
-    }
-    else {
-	j = 0;
-	hbase = 0;
-	rb_raise(rb_eArgError, "bignum cannot treat base %d", base);
+    j = SIZEOF_BDIGITS*CHAR_BIT*i;
+    switch (base) {
+      case 2: break;
+      case 3:
+	j = j * 647L / 1024;
+	break;
+      case 4: case 5: case 6: case 7:
+	j /= 2;
+	break;
+      case 8: case 9:
+	j /= 3;
+	break;
+      case 10: case 11: case 12: case 13: case 14: case 15:
+	j = j * 241L / 800;
+	break;
+      case 16: case 17: case 18: case 19: case 20: case 21:
+      case 22: case 23: case 24: case 25: case 26: case 27:
+      case 28: case 29: case 30: case 31:
+	j /= 4;
+	break;
+      case 32: case 33: case 34: case 35: case 36:
+	j /= 5;
+	break;
+      default:
+	rb_raise(rb_eArgError, "illegal radix %d", base);
+	break;
     }
+    j += 2;
+
+    hbase = base * base;
+#if SIZEOF_BDIGITS > 2
+    hbase *= hbase;
+#endif
 
     t = rb_big_clone(x);
     ds = BDIGITS(t);
@@ -420,10 +672,10 @@ rb_big2str(x, base)
 	    num %= hbase;
 	}
 	if (ds[i-1] == 0) i--;
-	k = 4;
+	k = SIZEOF_BDIGITS;
 	while (k--) {
 	    c = (char)(num % base);
-	    s[--j] = hexmap[(int)c];
+	    s[--j] = ruby_digitmap[(int)c];
 	    num /= base;
 	    if (i == 0 && num == 0) break;
 	}
@@ -437,10 +689,18 @@ rb_big2str(x, base)
 }
 
 static VALUE
-rb_big_to_s(x)
+rb_big_to_s(argc, argv, x)
+    int argc;
+    VALUE *argv;
     VALUE x;
 {
-    return rb_big2str(x, 10);
+    VALUE b;
+    int base;
+
+    rb_scan_args(argc, argv, "01", &b);
+    if (argc == 0) base = 10;
+    else base = NUM2INT(b);
+    return rb_big2str(x, base);
 }
 
 static unsigned long
@@ -452,7 +712,7 @@ big2ulong(x, type)
     BDIGIT_DBL num;
     BDIGIT *ds;
 
-    if (len > sizeof(long)/sizeof(BDIGIT))
+    if (len > SIZEOF_LONG/SIZEOF_BDIGITS)
 	rb_raise(rb_eRangeError, "bignum too big to convert into `%s'", type);
     ds = BDIGITS(x);
     num = 0;
@@ -486,6 +746,54 @@ rb_big2long(x)
     return num;
 }
 
+#if HAVE_LONG_LONG
+
+static unsigned LONG_LONG
+big2ull(x, type)
+    VALUE x;
+    char *type;
+{
+    long len = RBIGNUM(x)->len;
+    BDIGIT_DBL num;
+    BDIGIT *ds;
+
+    if (len > SIZEOF_LONG_LONG/SIZEOF_BDIGITS)
+	rb_raise(rb_eRangeError, "bignum too big to convert into `%s'", type);
+    ds = BDIGITS(x);
+    num = 0;
+    while (len--) {
+	num = BIGUP(num);
+	num += ds[len];
+    }
+    return num;
+}
+
+unsigned LONG_LONG
+rb_big2ull(x)
+    VALUE x;
+{
+    unsigned LONG_LONG num = big2ull(x, "unsigned long long");
+
+    if (!RBIGNUM(x)->sign) return -num;
+    return num;
+}
+
+LONG_LONG
+rb_big2ll(x)
+    VALUE x;
+{
+    unsigned LONG_LONG num = big2ull(x, "long long");
+
+    if ((LONG_LONG)num < 0 && (RBIGNUM(x)->sign
+			       || (LONG_LONG)num != LLONG_MIN)) {
+	rb_raise(rb_eRangeError, "bignum too big to convert into `long long'");
+    }
+    if (!RBIGNUM(x)->sign) return -(LONG_LONG)num;
+    return num;
+}
+
+#endif  /* HAVE_LONG_LONG */
+
 static VALUE
 dbl2big(d)
     double d;
@@ -537,6 +845,10 @@ rb_big2dbl(x)
     while (i--) {
 	d = ds[i] + BIGRAD*d;
     }
+    if (isinf(d)) {
+	rb_warn("Bignum out of Float range");
+	d = HUGE_VAL;
+    }
     if (!RBIGNUM(x)->sign) d = -d;
     return d;
 }
@@ -563,11 +875,10 @@ rb_big_cmp(x, y)
 	break;
 
       case T_FLOAT:
-	y = dbl2big(RFLOAT(y)->value);
-	break;
+	return rb_dbl_cmp(rb_big2dbl(x), RFLOAT(y)->value);
 
       default:
-	return rb_num_coerce_bin(x, y);
+	return rb_num_coerce_cmp(x, y);
     }
 
     if (RBIGNUM(x)->sign > RBIGNUM(y)->sign) return INT2FIX(1);
@@ -595,9 +906,12 @@ rb_big_eq(x, y)
       case T_BIGNUM:
 	break;
       case T_FLOAT:
-	return (rb_big2dbl(x) == RFLOAT(y)->value)?Qtrue:Qfalse;
+	if (rb_big2dbl(x) == RFLOAT(y)->value)
+	    return Qtrue;
+	else
+	    return Qfalse;
       default:
-	return Qfalse;
+	return rb_equal(y, x);
     }
     if (RBIGNUM(x)->sign != RBIGNUM(y)->sign) return Qfalse;
     if (RBIGNUM(x)->len != RBIGNUM(y)->len) return Qfalse;
@@ -605,6 +919,17 @@ rb_big_eq(x, y)
     return Qtrue;
 }
 
+static VALUE
+rb_big_eql(x, y)
+    VALUE x, y;
+{
+    if (TYPE(y) != T_BIGNUM) return Qfalse;
+    if (RBIGNUM(x)->sign != RBIGNUM(y)->sign) return Qfalse;
+    if (RBIGNUM(x)->len != RBIGNUM(y)->len) return Qfalse;
+    if (MEMCMP(BDIGITS(x),BDIGITS(y),BDIGIT,RBIGNUM(y)->len) != 0) return Qfalse;
+    return Qtrue;
+}
+
 static VALUE
 rb_big_uminus(x)
     VALUE x;
@@ -639,9 +964,8 @@ bigsub(x, y)
     VALUE z = 0;
     BDIGIT *zds;
     BDIGIT_DBL_SIGNED num;
-    long i;
-
-    i = RBIGNUM(x)->len;
+    long i = RBIGNUM(x)->len;
+    
     /* if x is larger than y, swap */
     if (RBIGNUM(x)->len < RBIGNUM(y)->len) {
 	z = x; x = y; y = z;	/* swap x y */
@@ -823,9 +1147,9 @@ bigdivrem(x, y, divp, modp)
     BDIGIT_DBL_SIGNED num;
     BDIGIT dd, q;
 
+    if (BIGZEROP(y)) rb_num_zerodiv();
     yds = BDIGITS(y);
-    if (ny == 0 && yds[0] == 0) rb_num_zerodiv();
-    if (nx < ny	|| nx == ny && BDIGITS(x)[nx - 1] < BDIGITS(y)[ny - 1]) {
+    if (nx < ny || (nx == ny && BDIGITS(x)[nx - 1] < BDIGITS(y)[ny - 1])) {
 	if (divp) *divp = rb_int2big(0);
 	if (modp) *modp = x;
 	return;
@@ -922,10 +1246,10 @@ bigdivrem(x, y, divp, modp)
 	for (i = 0;i < j;i++) zds[i] = zds[i+ny];
 	RBIGNUM(*divp)->len = i;
     }
-    if (modp) {			/* just normalize remainder */
+    if (modp) {			/* normalize remainder */
 	*modp = rb_big_clone(z);
 	zds = BDIGITS(*modp);
-	while (ny-- && !zds[ny]); ++ny;
+	while (--ny && !zds[ny]); ++ny;
 	if (dd) {
 	    t2 = 0; i = ny;
 	    while(i--) {
@@ -948,7 +1272,7 @@ bigdivmod(x, y, divp, modp)
     VALUE mod;
 
     bigdivrem(x, y, divp, &mod);
-    if (RBIGNUM(x)->sign != RBIGNUM(y)->sign && RBIGNUM(mod)->len > 0) {
+    if (RBIGNUM(x)->sign != RBIGNUM(y)->sign && !BIGZEROP(mod)) {
 	if (divp) *divp = bigadd(*divp, rb_int2big(1), 0);
 	if (modp) *modp = bigadd(mod, y, 1);
     }
@@ -983,7 +1307,6 @@ rb_big_div(x, y)
     return bignorm(z);
 }
 
-
 static VALUE
 rb_big_modulo(x, y)
     VALUE x, y;
@@ -1050,6 +1373,32 @@ rb_big_divmod(x, y)
     return rb_assoc_new(bignorm(div), bignorm(mod));
 }
 
+static VALUE
+rb_big_quo(x, y)
+    VALUE x, y;
+{
+    double dx = rb_big2dbl(x);
+    double dy;
+
+    switch (TYPE(y)) {
+      case T_FIXNUM:
+	dy = (double)FIX2LONG(y);
+	break;
+
+      case T_BIGNUM:
+	dy = rb_big2dbl(y);
+	break;
+
+      case T_FLOAT:
+	dy = RFLOAT(y)->value;
+	break;
+
+      default:
+	return rb_num_coerce_bin(x, y);
+    }
+    return rb_float_new(dx / dy);
+}
+
 VALUE
 rb_big_pow(x, y)
     VALUE x, y;
@@ -1071,20 +1420,18 @@ rb_big_pow(x, y)
       case T_FIXNUM:
 	yy = NUM2LONG(y);
 	if (yy > 0) {
-	    VALUE z;
+	    VALUE z = x;
 
-	    z = x;
 	    for (;;) {
-		yy = yy - 1;
+		yy -= 1;
 		if (yy == 0) break;
 		while (yy % 2 == 0) {
-		    yy = yy / 2;
+		    yy /= 2;
 		    x = rb_big_mul(x, x);
 		}
 		z = rb_big_mul(z, x);
 	    }
-	    if (!FIXNUM_P(z)) z = bignorm(z);
-	    return z;
+	    return bignorm(z);
 	}
 	d = (double)yy;
 	break;
@@ -1289,14 +1636,14 @@ rb_big_rshift(x, y)
 {
     BDIGIT *xds, *zds;
     int shift = NUM2INT(y);
-    int s1 = shift/BITSPERDIG;
-    int s2 = shift%BITSPERDIG;
+    long s1 = shift/BITSPERDIG;
+    long s2 = shift%BITSPERDIG;
     VALUE z;
     BDIGIT_DBL num = 0;
-    long i = RBIGNUM(x)->len;
-    long j;
+    long i, j;
 
     if (shift < 0) return rb_big_lshift(x, INT2FIX(-shift));
+
     if (s1 > RBIGNUM(x)->len) {
 	if (RBIGNUM(x)->sign)
 	    return INT2FIX(0);
@@ -1310,6 +1657,9 @@ rb_big_rshift(x, y)
     xds = BDIGITS(x);
     i = RBIGNUM(x)->len; j = i - s1;
     z = bignew(j, RBIGNUM(x)->sign);
+    if (!RBIGNUM(x)->sign) {
+	num = ((BDIGIT_DBL)~0) << BITSPERDIG;
+    }
     zds = BDIGITS(z);
     while (i--, j--) {
 	num = (num | xds[i]) >> s2;
@@ -1328,7 +1678,7 @@ rb_big_aref(x, y)
 {
     BDIGIT *xds;
     int shift;
-    int s1, s2;
+    long s1, s2;
 
     if (TYPE(y) == T_BIGNUM) {
 	if (!RBIGNUM(y)->sign || RBIGNUM(x)->sign)
@@ -1365,7 +1715,7 @@ rb_big_hash(x)
     for (i=0; i<len; i++) {
 	key ^= *digits++;
     }
-    return INT2FIX(key);
+    return LONG2FIX(key);
 }
 
 static VALUE
@@ -1377,7 +1727,7 @@ rb_big_coerce(x, y)
     }
     else {
 	rb_raise(rb_eTypeError, "Can't coerce %s to Bignum",
-		 rb_class2name(CLASS_OF(y)));
+		 rb_obj_classname(y));
     }
     /* not reached */
     return Qnil;
@@ -1394,22 +1744,20 @@ rb_big_abs(x)
     return x;
 }
 
-/* !!!warnig!!!!
-   this is not really a random number!!
-*/
-
 VALUE
-rb_big_rand(max, rand)
+rb_big_rand(max, rand_buf)
     VALUE max;
-    double rand;
+    double *rand_buf;
 {
     VALUE v;
-    long len;
+    long len = RBIGNUM(max)->len;
 
-    len = RBIGNUM(max)->len;
+    if (BIGZEROP(max)) {
+	return rb_float_new(rand_buf[0]);
+    }
     v = bignew(len,1);
     while (len--) {
-	BDIGITS(v)[len] = ((BDIGIT)~0) * rand;
+	BDIGITS(v)[len] = ((BDIGIT)~0) * rand_buf[len];
     }
 
     return rb_big_modulo((VALUE)v, max);
@@ -1419,14 +1767,7 @@ static VALUE
 rb_big_size(big)
     VALUE big;
 {
-    return INT2FIX(RBIGNUM(big)->len*sizeof(BDIGIT));
-}
-
-static VALUE
-rb_big_zero_p(big)
-    VALUE big;
-{
-    return Qfalse;
+    return LONG2FIX(RBIGNUM(big)->len*SIZEOF_BDIGITS);
 }
 
 void
@@ -1434,7 +1775,7 @@ Init_Bignum()
 {
     rb_cBignum = rb_define_class("Bignum", rb_cInteger);
 
-    rb_define_method(rb_cBignum, "to_s", rb_big_to_s, 0);
+    rb_define_method(rb_cBignum, "to_s", rb_big_to_s, -1);
     rb_define_method(rb_cBignum, "coerce", rb_big_coerce, 1);
     rb_define_method(rb_cBignum, "-@", rb_big_uminus, 0);
     rb_define_method(rb_cBignum, "+", rb_big_plus, 1);
@@ -1442,9 +1783,11 @@ Init_Bignum()
     rb_define_method(rb_cBignum, "*", rb_big_mul, 1);
     rb_define_method(rb_cBignum, "/", rb_big_div, 1);
     rb_define_method(rb_cBignum, "%", rb_big_modulo, 1);
+    rb_define_method(rb_cBignum, "div", rb_big_div, 1);
     rb_define_method(rb_cBignum, "divmod", rb_big_divmod, 1);
     rb_define_method(rb_cBignum, "modulo", rb_big_modulo, 1);
     rb_define_method(rb_cBignum, "remainder", rb_big_remainder, 1);
+    rb_define_method(rb_cBignum, "quo", rb_big_quo, 1);
     rb_define_method(rb_cBignum, "**", rb_big_pow, 1);
     rb_define_method(rb_cBignum, "&", rb_big_and, 1);
     rb_define_method(rb_cBignum, "|", rb_big_or, 1);
@@ -1456,11 +1799,9 @@ Init_Bignum()
 
     rb_define_method(rb_cBignum, "<=>", rb_big_cmp, 1);
     rb_define_method(rb_cBignum, "==", rb_big_eq, 1);
-    rb_define_method(rb_cBignum, "===", rb_big_eq, 1);
-    rb_define_method(rb_cBignum, "eql?", rb_big_eq, 1);
+    rb_define_method(rb_cBignum, "eql?", rb_big_eql, 1);
     rb_define_method(rb_cBignum, "hash", rb_big_hash, 0);
     rb_define_method(rb_cBignum, "to_f", rb_big_to_f, 0);
     rb_define_method(rb_cBignum, "abs", rb_big_abs, 0);
     rb_define_method(rb_cBignum, "size", rb_big_size, 0);
-    rb_define_method(rb_cBignum, "zero?", rb_big_zero_p, 0);
 }
