@@ -3,10 +3,10 @@
   compar.c -
 
   $Author: matz $
-  $Date: 2000/09/12 05:37:19 $
+  $Date: 2003/05/19 15:44:42 $
   created at: Thu Aug 26 14:39:48 JST 1993
 
-  Copyright (C) 1993-2000 Yukihiro Matsumoto
+  Copyright (C) 1993-2003 Yukihiro Matsumoto
 
 **********************************************************************/
 
@@ -16,34 +16,55 @@ VALUE rb_mComparable;
 
 static ID cmp;
 
-static VALUE
-cmp_eq(a)
-    VALUE *a;
+int
+rb_cmpint(val, a, b)
+    VALUE val, a, b;
 {
-    VALUE c = rb_funcall(a[0], cmp, 1, a[1]);
-    int t = NUM2INT(c);
-
-    if (t == 0) return Qtrue;
-    return Qfalse;
+    if (NIL_P(val)) {
+	rb_cmperr(a, b);
+    }
+    if (FIXNUM_P(val)) return FIX2INT(val);
+    if (TYPE(val) == T_BIGNUM) {
+	if (RBIGNUM(val)->sign) return 1;
+	return -1;
+    }
+    if (RTEST(rb_funcall(val, '>', 1, INT2FIX(0)))) return 1;
+    if (RTEST(rb_funcall(val, '<', 1, INT2FIX(0)))) return -1;
+    return 0;
 }
 
-static VALUE
-cmp_failed()
+void
+rb_cmperr(x, y)
+    VALUE x, y;
 {
-    return Qfalse;
+    const char *classname;
+
+    if (SPECIAL_CONST_P(y)) {
+	y = rb_inspect(y);
+	classname = StringValuePtr(y);
+    }
+    else {
+	classname = rb_obj_classname(y);
+    }
+    rb_raise(rb_eArgError, "comparison of %s with %s failed",
+	     rb_obj_classname(x), classname);
 }
 
+#define cmperr() (rb_cmperr(x, y), Qnil)
+
 static VALUE
 cmp_equal(x, y)
     VALUE x, y;
 {
-    VALUE a[2];
+    int c;
 
     if (x == y) return Qtrue;
 
-    a[0] = x; a[1] = y;
-    return rb_rescue2(cmp_eq, (VALUE)a, cmp_failed, 0,
-		      rb_eStandardError, rb_eNameError, 0);
+    c  = rb_funcall(x, cmp, 1, y);
+    if (NIL_P(c)) return Qnil;
+    if (c == INT2FIX(0)) return Qtrue;
+    if (rb_cmpint(c, x, y) == 0) return Qtrue;
+    return Qfalse;
 }
 
 static VALUE
@@ -51,9 +72,9 @@ cmp_gt(x, y)
     VALUE x, y;
 {
     VALUE c = rb_funcall(x, cmp, 1, y);
-    int t = NUM2INT(c);
 
-    if (t > 0) return Qtrue;
+    if (NIL_P(c)) return cmperr();
+    if (rb_cmpint(c, x, y) > 0) return Qtrue;
     return Qfalse;
 }
 
@@ -62,9 +83,9 @@ cmp_ge(x, y)
     VALUE x, y;
 {
     VALUE c = rb_funcall(x, cmp, 1, y);
-    int t = NUM2INT(c);
 
-    if (t >= 0) return Qtrue;
+    if (NIL_P(c)) return cmperr();
+    if (rb_cmpint(c, x, y) >= 0) return Qtrue;
     return Qfalse;
 }
 
@@ -73,9 +94,9 @@ cmp_lt(x, y)
     VALUE x, y;
 {
     VALUE c = rb_funcall(x, cmp, 1, y);
-    int t = NUM2INT(c);
 
-    if (t < 0) return Qtrue;
+    if (NIL_P(c)) return cmperr();
+    if (rb_cmpint(c, x, y) < 0) return Qtrue;
     return Qfalse;
 }
 
@@ -84,9 +105,9 @@ cmp_le(x, y)
     VALUE x, y;
 {
     VALUE c = rb_funcall(x, cmp, 1, y);
-    int t = NUM2INT(c);
 
-    if (t <= 0) return Qtrue;
+    if (NIL_P(c)) return cmperr();
+    if (rb_cmpint(c, x, y) <= 0) return Qtrue;
     return Qfalse;
 }
 
@@ -94,13 +115,8 @@ static VALUE
 cmp_between(x, min, max)
     VALUE x, min, max;
 {
-    VALUE c = rb_funcall(x, cmp, 1, min);
-    long t = NUM2LONG(c);
-    if (t < 0) return Qfalse;
-
-    c = rb_funcall(x, cmp, 1, max);
-    t = NUM2LONG(c);
-    if (t > 0) return Qfalse;
+    if (RTEST(cmp_lt(x, min))) return Qfalse;
+    if (RTEST(cmp_gt(x, max))) return Qfalse;
     return Qtrue;
 }
 
