@@ -2,11 +2,11 @@
 
   dir.c -
 
-  $Author: matz $
-  $Date: 2002/01/21 07:43:39 $
+  $Author: nobu $
+  $Date: 2003/06/23 08:41:07 $
   created at: Wed Jan  5 09:51:01 JST 1994
 
-  Copyright (C) 1993-2000 Yukihiro Matsumoto
+  Copyright (C) 1993-2003 Yukihiro Matsumoto
   Copyright (C) 2000  Network Applied Communication Laboratory, Inc.
   Copyright (C) 2000  Information-technology Promotion Agency, Japan
 
@@ -17,19 +17,14 @@
 #include <sys/types.h>
 #include <sys/stat.h>
 
-#ifdef HAVE_SYS_PARAM_H
-# include <sys/param.h>
-#else
-# define MAXPATHLEN 1024
-#endif
 #ifdef HAVE_UNISTD_H
 #include <unistd.h>
 #endif
 
-#if defined HAVE_DIRENT_H && !defined NT
+#if defined HAVE_DIRENT_H && !defined _WIN32
 # include <dirent.h>
 # define NAMLEN(dirent) strlen((dirent)->d_name)
-#elif defined HAVE_DIRECT_H && !defined NT
+#elif defined HAVE_DIRECT_H && !defined _WIN32
 # include <direct.h>
 # define NAMLEN(dirent) strlen((dirent)->d_name)
 #else
@@ -44,7 +39,7 @@
 # if HAVE_NDIR_H
 #  include <ndir.h>
 # endif
-# if defined(NT)
+# ifdef _WIN32
 #  include "win32/dir.h"
 # endif
 #endif
@@ -61,30 +56,40 @@ char *strchr _((char*,char));
 
 #include <ctype.h>
 
+#include "util.h"
+
 #ifndef HAVE_LSTAT
-#define lstat rb_sys_stat
+#define lstat(path,st) stat(path,st)
 #endif
 
 #define FNM_NOESCAPE	0x01
 #define FNM_PATHNAME	0x02
-#define FNM_PERIOD	0x04
-#define FNM_NOCASE	0x08
+#define FNM_DOTMATCH	0x04
+#define FNM_CASEFOLD	0x08
 
 #define FNM_NOMATCH	1
 #define FNM_ERROR	2
 
 #define downcase(c) (nocase && ISUPPER(c) ? tolower(c) : (c))
 
+#ifndef CharNext		/* defined as CharNext[AW] on Windows. */
+# if defined(DJGPP)
+#   define CharNext(p) ((p) + mblen(p, MB_CUR_MAX))
+# else
+#   define CharNext(p) ((p) + 1)
+# endif
+#endif
+
 #if defined DOSISH
 #define isdirsep(c) ((c) == '/' || (c) == '\\')
-static char *
+static const char *
 find_dirsep(s)
-    char *s;
+    const char *s;
 {
     while (*s) {
 	if (isdirsep(*s))
 	    return s;
-	s++;
+	s = CharNext(s);
     }
     return 0;
 }
@@ -100,7 +105,7 @@ range(pat, test, flags)
     int flags;
 {
     int not, ok = 0;
-    int nocase = flags & FNM_NOCASE;
+    int nocase = flags & FNM_CASEFOLD;
     int escape = !(flags & FNM_NOESCAPE);
 
     not = *pat == '!' || *pat == '^';
@@ -144,8 +149,8 @@ fnmatch(pat, string, flags)
     const char *s = string;
     int escape = !(flags & FNM_NOESCAPE);
     int pathname = flags & FNM_PATHNAME;
-    int period = flags & FNM_PERIOD;
-    int nocase = flags & FNM_NOCASE;
+    int period = !(flags & FNM_DOTMATCH);
+    int nocase = flags & FNM_CASEFOLD;
 
     while (c = *pat++) {
 	switch (c) {
@@ -169,8 +174,10 @@ fnmatch(pat, string, flags)
 	    }
 	    else if (ISDIRSEP(c)) {
 		s = find_dirsep(s);
-		if (s)
+		if (s) {
+                    s++;
 		    break;
+                }
 		return FNM_NOMATCH;
 	    }
 
@@ -179,7 +186,7 @@ fnmatch(pat, string, flags)
 	    pat--;
 	    while (*s) {
 		if ((c == '[' || downcase(*s) == test) &&
-		    !fnmatch(pat, s, flags & ~FNM_PERIOD))
+		    !fnmatch(pat, s, flags | FNM_DOTMATCH))
 		    return 0;
 		else if (ISDIRSEP(*s))
 		    break;
@@ -227,24 +234,30 @@ fnmatch(pat, string, flags)
 
 VALUE rb_cDir;
 
+struct dir_data {
+    DIR *dir;
+    char *path;
+};
+
 static void
 free_dir(dir)
-    DIR *dir;
+    struct dir_data *dir;
 {
-    if (dir) closedir(dir);
+    if (dir && dir->dir) closedir(dir->dir);
 }
 
 static VALUE dir_close _((VALUE));
 
+static VALUE dir_s_alloc _((VALUE));
 static VALUE
-dir_s_new(argc, argv, klass)
-    int argc;
-    VALUE *argv;
+dir_s_alloc(klass)
     VALUE klass;
 {
-    VALUE obj = Data_Wrap_Struct(klass, 0, free_dir, 0);
+    struct dir_data *dirp;
+    VALUE obj = Data_Make_Struct(klass, struct dir_data, 0, free_dir, dirp);
 
-    rb_obj_call_init(obj, argc, argv);
+    dirp->dir = NULL;
+    dirp->path = NULL;
 
     return obj;
 }
@@ -253,22 +266,25 @@ static VALUE
 dir_initialize(dir, dirname)
     VALUE dir, dirname;
 {
-    DIR *dirp;
+    struct dir_data *dp;
 
-    Check_SafeStr(dirname);
-    if (DATA_PTR(dir)) closedir(DATA_PTR(dir));
-    DATA_PTR(dir) = NULL;
-    dirp = opendir(RSTRING(dirname)->ptr);
-    if (dirp == NULL) {
+    SafeStringValue(dirname);
+    Data_Get_Struct(dir, struct dir_data, dp);
+    if (dp->dir) closedir(dp->dir);
+    if (dp->path) free(dp->path);
+    dp->dir = NULL;
+    dp->path = NULL;
+    dp->dir = opendir(RSTRING(dirname)->ptr);
+    if (dp->dir == NULL) {
 	if (errno == EMFILE || errno == ENFILE) {
 	    rb_gc();
-	    dirp = opendir(RSTRING(dirname)->ptr);
+	    dp->dir = opendir(RSTRING(dirname)->ptr);
 	}
-	if (dirp == NULL) {
+	if (dp->dir == NULL) {
 	    rb_sys_fail(RSTRING(dirname)->ptr);
 	}
     }
-    DATA_PTR(dir) = dirp;
+    dp->path = strdup(RSTRING(dirname)->ptr);
 
     return dir;
 }
@@ -277,12 +293,12 @@ static VALUE
 dir_s_open(klass, dirname)
     VALUE klass, dirname;
 {
-    VALUE dir = Data_Wrap_Struct(klass, 0, free_dir, 0);
+    struct dir_data *dp;
+    VALUE dir = Data_Make_Struct(klass, struct dir_data, 0, free_dir, dp);
 
     dir_initialize(dir, dirname);
     if (rb_block_given_p()) {
-	rb_ensure(rb_yield, dir, dir_close, dir);
-	return Qnil;
+	return rb_ensure(rb_yield, dir, dir_close, dir);
     }
 
     return dir;
@@ -294,23 +310,35 @@ dir_closed()
     rb_raise(rb_eIOError, "closed directory");
 }
 
-#define GetDIR(obj, dirp) {\
-    Data_Get_Struct(obj, DIR, dirp);\
-    if (dirp == NULL) dir_closed();\
+#define GetDIR(obj, dirp) do {\
+    Data_Get_Struct(obj, struct dir_data, dirp);\
+    if (dirp->dir == NULL) dir_closed();\
+} while (0)
+
+static VALUE
+dir_path(dir)
+    VALUE dir;
+{
+    struct dir_data *dirp;
+
+    GetDIR(dir, dirp);
+    if (!dirp->path) return Qnil;
+    return rb_str_new2(dirp->path);
 }
 
 static VALUE
 dir_read(dir)
     VALUE dir;
 {
-    DIR *dirp;
+    struct dir_data *dirp;
     struct dirent *dp;
 
     GetDIR(dir, dirp);
     errno = 0;
-    dp = readdir(dirp);
-    if (dp)
+    dp = readdir(dirp->dir);
+    if (dp) {
 	return rb_tainted_str_new(dp->d_name, NAMLEN(dp));
+    }
     else if (errno == 0) {	/* end of stream */
 	return Qnil;
     }
@@ -324,13 +352,13 @@ static VALUE
 dir_each(dir)
     VALUE dir;
 {
-    DIR *dirp;
+    struct dir_data *dirp;
     struct dirent *dp;
 
     GetDIR(dir, dirp);
-    for (dp = readdir(dirp); dp != NULL; dp = readdir(dirp)) {
+    for (dp = readdir(dirp->dir); dp != NULL; dp = readdir(dirp->dir)) {
 	rb_yield(rb_tainted_str_new(dp->d_name, NAMLEN(dp)));
-	if (DATA_PTR(dir) == NULL) dir_closed();
+	if (dirp->dir == NULL) dir_closed();
     }
     return dir;
 }
@@ -340,11 +368,11 @@ dir_tell(dir)
     VALUE dir;
 {
 #ifdef HAVE_TELLDIR
-    DIR *dirp;
+    struct dir_data *dirp;
     long pos;
 
     GetDIR(dir, dirp);
-    pos = telldir(dirp);
+    pos = telldir(dirp->dir);
     return rb_int2inum(pos);
 #else
     rb_notimplement();
@@ -355,11 +383,11 @@ static VALUE
 dir_seek(dir, pos)
     VALUE dir, pos;
 {
-    DIR *dirp;
+    struct dir_data *dirp;
 
 #ifdef HAVE_SEEKDIR
     GetDIR(dir, dirp);
-    seekdir(dirp, NUM2INT(pos));
+    seekdir(dirp->dir, NUM2INT(pos));
     return dir;
 #else
     rb_notimplement();
@@ -371,17 +399,17 @@ dir_set_pos(dir, pos)
     VALUE dir, pos;
 {
     dir_seek(dir, pos);
-    return dir;
+    return pos;
 }
 
 static VALUE
 dir_rewind(dir)
     VALUE dir;
 {
-    DIR *dirp;
+    struct dir_data *dirp;
 
     GetDIR(dir, dirp);
-    rewinddir(dirp);
+    rewinddir(dirp->dir);
     return dir;
 }
 
@@ -389,28 +417,50 @@ static VALUE
 dir_close(dir)
     VALUE dir;
 {
-    DIR *dirp;
+    struct dir_data *dirp;
 
-    Data_Get_Struct(dir, DIR, dirp);
-    if (dirp == NULL) dir_closed();
-    closedir(dirp);
-    DATA_PTR(dir) = NULL;
+    GetDIR(dir, dirp);
+    closedir(dirp->dir);
+    dirp->dir = NULL;
 
     return Qnil;
 }
 
+static void
+dir_chdir(path)
+    const char *path;
+{
+    if (chdir(path) < 0)
+	rb_sys_fail(path);
+}
+
+static int chdir_blocking = 0;
+static VALUE chdir_thread = Qnil;
+
+static VALUE
+chdir_restore(path)
+    char *path;
+{
+    chdir_blocking--;
+    if (chdir_blocking == 0)
+	chdir_thread = Qnil;
+    dir_chdir(path);
+    free(path);
+    return Qnil;
+}
+
 static VALUE
 dir_s_chdir(argc, argv, obj)
     int argc;
     VALUE *argv;
     VALUE obj;
 {
-    VALUE path;
+    VALUE path = Qnil;
     char *dist = "";
 
     rb_secure(2);
     if (rb_scan_args(argc, argv, "01", &path) == 1) {
-	Check_SafeStr(path);
+	SafeStringValue(path);
 	dist = RSTRING(path)->ptr;
     }
     else {
@@ -421,32 +471,34 @@ dir_s_chdir(argc, argv, obj)
 	}
     }
 
-    if (chdir(dist) < 0)
-	rb_sys_fail(dist);
+    if (chdir_blocking > 0) {
+	if (!rb_block_given_p() || rb_thread_current() != chdir_thread)
+	    rb_warn("conflicting chdir during another chdir block");
+    }
+
+    if (rb_block_given_p()) {
+	char *cwd = my_getcwd();
+	chdir_blocking++;
+	if (chdir_thread == Qnil)
+	    chdir_thread = rb_thread_current();
+	dir_chdir(dist);
+	return rb_ensure(rb_yield, path, chdir_restore, (VALUE)cwd);
+    }
+    dir_chdir(dist);
 
     return INT2FIX(0);
 }
 
-static char *
-my_getcwd()
-{
-    int size = MAXPATHLEN;
-    char *buf = xmalloc(size);
-
-    while (!getcwd(buf, size)) {
-	if (errno != ERANGE) rb_sys_fail(buf);
-	size *= 2;
-	buf = xrealloc(buf, size);
-    }
-    return buf;
-}
-
 static VALUE
 dir_s_getwd(dir)
     VALUE dir;
 {
-    char *path = my_getcwd();
-    VALUE cwd = rb_tainted_str_new2(path);
+    char *path;
+    VALUE cwd;
+
+    rb_secure(4);
+    path = my_getcwd();
+    cwd = rb_tainted_str_new2(path);
 
     free(path);
     return cwd;
@@ -458,7 +510,7 @@ dir_s_chroot(dir, path)
 {
 #if defined(HAVE_CHROOT) && !defined(__CHECKER__)
     rb_secure(2);
-    Check_SafeStr(path);
+    SafeStringValue(path);
 
     if (chroot(RSTRING(path)->ptr) == -1)
 	rb_sys_fail(RSTRING(path)->ptr);
@@ -486,9 +538,9 @@ dir_s_mkdir(argc, argv, obj)
 	mode = 0777;
     }
 
-    Check_SafeStr(path);
+    SafeStringValue(path);
     rb_secure(2);
-#if !defined(NT)
+#ifndef _WIN32
     if (mkdir(RSTRING(path)->ptr, mode) == -1)
 	rb_sys_fail(RSTRING(path)->ptr);
 #else
@@ -503,7 +555,7 @@ static VALUE
 dir_s_rmdir(obj, dir)
     VALUE obj, dir;
 {
-    Check_SafeStr(dir);
+    SafeStringValue(dir);
     rb_secure(2);
     if (rmdir(RSTRING(dir)->ptr) < 0)
 	rb_sys_fail(RSTRING(dir)->ptr);
@@ -528,8 +580,8 @@ has_magic(s, send, flags)
 	  case '*':
 	    return Qtrue;
 
-	  case '[':		/* Only accept an open brace if there is a close */
-	    open++;		/* brace to match it.  Bracket expressions must be */
+	  case '[':	/* Only accept an open brace if there is a close */
+	    open++;	/* brace to match it.  Bracket expressions must be */
 	    continue;	/* complete, according to Posix.2 */
 	  case ']':
 	    if (open)
@@ -557,7 +609,7 @@ extract_path(p, pend)
     alloc = ALLOC_N(char, len+1);
     memcpy(alloc, p, len);
     if (len > 1 && pend[-1] == '/'
-#if defined DOSISH
+#if defined DOSISH_DRIVE_LETTER
     && pend[-2] != ':'
 #endif
     ) {
@@ -582,12 +634,62 @@ extract_elem(path)
     return extract_path(path, pend);
 }
 
+static void
+remove_backslashes(p)
+    char *p;
+{
+    char *pend = p + strlen(p);
+    char *t = p;
+
+    while (p < pend) {
+	if (*p == '\\') {
+	    if (++p == pend) break;
+	}
+	*t++ = *p++;
+    }
+    *t = '\0';
+}
+
 #ifndef S_ISDIR
 #   define S_ISDIR(m) ((m & S_IFMT) == S_IFDIR)
 #endif
 
-void
-rb_glob_helper(path, sub, flags, func, arg)
+struct glob_args {
+    void (*func) _((const char*, VALUE));
+    const char *c;
+    VALUE v;
+};
+
+static VALUE glob_func_caller _((VALUE));
+
+static VALUE
+glob_func_caller(val)
+    VALUE val;
+{
+    struct glob_args *args = (struct glob_args *)val;
+    (*args->func)(args->c, args->v);
+    return Qnil;
+}
+
+static int
+glob_call_func(func, path, arg)
+    void (*func) _((const char*, VALUE));
+    const char *path;
+    VALUE arg;
+{
+    int status;
+    struct glob_args args;
+
+    args.func = func;
+    args.c = path;
+    args.v = arg;
+
+    rb_protect(glob_func_caller, (VALUE)&args, &status);
+    return status;
+}
+
+static int
+glob_helper(path, sub, flags, func, arg)
     char *path;
     char *sub;
     int flags;
@@ -596,16 +698,28 @@ rb_glob_helper(path, sub, flags, func, arg)
 {
     struct stat st;
     char *p, *m;
+    int status = 0;
 
     p = sub ? sub : path;
     if (!has_magic(p, 0, flags)) {
-	if (rb_sys_stat(path, &st) == 0) {
-	    (*func)(path, arg);
+#if defined DOSISH
+	remove_backslashes(path);
+#else
+	if (!(flags & FNM_NOESCAPE)) remove_backslashes(p);
+#endif
+	if (lstat(path, &st) == 0) {
+	    status = glob_call_func(func, path, arg);
+	    if (status) return status;
 	}
-	return;
+	else if (errno != ENOENT) {
+	    /* In case stat error is other than ENOENT and
+	       we may want to know what is wrong. */
+	    rb_sys_warning(path);
+	}
+	return 0;
     }
 
-    while (p) {
+    while (p && !status) {
 	if (*p == '/') p++;
 	m = strchr(p, '/');
 	if (has_magic(p, m, flags)) {
@@ -624,8 +738,10 @@ rb_glob_helper(path, sub, flags, func, arg)
 	    else dir = base;
 
 	    magic = extract_elem(p);
-	    if (rb_sys_stat(dir, &st) < 0) {
+	    if (stat(dir, &st) < 0) {
+	        if (errno != ENOENT) rb_sys_warning(dir);
 	        free(base);
+	        free(magic);
 	        break;
 	    }
 	    if (S_ISDIR(st.st_mode)) {
@@ -633,25 +749,29 @@ rb_glob_helper(path, sub, flags, func, arg)
 		    int n = strlen(base);
 		    recursive = 1;
 		    buf = ALLOC_N(char, n+strlen(m)+3);
-		    sprintf(buf, "%s%s%s", base, (*base)?"":".", m);
-		    rb_glob_helper(buf, buf+n, flags, func, arg);
+		    sprintf(buf, "%s%s", base, *base ? m : m+1);
+		    status = glob_helper(buf, buf+n, flags, func, arg);
 		    free(buf);
+		    if (status) goto finalize;
 		}
 		dirp = opendir(dir);
 		if (dirp == NULL) {
+		    rb_sys_warning(dir);
 		    free(base);
+		    free(magic);
 		    break;
 		}
 	    }
 	    else {
-	      free(base);
-	      break;
+		free(base);
+		free(magic);
+		break;
 	    }
-	    
-#if defined DOSISH
+
+#if defined DOSISH_DRIVE_LETTER
 #define BASE (*base && !((isdirsep(*base) && !base[1]) || (base[1] == ':' && isdirsep(base[2]) && !base[3])))
 #else
-#define BASE (*base && !(*base == '/' && !base[1]))
+#define BASE (*base && !(isdirsep(*base) && !base[1]))
 #endif
 
 	    for (dp = readdir(dirp); dp != NULL; dp = readdir(dirp)) {
@@ -659,25 +779,31 @@ rb_glob_helper(path, sub, flags, func, arg)
 		    if (strcmp(".", dp->d_name) == 0 || strcmp("..", dp->d_name) == 0)
 			continue;
 		    buf = ALLOC_N(char, strlen(base)+NAMLEN(dp)+strlen(m)+6);
-		    sprintf(buf, "%s%s%s", base, (BASE)?"/":"", dp->d_name);
+		    sprintf(buf, "%s%s%s", base, (BASE) ? "/" : "", dp->d_name);
 		    if (lstat(buf, &st) < 0) {
+			if (errno != ENOENT) rb_sys_warning(buf);
+			free(buf);
 			continue;
 		    }
 		    if (S_ISDIR(st.st_mode)) {
 			char *t = buf+strlen(buf);
 		        strcpy(t, "/**");
 			strcpy(t+3, m);
-			rb_glob_helper(buf, t, flags, func, arg);
+			status = glob_helper(buf, t, flags, func, arg);
+			free(buf);
+			if (status) break;
+			continue;
 		    }
 		    free(buf);
 		    continue;
 		}
 		if (fnmatch(magic, dp->d_name, flags) == 0) {
 		    buf = ALLOC_N(char, strlen(base)+NAMLEN(dp)+2);
-		    sprintf(buf, "%s%s%s", base, (BASE)?"/":"", dp->d_name);
+		    sprintf(buf, "%s%s%s", base, (BASE) ? "/" : "", dp->d_name);
 		    if (!m) {
-			(*func)(buf, arg);
+			status = glob_call_func(func, buf, arg);
 			free(buf);
+			if (status) break;
 			continue;
 		    }
 		    tmp = ALLOC(struct d_link);
@@ -687,50 +813,72 @@ rb_glob_helper(path, sub, flags, func, arg)
 		}
 	    }
 	    closedir(dirp);
+	  finalize:
 	    free(base);
 	    free(magic);
-	    while (link) {
-		stat(link->path, &st); /* should success */
-		if (S_ISDIR(st.st_mode)) {
-		    int len = strlen(link->path);
-		    int mlen = strlen(m);
-		    char *t = ALLOC_N(char, len+mlen+1);
+	    if (link) {
+		while (link) {
+		    if (status == 0) {
+			if (stat(link->path, &st) == 0) {
+			    if (S_ISDIR(st.st_mode)) {
+				int len = strlen(link->path);
+				int mlen = strlen(m);
+				char *t = ALLOC_N(char, len+mlen+1);
 
-		    sprintf(t, "%s%s", link->path, m);
-		    rb_glob_helper(t, t+len, flags, func, arg);
-		    free(t);
+				sprintf(t, "%s%s", link->path, m);
+				status = glob_helper(t, t+len, flags, func, arg);
+				free(t);
+			    }
+			}
+			else {
+			    rb_sys_warning(link->path);
+			}
+		    }
+		    tmp = link;
+		    link = link->next;
+		    free(tmp->path);
+		    free(tmp);
 		}
-		tmp = link;
-		link = link->next;
-		free(tmp->path);
-		free(tmp);
+		break;
 	    }
 	}
 	p = m;
     }
+    return status;
+}
+
+static void
+rb_glob2(path, flags, func, arg)
+    char *path;
+    int flags;
+    void (*func) _((const char*, VALUE));
+    VALUE arg;
+{
+    int status = glob_helper(path, 0, flags, func, arg);
+    if (status) rb_jump_tag(status);
 }
 
 void
 rb_glob(path, func, arg)
     char *path;
-    void (*func)();
+    void (*func) _((const char*, VALUE));
     VALUE arg;
 {
-    rb_glob_helper(path, 0, FNM_PERIOD, func, arg);
+    rb_glob2(path, 0, func, arg);
 }
 
 void
-rb_iglob(path, func, arg)
+rb_globi(path, func, arg)
     char *path;
     void (*func) _((const char*, VALUE));
     VALUE arg;
 {
-    rb_glob_helper(path, 0, FNM_PERIOD|FNM_NOCASE, func, arg);
+    rb_glob2(path, FNM_CASEFOLD, func, arg);
 }
 
 static void
 push_pattern(path, ary)
-    char *path;
+    const char *path;
     VALUE ary;
 {
     VALUE str = rb_tainted_str_new2(path);
@@ -744,19 +892,21 @@ push_pattern(path, ary)
 }
 
 static void
-push_globs(ary, s)
+push_globs(ary, s, flags)
     VALUE ary;
     char *s;
+    int flags;
 {
-    rb_glob(s, push_pattern, ary);
+    rb_glob2(s, flags, push_pattern, ary);
 }
 
 static void
-push_braces(ary, s)
+push_braces(ary, s, flags)
     VALUE ary;
     char *s;
+    int flags;
 {
-    char buffer[MAXPATHLEN], *buf = buffer;
+    char *buf;
     char *p, *t, *b;
     char *lbrace, *rbrace;
     int nest = 0;
@@ -779,10 +929,9 @@ push_braces(ary, s)
 	p++;
     }
 
-    if (lbrace) {
+    if (lbrace && rbrace) {
 	int len = strlen(s);
-	if (len >= MAXPATHLEN)
-	    buf = xmalloc(len + 1);
+	buf = xmalloc(len + 1);
 	memcpy(buf, s, lbrace-s);
 	b = buf + (lbrace-s);
 	p = lbrace;
@@ -794,64 +943,91 @@ push_braces(ary, s)
 	    }
 	    memcpy(b, t, p-t);
 	    strcpy(b+(p-t), rbrace+1);
-	    push_braces(ary, buf);
+	    push_braces(ary, buf, flags);
 	}
-	if (buf != buffer)
-	    free(buf);
+	free(buf);
     }
     else {
-	push_globs(ary, s);
+	push_globs(ary, s, flags);
     }
 }
 
-#define isdelim(c) ((c)==' '||(c)=='\t'||(c)=='\n'||(c)=='\0')
+#define isdelim(c) ((c)=='\0')
 
 static VALUE
-dir_s_glob(dir, str)
-    VALUE dir, str;
+rb_push_glob(str, flags)
+    VALUE str;
+    int flags;
 {
     char *p, *pend;
-    char buffer[MAXPATHLEN], *buf;
+    char *buf;
     char *t;
-    int nest;
-    VALUE ary = 0;
+    int nest, maxnest;
+    int noescape = flags & FNM_NOESCAPE;
+    VALUE ary;
 
-    Check_SafeStr(str);
-    if (!rb_block_given_p()) {
-	ary = rb_ary_new();
-    }
-    if (RSTRING(str)->len >= MAXPATHLEN)
-	buf = xmalloc(RSTRING(str)->len + 1);
+    if (rb_block_given_p())
+	ary = 0;
     else
-	buf = buffer;
+	ary = rb_ary_new();
+
+    SafeStringValue(str);
+    buf = xmalloc(RSTRING(str)->len + 1);
 
     p = RSTRING(str)->ptr;
     pend = p + RSTRING(str)->len;
 
     while (p < pend) {
 	t = buf;
-	nest = 0;
+	nest = maxnest = 0;
 	while (p < pend && isdelim(*p)) p++;
 	while (p < pend && !isdelim(*p)) {
-	    if (*p == '{') nest+=2;
-	    if (*p == '}') nest+=3;
+	    if (*p == '{') nest++, maxnest++;
+	    if (*p == '}') nest--;
+	    if (!noescape && *p == '\\') {
+		*t++ = *p++;
+		if (p == pend) break;
+	    }
 	    *t++ = *p++;
 	}
 	*t = '\0';
-	if (nest == 0) {
-	    push_globs(ary, buf);
+	if (maxnest == 0) {
+	    push_globs(ary, buf, flags);
 	}
-	else if (nest % 5 == 0) {
-	    push_braces(ary, buf);
+	else if (nest == 0) {
+	    push_braces(ary, buf, flags);
 	}
 	/* else unmatched braces */
     }
-    if (buf != buffer)
-	free(buf);
-    if (!ary) return Qnil;
+    free(buf);
+
     return ary;
 }
 
+static VALUE
+dir_s_aref(obj, str)
+    VALUE obj, str;
+{
+    return rb_push_glob(str, 0);
+}
+
+static VALUE
+dir_s_glob(argc, argv, obj)
+    int argc;
+    VALUE *argv;
+    VALUE obj;
+{
+    VALUE str, rflags;
+    int flags;
+
+    if (rb_scan_args(argc, argv, "11", &str, &rflags) == 2)
+	flags = NUM2INT(rflags);
+    else
+	flags = 0;
+
+    return rb_push_glob(str, flags);
+}
+
 static VALUE
 dir_foreach(io, dirname)
     VALUE io, dirname;
@@ -873,6 +1049,30 @@ dir_entries(io, dirname)
     return rb_ensure(rb_Array, dir, dir_close, dir);
 }
 
+static VALUE
+file_s_fnmatch(argc, argv, obj)
+    int argc;
+    VALUE *argv;
+    VALUE obj;
+{
+    VALUE pattern, path;
+    VALUE rflags;
+    int flags;
+
+    if (rb_scan_args(argc, argv, "21", &pattern, &path, &rflags) == 3)
+	flags = NUM2INT(rflags);
+    else
+	flags = 0;
+
+    StringValue(pattern);
+    StringValue(path);
+
+    if (fnmatch(RSTRING(pattern)->ptr, RSTRING(path)->ptr, flags) == 0)
+	return Qtrue;
+
+    return Qfalse;
+}
+
 void
 Init_Dir()
 {
@@ -880,12 +1080,13 @@ Init_Dir()
 
     rb_include_module(rb_cDir, rb_mEnumerable);
 
-    rb_define_singleton_method(rb_cDir, "new", dir_s_new, -1);
+    rb_define_alloc_func(rb_cDir, dir_s_alloc);
     rb_define_singleton_method(rb_cDir, "open", dir_s_open, 1);
     rb_define_singleton_method(rb_cDir, "foreach", dir_foreach, 1);
     rb_define_singleton_method(rb_cDir, "entries", dir_entries, 1);
 
     rb_define_method(rb_cDir,"initialize", dir_initialize, 1);
+    rb_define_method(rb_cDir,"path", dir_path, 0);
     rb_define_method(rb_cDir,"read", dir_read, 0);
     rb_define_method(rb_cDir,"each", dir_each, 0);
     rb_define_method(rb_cDir,"rewind", dir_rewind, 0);
@@ -904,6 +1105,14 @@ Init_Dir()
     rb_define_singleton_method(rb_cDir,"delete", dir_s_rmdir, 1);
     rb_define_singleton_method(rb_cDir,"unlink", dir_s_rmdir, 1);
 
-    rb_define_singleton_method(rb_cDir,"glob", dir_s_glob, 1);
-    rb_define_singleton_method(rb_cDir,"[]", dir_s_glob, 1);
+    rb_define_singleton_method(rb_cDir,"glob", dir_s_glob, -1);
+    rb_define_singleton_method(rb_cDir,"[]", dir_s_aref, 1);
+
+    rb_define_singleton_method(rb_cFile,"fnmatch", file_s_fnmatch, -1);
+    rb_define_singleton_method(rb_cFile,"fnmatch?", file_s_fnmatch, -1);
+
+    rb_file_const("FNM_NOESCAPE", INT2FIX(FNM_NOESCAPE));
+    rb_file_const("FNM_PATHNAME", INT2FIX(FNM_PATHNAME));
+    rb_file_const("FNM_DOTMATCH", INT2FIX(FNM_DOTMATCH));
+    rb_file_const("FNM_CASEFOLD", INT2FIX(FNM_CASEFOLD));
 }
