@@ -1,12 +1,15 @@
-/************************************************
+/**********************************************************************
 
   env.h -
 
-  $Author: matz $
-  $Date: 1999/08/13 05:45:04 $
+  $Author: michal $
+  $Date: 2003/01/16 07:34:01 $
   created at: Mon Jul 11 11:53:03 JST 1994
 
-************************************************/
+  Copyright (C) 1993-2003 Yukihiro Matsumoto
+
+**********************************************************************/
+
 #ifndef ENV_H
 #define ENV_H
 
@@ -15,27 +18,32 @@ extern struct FRAME {
     int argc;
     VALUE *argv;
     ID last_func;
+    ID orig_func;
     VALUE last_class;
     VALUE cbase;
     struct FRAME *prev;
     struct FRAME *tmp;
-    char *file;
-    int line;
+    struct RNode *node;
     int iter;
+    int flags;
 } *ruby_frame;
 
 void rb_gc_mark_frame _((struct FRAME *));
 
+#define FRAME_ALLOCA 0
+#define FRAME_MALLOC 1
+
 extern struct SCOPE {
     struct RBasic super;
     ID *local_tbl;
     VALUE *local_vars;
-    int flag;
+    int flags;
 } *ruby_scope;
 
 #define SCOPE_ALLOCA  0
 #define SCOPE_MALLOC  1
 #define SCOPE_NOSTACK 2
+#define SCOPE_DONT_RECYCLE 4
 
 extern int ruby_in_eval;
 
