@@ -2,17 +2,18 @@
 
   error.c -
 
-  $Author: matz $
-  $Date: 2002/01/21 07:43:39 $
+  $Author: nobu $
+  $Date: 2003/07/30 06:10:08 $
   created at: Mon Aug  9 16:11:34 JST 1993
 
-  Copyright (C) 1993-2000 Yukihiro Matsumoto
+  Copyright (C) 1993-2003 Yukihiro Matsumoto
 
 **********************************************************************/
 
 #include "ruby.h"
 #include "env.h"
 #include "version.h"
+#include "st.h"
 
 #include <stdio.h>
 #ifdef HAVE_STDARG_PROTOTYPES
@@ -23,33 +24,35 @@
 #define va_init_list(a,b) va_start(a)
 #endif
 
-#if defined __CYGWIN__
-# include <cygwin/version.h>
-# if (CYGWIN_VERSION_API_MAJOR > 0) || (CYGWIN_VERSION_API_MINOR >= 8)
-#  define sys_nerr _sys_nerr
-# endif
-#endif
-
 int ruby_nerrs;
 
+static int
+err_position(buf, len)
+    char *buf;
+    long len;
+{
+    ruby_set_current_source();
+    if (!ruby_sourcefile) {
+	return 0;
+    }
+    else if (ruby_sourceline == 0) {
+	return snprintf(buf, len, "%s: ", ruby_sourcefile);
+    }
+    else {
+	return snprintf(buf, len, "%s:%d: ", ruby_sourcefile, ruby_sourceline);
+    }
+}
+
 static void
 err_snprintf(buf, len, fmt, args)
-    char *buf, *fmt;
-    int len;
+    char *buf;
+    long len;
+    const char *fmt;
     va_list args;
 {
-    int n;
+    long n;
 
-    if (!ruby_sourcefile) {
-	vsnprintf(buf, len, fmt, args);
-	return;
-    }
-    else if (ruby_sourceline == 0) {
-	n = snprintf(buf, len, "%s: ", ruby_sourcefile);
-    }
-    else {
-	n = snprintf(buf, len, "%s:%d: ", ruby_sourcefile, ruby_sourceline);
-    }
+    n = err_position(buf, len);
     if (len > n) {
 	vsnprintf((char*)buf+n, len-n, fmt, args);
     }
@@ -110,9 +113,8 @@ warn_print(fmt, args)
     char buf[BUFSIZ];
 
     err_snprintf(buf, BUFSIZ, fmt, args);
-    fputs(buf, stderr);
-    fputs("\n", stderr);
-    fflush(stderr);
+    rb_write_error(buf);
+    rb_write_error("\n");
 }
 
 void
@@ -127,6 +129,8 @@ rb_warn(fmt, va_alist)
     char buf[BUFSIZ];
     va_list args;
 
+    if (NIL_P(ruby_verbose)) return;
+
     snprintf(buf, BUFSIZ, "warning: %s", fmt);
 
     va_init_list(args, fmt);
@@ -156,6 +160,15 @@ rb_warning(fmt, va_alist)
     va_end(args);
 }
 
+static VALUE
+rb_warn_m(self, mesg)
+    VALUE self, mesg;
+{
+    rb_io_write(rb_stderr, mesg);
+    rb_io_write(rb_stderr, rb_default_rs);
+    return mesg;
+}
+
 void
 #ifdef HAVE_STDARG_PROTOTYPES
 rb_bug(const char *fmt, ...)
@@ -174,7 +187,9 @@ rb_bug(fmt, va_alist)
     va_init_list(args, fmt);
     warn_print(buf, args);
     va_end(args);
-    fprintf(stderr, "ruby %s (%s) [%s]\n", RUBY_VERSION, RUBY_RELEASE_DATE, RUBY_PLATFORM);
+    snprintf(buf, BUFSIZ, "ruby %s (%s) [%s]\n", RUBY_VERSION, RUBY_RELEASE_DATE, RUBY_PLATFORM);
+    rb_write_error(buf);
+    rb_write_error("\n");
     abort();
 }
 
@@ -182,30 +197,30 @@ static struct types {
     int type;
     const char *name;
 } builtin_types[] = {
-    T_NIL,	"nil",
-    T_OBJECT,	"Object",
-    T_CLASS,	"Class",
-    T_ICLASS,	"iClass",	/* internal use: mixed-in module holder */
-    T_MODULE,	"Module",
-    T_FLOAT,	"Float",
-    T_STRING,	"String",
-    T_REGEXP,	"Regexp",
-    T_ARRAY,	"Array",
-    T_FIXNUM,	"Fixnum",
-    T_HASH,	"Hash",
-    T_STRUCT,	"Struct",
-    T_BIGNUM,	"Bignum",
-    T_FILE,	"File",
-    T_TRUE,	"true",
-    T_FALSE,	"false",
-    T_SYMBOL,	"Symbol",	/* :symbol */
-    T_DATA,	"Data",		/* internal use: wrapped C pointers */
-    T_MATCH,	"Match",	/* data of $~ */
-    T_VARMAP,	"Varmap",	/* internal use: dynamic variables */
-    T_SCOPE,	"Scope",	/* internal use: variable scope */
-    T_NODE,	"Node",		/* internal use: syntax tree node */
-    T_UNDEF,	"undef",	/* internal use: #undef; should not happen */
-    -1,		0,
+    {T_NIL,	"nil"},
+    {T_OBJECT,	"Object"},
+    {T_CLASS,	"Class"},
+    {T_ICLASS,	"iClass"},	/* internal use: mixed-in module holder */
+    {T_MODULE,	"Module"},
+    {T_FLOAT,	"Float"},
+    {T_STRING,	"String"},
+    {T_REGEXP,	"Regexp"},
+    {T_ARRAY,	"Array"},
+    {T_FIXNUM,	"Fixnum"},
+    {T_HASH,	"Hash"},
+    {T_STRUCT,	"Struct"},
+    {T_BIGNUM,	"Bignum"},
+    {T_FILE,	"File"},
+    {T_TRUE,	"true"},
+    {T_FALSE,	"false"},
+    {T_SYMBOL,	"Symbol"},	/* :symbol */
+    {T_DATA,	"Data"},	/* internal use: wrapped C pointers */
+    {T_MATCH,	"MatchData"},	/* data of $~ */
+    {T_VARMAP,	"Varmap"},	/* internal use: dynamic variables */
+    {T_SCOPE,	"Scope"},	/* internal use: variable scope */
+    {T_NODE,	"Node"},	/* internal use: syntax tree node */
+    {T_UNDEF,	"undef"},	/* internal use: #undef; should not happen */
+    {-1,	0}
 };
 
 void
@@ -234,7 +249,7 @@ rb_check_type(x, t)
 		    etype = RSTRING(rb_obj_as_string(x))->ptr;
 		}
 		else {
-		    etype = rb_class2name(CLASS_OF(x));
+		    etype = rb_obj_classname(x);
 		}
 		rb_raise(rb_eTypeError, "wrong argument type %s (expected %s)",
 			 etype, type->name);
@@ -259,12 +274,13 @@ VALUE rb_eTypeError;
 VALUE rb_eArgError;
 VALUE rb_eIndexError;
 VALUE rb_eRangeError;
+VALUE rb_eNameError;
+VALUE rb_eNoMethodError;
 VALUE rb_eSecurityError;
 VALUE rb_eNotImpError;
 VALUE rb_eNoMemError;
 
 VALUE rb_eScriptError;
-VALUE rb_eNameError;
 VALUE rb_eSyntaxError;
 VALUE rb_eLoadError;
 
@@ -277,10 +293,7 @@ rb_exc_new(etype, ptr, len)
     const char *ptr;
     long len;
 {
-    VALUE exc = rb_obj_alloc(etype);
-
-    rb_iv_set(exc, "mesg", rb_str_new(ptr, len));
-    return exc;
+    return rb_funcall(etype, rb_intern("new"), 1, rb_str_new(ptr, len));
 }
 
 VALUE
@@ -295,11 +308,8 @@ VALUE
 rb_exc_new3(etype, str)
     VALUE etype, str;
 {
-    char *s;
-    int len;
-
-    s = rb_str2cstr(str, &len);
-    return rb_exc_new(etype, s, len);
+    StringValue(str);
+    return rb_exc_new(etype, RSTRING(str)->ptr, RSTRING(str)->len);
 }
 
 static VALUE
@@ -311,9 +321,10 @@ exc_initialize(argc, argv, exc)
     VALUE mesg;
 
     if (rb_scan_args(argc, argv, "01", &mesg) == 1) {
-	STR2CSTR(mesg);		/* ensure mesg can be converted to String */
+	StringValue(mesg);	/* ensure mesg can be converted to String */
     }
     rb_iv_set(exc, "mesg", mesg);
+    rb_iv_set(exc, "bt", Qnil);
 
     return exc;
 }
@@ -324,16 +335,12 @@ exc_exception(argc, argv, self)
     VALUE *argv;
     VALUE self;
 {
-    VALUE etype, exc;
+    VALUE exc;
 
     if (argc == 0) return self;
     if (argc == 1 && self == argv[0]) return self;
-    etype = CLASS_OF(self);
-    while (FL_TEST(etype, FL_SINGLETON)) {
-	etype = RCLASS(etype)->super;
-    }
-    exc = rb_obj_alloc(etype);
-    rb_obj_call_init(exc, argc, argv);
+    exc = rb_obj_clone(self);
+    exc_initialize(argc, argv, exc);
 
     return exc;
 }
@@ -342,7 +349,7 @@ static VALUE
 exc_to_s(exc)
     VALUE exc;
 {
-    VALUE mesg = rb_iv_get(exc, "mesg");
+    VALUE mesg = rb_attr_get(exc, rb_intern("mesg"));
 
     if (NIL_P(mesg)) return rb_class_path(CLASS_OF(exc));
     if (OBJ_TAINTED(exc)) OBJ_TAINT(mesg);
@@ -361,12 +368,12 @@ exc_inspect(exc)
 	return rb_str_dup(rb_class_path(klass));
     }
 
-    str = rb_str_new2("#<");
+    str = rb_str_buf_new2("#<");
     klass = rb_class_path(klass);
-    rb_str_append(str, klass);
-    rb_str_cat(str, ": ", 2);
-    rb_str_append(str, exc);
-    rb_str_cat(str, ">", 1);
+    rb_str_buf_append(str, klass);
+    rb_str_buf_cat(str, ": ", 2);
+    rb_str_buf_append(str, exc);
+    rb_str_buf_cat(str, ">", 1);
 
     return str;
 }
@@ -385,7 +392,7 @@ static VALUE
 check_backtrace(bt)
     VALUE bt;
 {
-    int i;
+    long i;
     static char *err = "backtrace must be Array of String";
 
     if (!NIL_P(bt)) {
@@ -412,130 +419,211 @@ exc_set_backtrace(exc, bt)
     return rb_iv_set(exc, "bt", check_backtrace(bt));
 }
 
-#ifdef __BEOS__
-typedef struct {
-   VALUE *list;
-   int n;
-} syserr_list_entry;
-
-typedef struct {
-   int ix;
-   int n;
-} syserr_index_entry;
-
-static VALUE syserr_error;
-static VALUE syserr_list_b_general[16+1];
-static VALUE syserr_list_b_os0[2+1];
-static VALUE syserr_list_b_os1[5+1];
-static VALUE syserr_list_b_os2[2+1];
-static VALUE syserr_list_b_os3[3+1];
-static VALUE syserr_list_b_os4[1+1];
-static VALUE syserr_list_b_app[15+1];
-static VALUE syserr_list_b_interface[0+1];
-static VALUE syserr_list_b_media[8+1];
-static VALUE syserr_list_b_midi[0+1];
-static VALUE syserr_list_b_storage[15+1];
-static VALUE syserr_list_b_posix[38+1];
-static VALUE syserr_list_b_mail[8+1];
-static VALUE syserr_list_b_print[1+1];
-static VALUE syserr_list_b_device[14+1];
-
-# define SYSERR_LIST_B(n) {(n), sizeof(n)/sizeof(VALUE)}
-static const syserr_list_entry syserr_list[] = {
-   SYSERR_LIST_B(syserr_list_b_general),
-   SYSERR_LIST_B(syserr_list_b_os0),
-   SYSERR_LIST_B(syserr_list_b_os1),
-   SYSERR_LIST_B(syserr_list_b_os2),
-   SYSERR_LIST_B(syserr_list_b_os3),
-   SYSERR_LIST_B(syserr_list_b_os4),
-   SYSERR_LIST_B(syserr_list_b_app),
-   SYSERR_LIST_B(syserr_list_b_interface),
-   SYSERR_LIST_B(syserr_list_b_media),
-   SYSERR_LIST_B(syserr_list_b_midi),
-   SYSERR_LIST_B(syserr_list_b_storage),
-   SYSERR_LIST_B(syserr_list_b_posix),
-   SYSERR_LIST_B(syserr_list_b_mail),
-   SYSERR_LIST_B(syserr_list_b_print),
-   SYSERR_LIST_B(syserr_list_b_device),
-};
-# undef SYSERR_LIST_B
-
-static const syserr_index_entry syserr_index[]= {
-     {0, 1},  {1, 5},  {6, 1},  {7, 1}, {8, 1}, {9, 1}, {10, 1}, {11, 1},
-     {12, 1}, {13, 1}, {14, 1}, {0, 0},
-};
-#else
-static VALUE *syserr_list;
-#endif
-
-#if !defined(NT) && !defined(__FreeBSD__) && !defined(__NetBSD__) && !defined(__OpenBSD__) && !defined(sys_nerr)
-extern int sys_nerr;
-#endif
+static VALUE
+exit_initialize(argc, argv, exc)
+    int argc;
+    VALUE *argv;
+    VALUE exc;
+{
+    VALUE status = INT2NUM(0);
+    if (argc > 0 && FIXNUM_P(argv[0])) {
+	status = *argv++;
+	--argc;
+    }
+    exc_initialize(argc, argv, exc);
+    rb_iv_set(exc, "status", status);
+    return exc;
+}
 
 static VALUE
-set_syserr(i, name)
-    int i;
+exit_status(exc)
+    VALUE exc;
+{
+    return rb_attr_get(exc, rb_intern("status"));
+}
+
+void
+#ifdef HAVE_STDARG_PROTOTYPES
+rb_name_error(ID id, const char *fmt, ...)
+#else
+rb_name_error(id, fmt, va_alist)
+    ID id;
+    const char *fmt;
+    va_dcl
+#endif
+{
+    VALUE exc, argv[2];
+    va_list args;
+    char buf[BUFSIZ];
+
+    va_init_list(args, fmt);
+    vsnprintf(buf, BUFSIZ, fmt, args);
+    va_end(args);
+
+    argv[0] = rb_str_new2(buf);
+    argv[1] = ID2SYM(id);
+    exc = rb_class_new_instance(2, argv, rb_eNameError);
+    rb_exc_raise(exc);
+}
+
+static VALUE
+name_err_initialize(argc, argv, self)
+    int argc;
+    VALUE *argv;
+    VALUE self;
+{
+    VALUE name = (argc > 1) ? argv[--argc] : Qnil;
+    exc_initialize(argc, argv, self);
+    rb_iv_set(self, "name", name);
+    return self;
+}
+
+static VALUE
+name_err_name(self)
+    VALUE self;
+{
+    return rb_attr_get(self, rb_intern("name"));
+}
+
+static VALUE
+nometh_err_initialize(argc, argv, self)
+    int argc;
+    VALUE *argv;
+    VALUE self;
+{
+    VALUE args = (argc > 2) ? argv[--argc] : Qnil;
+    name_err_initialize(argc, argv, self);
+    rb_iv_set(self, "args", args);
+    return self;
+}
+
+static VALUE
+nometh_err_args(self)
+    VALUE self;
+{
+    return rb_attr_get(self, rb_intern("args"));
+}
+
+void
+rb_invalid_str(str, type)
+    const char *str, *type;
+{
+    VALUE s = rb_str_inspect(rb_str_new2(str));
+
+    rb_raise(rb_eArgError, "invalid value for %s: %s", type, RSTRING(s)->ptr);
+}
+
+static st_table *syserr_tbl;
+
+static VALUE
+set_syserr(n, name)
+    int n;
     const char *name;
 {
-#ifdef __BEOS__
-   VALUE *list;
-   int ix, offset;
-#endif
-    VALUE error = rb_define_class_under(rb_mErrno, name, rb_eSystemCallError);
-    rb_define_const(error, "Errno", INT2NUM(i));
-#ifdef __BEOS__
-   if (i == B_ERROR) {
-      syserr_error = error;
-      rb_global_variable(&syserr_error);
-      return error;
-   }
-   i -= B_GENERAL_ERROR_BASE;
-   ix = (i >> 12) & 0xf;
-   offset = (i >> 8) & 0xf;
-   if (offset < syserr_index[ix].n) {
-      ix = syserr_index[ix].ix;
-      if ((i & 0xff) < syserr_list[ix + offset].n) {
-	 list = syserr_list[ix + offset].list;
-	 list[i & 0xff] = error;
-	 rb_global_variable(&list[i & 0xff]);
-      }
-   }
-#else
-    if (i <= sys_nerr) {
-	syserr_list[i] = error;
+    VALUE error;
+
+    if (!st_lookup(syserr_tbl, n, &error)) {
+	error = rb_define_class_under(rb_mErrno, name, rb_eSystemCallError);
+	rb_define_const(error, "Errno", INT2NUM(n));
+	st_add_direct(syserr_tbl, n, error);
+    }
+    else {
+	rb_define_const(rb_mErrno, name, error);
     }
-#endif
     return error;
 }
 
+static VALUE
+get_syserr(n)
+    int n;
+{
+    VALUE error;
+
+    if (!st_lookup(syserr_tbl, n, &error)) {
+	char name[8];	/* some Windows' errno have 5 digits. */
+
+	snprintf(name, sizeof(name), "E%03d", n);
+	error = set_syserr(n, name);
+    }
+    return error;
+}
+
+static VALUE
+syserr_initialize(argc, argv, self)
+    int argc;
+    VALUE *argv;
+    VALUE self;
+{
+#if !defined(_WIN32) && !defined(__VMS)
+    char *strerror();
+#endif
+    char *err;
+    char *buf;
+    VALUE mesg, error;
+    VALUE klass = rb_obj_class(self);
+
+    if (klass == rb_eSystemCallError) {
+	rb_scan_args(argc, argv, "11", &mesg, &error);
+	if (argc == 1 && FIXNUM_P(mesg)) {
+	    error = mesg; mesg = Qnil;
+	}
+	if (!NIL_P(error) && st_lookup(syserr_tbl, NUM2LONG(error), &klass)) {
+	    /* change class */
+	    if (TYPE(self) != T_OBJECT) { /* insurance to avoid type crash */
+		rb_raise(rb_eTypeError, "invalid instance type");
+	    }
+	    RBASIC(self)->klass = klass;
+	}
+    }
+    else {
+	rb_scan_args(argc, argv, "01", &mesg);
+	error = rb_const_get_at(klass, rb_intern("Errno"));
+    }
+    if (!NIL_P(error)) err = strerror(NUM2LONG(error));
+    else err = "unknown error";
+    if (!NIL_P(mesg)) {
+	StringValue(mesg);
+	buf = ALLOCA_N(char, strlen(err)+RSTRING(mesg)->len+4);
+	sprintf(buf, "%s - %.*s", err, (int)RSTRING(mesg)->len, RSTRING(mesg)->ptr);
+	mesg = rb_str_new2(buf);
+    }
+    else {
+	mesg = rb_str_new2(err);
+    }
+    exc_initialize(1, &mesg, self);
+    rb_iv_set(self, "errno", error);
+    return self;
+}
+
 static VALUE
 syserr_errno(self)
     VALUE self;
 {
-    return rb_iv_get(self, "errno");
+    return rb_attr_get(self, rb_intern("errno"));
 }
 
-#ifdef __BEOS__
 static VALUE
-get_syserr(int i)
+syserr_eqq(self, exc)
+    VALUE self, exc;
 {
-   VALUE *list;
-   int ix, offset;
-   
-   if (i == B_ERROR) return syserr_error;
-   i -= B_GENERAL_ERROR_BASE;
-   ix = (i >> 12) & 0xf;
-   offset = (i >> 8) & 0xf;
-   if (offset < syserr_index[ix].n) {
-      ix = syserr_index[ix].ix;
-      if ((i & 0xff) < syserr_list[ix + offset].n) {
-	 list = syserr_list[ix + offset].list;
-	 return list[i & 0xff];
-      }
-   }
-   return 0;
+    VALUE num;
+
+    if (!rb_obj_is_kind_of(exc, rb_eSystemCallError)) return Qfalse;
+    if (self == rb_eSystemCallError) return Qtrue;
+
+    num = rb_attr_get(exc, rb_intern("errno"));
+    if (NIL_P(num)) {
+	VALUE klass = CLASS_OF(exc);
+
+	while (TYPE(klass) == T_ICLASS || FL_TEST(klass, FL_SINGLETON)) {
+	    klass = (VALUE)RCLASS(klass)->super;
+	}
+	num = rb_const_get(klass, rb_intern("Errno"));
+    }
+    if (rb_const_get(self, rb_intern("Errno")) == num)
+	return Qtrue;
+    return Qfalse;
 }
-#endif /* __BEOS__ */
 
 static void init_syserr _((void));
 
@@ -554,29 +642,37 @@ Init_Exception()
     rb_define_method(rb_eException, "set_backtrace", exc_set_backtrace, 1);
 
     rb_eSystemExit  = rb_define_class("SystemExit", rb_eException);
+    rb_define_method(rb_eSystemExit, "initialize", exit_initialize, -1);
+    rb_define_method(rb_eSystemExit, "status", exit_status, 0);
+
     rb_eFatal  	    = rb_define_class("fatal", rb_eException);
-    rb_eInterrupt   = rb_define_class("Interrupt", rb_eException);
     rb_eSignal      = rb_define_class("SignalException", rb_eException);
+    rb_eInterrupt   = rb_define_class("Interrupt", rb_eSignal);
 
     rb_eStandardError = rb_define_class("StandardError", rb_eException);
-    rb_eTypeError   = rb_define_class("TypeError", rb_eStandardError);
-    rb_eArgError    = rb_define_class("ArgumentError", rb_eStandardError);
-    rb_eIndexError  = rb_define_class("IndexError", rb_eStandardError);
-    rb_eRangeError  = rb_define_class("RangeError", rb_eStandardError);
+    rb_eTypeError     = rb_define_class("TypeError", rb_eStandardError);
+    rb_eArgError      = rb_define_class("ArgumentError", rb_eStandardError);
+    rb_eIndexError    = rb_define_class("IndexError", rb_eStandardError);
+    rb_eRangeError    = rb_define_class("RangeError", rb_eStandardError);
+    rb_eNameError     = rb_define_class("NameError", rb_eStandardError);
+    rb_define_method(rb_eNameError, "initialize", name_err_initialize, -1);
+    rb_define_method(rb_eNameError, "name", name_err_name, 0);
+    rb_eNoMethodError = rb_define_class("NoMethodError", rb_eNameError);
+    rb_define_method(rb_eNoMethodError, "initialize", nometh_err_initialize, -1);
+    rb_define_method(rb_eNoMethodError, "args", nometh_err_args, 0);
 
     rb_eScriptError = rb_define_class("ScriptError", rb_eException);
     rb_eSyntaxError = rb_define_class("SyntaxError", rb_eScriptError);
-    rb_eNameError   = rb_define_class("NameError", rb_eScriptError);
     rb_eLoadError   = rb_define_class("LoadError", rb_eScriptError);
     rb_eNotImpError = rb_define_class("NotImplementedError", rb_eScriptError);
-    /* backward compatibility -- will be removed in the future */
-    rb_define_global_const("NotImplementError", rb_eNotImpError);
 
     rb_eRuntimeError = rb_define_class("RuntimeError", rb_eStandardError);
     rb_eSecurityError = rb_define_class("SecurityError", rb_eStandardError);
     rb_eNoMemError = rb_define_class("NoMemoryError", rb_eException);
 
     init_syserr();
+
+    rb_define_global_function("warn", rb_warn_m, 1);
 }
 
 void
@@ -648,50 +744,43 @@ void
 rb_sys_fail(mesg)
     const char *mesg;
 {
-#ifndef NT
-    char *strerror();
-#endif
-    char *err;
-    char *buf;
     extern int errno;
     int n = errno;
-    VALUE ee;
-
-    err = strerror(errno);
-    if (mesg) {
-	volatile VALUE tmp = rb_str_inspect(rb_str_new2(mesg));
-
-	buf = ALLOCA_N(char, strlen(err)+RSTRING(tmp)->len+4);
-	sprintf(buf, "%s - %s", err, RSTRING(tmp)->ptr);
-    }
-    else {
-	buf = ALLOCA_N(char, strlen(err)+1);
-	strcpy(buf, err);
-    }
+    VALUE arg;
 
     errno = 0;
-#ifdef __BEOS__
-    ee = get_syserr(n);
-    if (!ee) {
-	char name[6];
-      
-	sprintf(name, "E%03d", n);
-	ee = set_syserr(n, name);
-   }
-#else
-    if (n > sys_nerr || !syserr_list[n]) {
-	char name[6];
+    if (n == 0) {
+	rb_bug("rb_sys_fail() - errno == 0");
+    }
 
-	sprintf(name, "E%03d", n);
-	ee = set_syserr(n, name);
-    }
-    else {
-	ee = syserr_list[n];
-    }
+    arg = mesg ? rb_str_new2(mesg) : Qnil;
+    rb_exc_raise(rb_class_new_instance(1, &arg, get_syserr(n)));
+}
+
+void
+#ifdef HAVE_STDARG_PROTOTYPES
+rb_sys_warning(const char *fmt, ...)
+#else
+rb_sys_warning(fmt, va_alist)
+     const char *fmt;
+     va_dcl
 #endif
-    ee = rb_exc_new2(ee, buf);
-    rb_iv_set(ee, "errno", INT2NUM(n));
-    rb_exc_raise(ee);
+{
+     char buf[BUFSIZ];
+     va_list args;
+     int errno_save;
+     
+     errno_save = errno;
+
+     if (!RTEST(ruby_verbose)) return;
+
+     snprintf(buf, BUFSIZ, "warning: %s", fmt);
+     snprintf(buf+strlen(buf), BUFSIZ-strlen(buf), ": %s", strerror(errno_save));
+     
+     va_init_list(args, fmt);
+     warn_print(buf, args);
+     va_end(args);
+     errno = errno_save;
 }
 
 void
@@ -708,29 +797,23 @@ rb_error_frozen(what)
     rb_raise(rb_eTypeError, "can't modify frozen %s", what);
 }
 
+void
+rb_check_frozen(obj)
+    VALUE obj;
+{
+    if (OBJ_FROZEN(obj)) rb_error_frozen(rb_obj_classname(obj));
+}
+
 static void
 init_syserr()
 {
-#ifdef __BEOS__
-   int i, ix, offset;
-#endif
+    syserr_tbl = st_init_numtable();
     rb_eSystemCallError = rb_define_class("SystemCallError", rb_eStandardError);
+    rb_define_method(rb_eSystemCallError, "initialize", syserr_initialize, -1);
     rb_define_method(rb_eSystemCallError, "errno", syserr_errno, 0);
+    rb_define_singleton_method(rb_eSystemCallError, "===", syserr_eqq, 1);
 
     rb_mErrno = rb_define_module("Errno");
-#ifdef __BEOS__
-   for (i = 0; syserr_index[i].n != 0; i++) {
-      ix = syserr_index[i].ix;
-      for (offset = 0; offset < syserr_index[i].n; offset++) {
-	 MEMZERO(syserr_list[ix + offset].list, VALUE, syserr_list[ix + offset].n);
-      }
-   }
-   set_syserr(B_ERROR, "ERROR");
-#else
-    syserr_list = ALLOC_N(VALUE, sys_nerr+1);
-    MEMZERO(syserr_list, VALUE, sys_nerr+1);
-#endif
-
 #ifdef EPERM
     set_syserr(EPERM, "EPERM");
 #endif
@@ -1110,7 +1193,7 @@ err_append(s)
 	    ruby_errinfo = rb_exc_new2(rb_eSyntaxError, s);
 	}
 	else {
-	    VALUE str = rb_str_to_str(ruby_errinfo);
+	    VALUE str = rb_obj_as_string(ruby_errinfo);
 
 	    rb_str_cat2(str, "\n");
 	    rb_str_cat2(str, s);
@@ -1118,8 +1201,7 @@ err_append(s)
 	}
     }
     else {
-	fputs(s, stderr);
-	fputs("\n", stderr);
-	fflush(stderr);
+	rb_write_error(s);
+	rb_write_error("\n");
     }
 }
