@@ -6,6 +6,7 @@ dir_config('termcap')
 
 make=false
 have_library("mytinfo", "tgetent") if /bow/ =~ RUBY_PLATFORM
+have_library("tinfo", "tgetent") or have_library("termcap", "tgetent")
 if have_header("ncurses.h") and have_library("ncurses", "initscr")
   make=true
 elsif have_header("ncurses/curses.h") and have_library("ncurses", "initscr")
@@ -13,14 +14,13 @@ elsif have_header("ncurses/curses.h") and have_library("ncurses", "initscr")
 elsif have_header("curses_colr/curses.h") and have_library("cur_colr", "initscr")
   make=true
 else
-  have_library("termcap", "tgetent") 
-  if have_library("curses", "initscr")
+  if have_header("curses.h") and have_library("curses", "initscr")
     make=true
   end
 end
 
 if make
-  for f in %w(isendwin ungetch beep doupdate flash deleteln wdeleteln)
+  for f in %w(isendwin ungetch beep getnstr wgetnstr doupdate flash deleteln wdeleteln keypad keyname init_color wresize resizeterm)
     have_func(f)
   end
   create_makefile("curses")
