@@ -1,17 +1,16 @@
 require 'mkmf'
 
-$CPPFLAGS += " -Dss_family=__ss_family -Dss_len=__ss_len"
-
 case RUBY_PLATFORM
+when /bccwin32/
+  test_func = "WSACleanup"
+  have_library("ws2_32", "WSACleanup")
+  have_func("closesocket")
 when /mswin32|mingw/
   test_func = "WSACleanup"
   have_library("wsock32", "WSACleanup")
   have_func("closesocket")
 when /cygwin/
-#  $LDFLAGS << " -L/usr/lib" if File.directory?("/usr/lib")
-#  $CFLAGS << " -I/usr/include"
   test_func = "socket"
-#  have_library("bind", "gethostbyaddr")
 when /beos/
   test_func = "socket"
   have_library("net", "socket")
@@ -26,7 +25,8 @@ else
 end
 
 $ipv6 = false
-if enable_config("ipv6", false)
+default_ipv6 = /cygwin/ !~ RUBY_PLATFORM
+if enable_config("ipv6", default_ipv6)
   if try_link(<<EOF)
 #include <sys/types.h>
 #include <sys/socket.h>
@@ -45,19 +45,13 @@ $ipv6lib = nil
 $ipv6libdir = nil
 $ipv6trylibc = nil
 if $ipv6
-  if egrep_cpp("yes", <<EOF)
+  if macro_defined?("IPV6_INRIA_VERSION", <<EOF)
 #include <netinet/in.h>
-#ifdef IPV6_INRIA_VERSION
-yes
-#endif
 EOF
     $ipv6type = "inria"
     $CFLAGS="-DINET6 "+$CFLAGS
-  elsif egrep_cpp("yes", <<EOF)
+  elsif macro_defined?("__KAME__", <<EOF)
 #include <netinet/in.h>
-#ifdef __KAME__
-yes
-#endif
 EOF
     $ipv6type = "kame"
     $ipv6lib="inet6"
@@ -69,31 +63,22 @@ EOF
     $ipv6lib="inet6"
     $ipv6libdir="/usr/inet6/lib"
     $CFLAGS="-DINET6 -I/usr/inet6/include "+$CFLAGS
-  elsif egrep_cpp("yes", <<EOF)
+  elsif macro_defined?("_TOSHIBA_INET6", <<EOF)
 #include <sys/param.h>
-#ifdef _TOSHIBA_INET6
-yes
-#endif
 EOF
     $ipv6type = "toshiba"
     $ipv6lib="inet6"
     $ipv6libdir="/usr/local/v6/lib"
     $CFLAGS="-DINET6 "+$CFLAGS
-  elsif egrep_cpp("yes", <<EOF)
+  elsif macro_defined?("__V6D__", <<EOF)
 #include </usr/local/v6/include/sys/v6config.h>
-#ifdef __V6D__
-yes
-#endif
 EOF
     $ipv6type = "v6d"
     $ipv6lib="v6"
     $ipv6libdir="/usr/local/v6/lib"
     $CFLAGS="-DINET6 -I/usr/local/v6/include "+$CFLAGS
-  elsif egrep_cpp("yes", <<EOF)
+  elsif macro_defined?("_ZETA_MINAMI_INET6", <<EOF)
 #include <sys/param.h>
-#ifdef _ZETA_MINAMI_INET6
-yes
-#endif
 EOF
     $ipv6type = "zeta"
     $ipv6lib="inet6"
@@ -163,6 +148,29 @@ main()
 }
 EOF
     $CFLAGS="-DHAVE_SOCKADDR_STORAGE "+$CFLAGS
+else      #   doug's fix, NOW add -Dss_family... only if required!
+$CPPFLAGS += " -Dss_family=__ss_family -Dss_len=__ss_len"
+  if try_link(<<EOF)
+#ifdef _WIN32
+# include <windows.h>
+# include <winsock.h>
+#else
+# include <sys/types.h>
+# include <netdb.h>
+# include <string.h>
+# include <sys/socket.h>
+#endif
+int
+main()
+{
+   struct sockaddr_storage ss;
+
+   ss.ss_family;
+   return 0;
+}
+EOF
+    $CFLAGS="-DHAVE_SOCKADDR_STORAGE "+$CFLAGS
+end
 end
 
   if try_link(<<EOF)
@@ -183,11 +191,16 @@ EOF
     $CFLAGS="-DHAVE_SA_LEN "+$CFLAGS
 end
 
-have_header("netinet/tcp.h") if not /cygwin/ === RUBY_PLATFORM # for cygwin 1.1.5
+have_header("netinet/tcp.h") if not /cygwin/ =~ RUBY_PLATFORM # for cygwin 1.1.5
 have_header("netinet/udp.h")
 
+if have_func("sendmsg") or have_func("recvmsg")
+  have_struct_member('struct msghdr', 'msg_control', header=['sys/types.h', 'sys/socket.h'])
+  have_struct_member('struct msghdr', 'msg_accrights', header=['sys/types.h', 'sys/socket.h'])
+end
+
 $getaddr_info_ok = false
-if not enable_config("wide-getaddrinfo", false) and try_run(<<EOF)
+if !enable_config("wide-getaddrinfo", false) and try_run(<<EOF)
 #include <sys/types.h>
 #include <netdb.h>
 #include <string.h>
@@ -207,6 +220,7 @@ main()
   for (passive = 0; passive <= 1; passive++) {
     memset(&hints, 0, sizeof(hints));
     hints.ai_family = AF_UNSPEC;
+    hints.ai_protocol = IPPROTO_TCP;
     hints.ai_flags = passive ? AI_PASSIVE : 0;
     hints.ai_socktype = SOCK_STREAM;
     if ((gaierr = getaddrinfo(NULL, "54321", &hints, &aitop)) != 0) {
@@ -283,6 +297,7 @@ if $ipv6 and not $getaddr_info_ok
 Fatal: --enable-ipv6 is specified, and your OS seems to support IPv6 feature.
 But your getaddrinfo() and getnameinfo() are appeared to be broken.  Sorry,
 you cannot compile IPv6 socket classes with broken these functions.
+You can try --enable-wide-getaddrinfo.
 EOS
   exit
 end
@@ -339,13 +354,14 @@ EOF
 end
 
 have_header("sys/un.h")
+have_header("sys/uio.h")
 
 if have_func(test_func)
   have_func("hsterror")
   unless have_func("gethostname")
     have_func("uname")
   end
-  if ENV["SOCKS_SERVER"] or enable_config("socks", false)
+  if enable_config("socks", ENV["SOCKS_SERVER"])
     if have_library("socks5", "SOCKSinit")
       $CFLAGS+=" -DSOCKS5 -DSOCKS"
     elsif have_library("socks", "Rconnect")
