@@ -36,10 +36,16 @@
 
 #include "config.h"
 #include <sys/types.h>
-#ifndef NT
-#include <sys/socket.h>
+#ifndef _WIN32
+#if defined(__BEOS__)
+# include <net/socket.h>
+#else
+# include <sys/socket.h>
+#endif
 #include <netinet/in.h>
+#if defined(HAVE_ARPA_INET_H)
 #include <arpa/inet.h>
+#endif
 #if defined(HAVE_ARPA_NAMESER_H)
 #include <arpa/nameser.h>
 #endif
@@ -48,7 +54,7 @@
 #include <resolv.h>
 #endif
 #endif
-#ifdef NT
+#ifdef _WIN32
 #include <winsock2.h>
 #include <stdio.h>
 #define snprintf _snprintf
@@ -174,26 +180,30 @@ getnameinfo(sa, salen, host, hostlen, serv, servlen, flags)
 		/* what we should do? */
 	} else if (flags & NI_NUMERICSERV) {
 		snprintf(numserv, sizeof(numserv), "%d", ntohs(port));
-		if (strlen(numserv) > servlen)
+		if (strlen(numserv) + 1 > servlen)
 			return ENI_MEMORY;
 		strcpy(serv, numserv);
 	} else {
+#if defined(HAVE_GETSERVBYPORT)
 		sp = getservbyport(port, (flags & NI_DGRAM) ? "udp" : "tcp");
 		if (sp) {
-			if (strlen(sp->s_name) > servlen)
+			if (strlen(sp->s_name) + 1 > servlen)
 				return ENI_MEMORY;
 			strcpy(serv, sp->s_name);
 		} else
 			return ENI_NOSERVNAME;
+#else
+		return ENI_NOSERVNAME;
+#endif
 	}
 
 	switch (sa->sa_family) {
 	case AF_INET:
-		v4a = ((struct sockaddr_in *)sa)->sin_addr.s_addr;
+		v4a = ntohl(((struct sockaddr_in *)sa)->sin_addr.s_addr);
 		if (IN_MULTICAST(v4a) || IN_EXPERIMENTAL(v4a))
 			flags |= NI_NUMERICHOST;
 		v4a >>= IN_CLASSA_NSHIFT;
-		if (v4a == 0 || v4a == IN_LOOPBACKNET)
+		if (v4a == 0)
 			flags |= NI_NUMERICHOST;			
 		break;
 #ifdef INET6
@@ -226,7 +236,7 @@ getnameinfo(sa, salen, host, hostlen, serv, servlen, flags)
 				p = strchr(hp->h_name, '.');
 				if (p) *p = '\0';
 			}
-			if (strlen(hp->h_name) > hostlen) {
+			if (strlen(hp->h_name) + 1 > hostlen) {
 #ifdef INET6
 				freehostent(hp);
 #endif
