@@ -2,7 +2,7 @@
 
 require 'mkmf'
 
-if RUBY_PLATFORM !~ /mswin32|mingw|cygwin/
+if RUBY_PLATFORM !~ /mswin32|mingw|cygwin|bccwin32/
   have_library("nsl", "t_open")
   have_library("socket", "socket")
   have_library("dl", "dlopen")
@@ -48,8 +48,8 @@ def find_tk(tklib, stubs)
 end
 
 if have_header("tcl.h") && have_header("tk.h") &&
-    (/mswin32|mingw|cygwin/ =~ RUBY_PLATFORM || find_library("X11", "XOpenDisplay",
-	"/usr/X11/lib", "/usr/X11R6/lib", "/usr/openwin/lib")) &&
+    (/mswin32|mingw|cygwin|bccwin32/ =~ RUBY_PLATFORM || find_library("X11", "XOpenDisplay",
+	"/usr/X11/lib", "/usr/lib/X11", "/usr/X11R6/lib", "/usr/openwin/lib")) &&
     find_tcl(tcllib, stubs) &&
     find_tk(tklib, stubs)
   $CPPFLAGS += ' -DUSE_TCL_STUBS -DUSE_TK_STUBS' if stubs
