@@ -1,8 +1,8 @@
-
+#
 #		tkcanvas.rb - Tk canvas classes
-#			$Date: 2002/02/28 06:52:49 $
+#			$Date: 2003/08/03 08:53:06 $
 #			by Yukihiro Matsumoto <matz@caelum.co.jp>
-#			$Date: 2002/02/28 06:52:49 $
+#			$Date: 2003/08/03 08:53:06 $
 #			by Hidetoshi Nagai <nagai@ai.kyutech.ac.jp>
 
 require "tk"
@@ -11,7 +11,7 @@ require 'tkfont'
 module TkTreatCItemFont
   include TkTreatItemFont
 
-  ItemCMD = ['itemconfigure', TkComm::None]
+  ItemCMD = ['itemconfigure'.freeze, TkComm::None].freeze
   def __conf_cmd(idx)
     ItemCMD[idx]
   end
@@ -29,10 +29,12 @@ class TkCanvas<TkWindow
   include TkTreatCItemFont
   include Scrollable
 
+  TkCommandNames = ['canvas'.freeze].freeze
   WidgetClassName = 'Canvas'.freeze
   WidgetClassNames[WidgetClassName] = self
-  def self.to_eval
-    WidgetClassName
+
+  def __destroy_hook__
+    TkcItem::CItemID_TBL.delete(@path)
   end
 
   def create_self(keys)
@@ -54,6 +56,7 @@ class TkCanvas<TkWindow
 
   def addtag(tag, mode, *args)
     tk_send 'addtag', tagid(tag), mode, *args
+    self
   end
   def addtag_above(tagOrId, target)
     addtag(tagOrId, 'above', tagid(target))
@@ -83,10 +86,17 @@ class TkCanvas<TkWindow
 
   def itembind(tag, context, cmd=Proc.new, args=nil)
     _bind([path, "bind", tagid(tag)], context, cmd, args)
+    self
   end
 
   def itembind_append(tag, context, cmd=Proc.new, args=nil)
     _bind_append([path, "bind", tagid(tag)], context, cmd, args)
+    self
+  end
+
+  def itembind_remove(tag, context)
+    _bind_remove([path, "bind", tagid(tag)], context)
+    self
   end
 
   def itembindinfo(tag, context=nil)
@@ -94,35 +104,43 @@ class TkCanvas<TkWindow
   end
 
   def canvasx(x, *args)
-    tk_tcl2ruby(tk_send 'canvasx', x, *args)
+    tk_tcl2ruby(tk_send('canvasx', x, *args))
   end
   def canvasy(y, *args)
-    tk_tcl2ruby(tk_send 'canvasy', y, *args)
+    tk_tcl2ruby(tk_send('canvasy', y, *args))
   end
 
   def coords(tag, *args)
     if args == []
       tk_split_list(tk_send('coords', tagid(tag)))
     else
-      tk_send('coords', tagid(tag), *args)
+      tk_send('coords', tagid(tag), *(args.flatten))
     end
   end
 
   def dchars(tag, first, last=None)
     tk_send 'dchars', tagid(tag), first, last
+    self
   end
 
   def delete(*args)
+    if TkcItem::CItemID_TBL[self.path]
+      find('withtag', *args).each{|item| 
+	TkcItem::CItemID_TBL[self.path].delete(item.id)
+      }
+    end
     tk_send 'delete', *args.collect{|t| tagid(t)}
+    self
   end
   alias remove delete
 
   def dtag(tag, tag_to_del=None)
     tk_send 'dtag', tagid(tag), tag_to_del
+    self
   end
 
   def find(mode, *args)
-    list(tk_send 'find', mode, *args).collect!{|id| 
+    list(tk_send('find', mode, *args)).collect!{|id| 
       TkcItem.id2obj(self, id)
     }
   end
@@ -151,6 +169,7 @@ class TkCanvas<TkWindow
   def itemfocus(tagOrId=nil)
     if tagOrId
       tk_send 'focus', tagid(tagOrId)
+      self
     else
       ret = tk_send('focus')
       if ret == ""
@@ -169,18 +188,20 @@ class TkCanvas<TkWindow
 
   def icursor(tagOrId, index)
     tk_send 'icursor', tagid(tagOrId), index
+    self
   end
 
   def index(tagOrId, index)
-    tk_send 'index', tagid(tagOrId), index
+    number(tk_send('index', tagid(tagOrId), index))
   end
 
   def insert(tagOrId, index, string)
     tk_send 'insert', tagid(tagOrId), index, string
+    self
   end
 
   def itemcget(tagOrId, option)
-    case option
+    case option.to_s
     when 'dash', 'activedash', 'disableddash'
       conf = tk_send('itemcget', tagid(tagOrId), "-#{option}")
       if conf =~ /^[0-9]/
@@ -191,12 +212,13 @@ class TkCanvas<TkWindow
     when 'text', 'label', 'show', 'data', 'file', 'maskdata', 'maskfile'
       tk_send 'itemcget', tagid(tagOrId), "-#{option}"
     else
-      tk_tcl2ruby tk_send 'itemcget', tagid(tagOrId), "-#{option}"
+      tk_tcl2ruby tk_send('itemcget', tagid(tagOrId), "-#{option}")
     end
   end
 
   def itemconfigure(tagOrId, key, value=None)
     if key.kind_of? Hash
+      key = _symbolkey2str(key)
       if ( key['font'] || key['kanjifont'] \
 	  || key['latinfont'] || key['asciifont'] )
 	tagfont_configure(tagOrId, key.dup)
@@ -205,13 +227,16 @@ class TkCanvas<TkWindow
       end
 
     else
-      if ( key == 'font' || key == 'kanjifont' \
-	  || key == 'latinfont' || key == 'asciifont' )
+      if ( key == 'font' || key == :font || 
+           key == 'kanjifont' || key == :kanjifont || 
+	   key == 'latinfont' || key == :latinfont || 
+           key == 'asciifont' || key == :asciifont )
 	tagfont_configure(tagid(tagOrId), {key=>value})
       else
 	tk_send 'itemconfigure', tagid(tagOrId), "-#{key}", value
       end
     end
+    self
   end
 #  def itemconfigure(tagOrId, key, value=None)
 #    if key.kind_of? Hash
@@ -226,10 +251,10 @@ class TkCanvas<TkWindow
 
   def itemconfiginfo(tagOrId, key=nil)
     if key
-      case key
+      case key.to_s
       when 'dash', 'activedash', 'disableddash'
-	conf = tk_split_simplelist(tk_send 'itemconfigure', 
-				   tagid(tagOrId), "-#{key}")
+	conf = tk_split_simplelist(tk_send('itemconfigure', 
+					   tagid(tagOrId), "-#{key}"))
 	if conf[3] && conf[3] =~ /^[0-9]/
 	  conf[3] = list(conf[3])
 	end
@@ -237,17 +262,17 @@ class TkCanvas<TkWindow
 	  conf[4] = list(conf[4])
 	end
       when 'text', 'label', 'show', 'data', 'file', 'maskdata', 'maskfile'
-	conf = tk_split_simplelist(tk_send 'itemconfigure', 
-				   tagid(tagOrId), "-#{key}")
+	conf = tk_split_simplelist(tk_send('itemconfigure', 
+					   tagid(tagOrId), "-#{key}"))
       else
-	conf = tk_split_list(tk_send 'itemconfigure', 
-			     tagid(tagOrId), "-#{key}")
+	conf = tk_split_list(tk_send('itemconfigure', 
+				     tagid(tagOrId), "-#{key}"))
       end
       conf[0] = conf[0][1..-1]
       conf
     else
-      tk_split_simplelist(tk_send 'itemconfigure', 
-			  tagid(tagOrId)).collect{|conflist|
+      tk_split_simplelist(tk_send('itemconfigure', 
+				  tagid(tagOrId))).collect{|conflist|
 	conf = tk_split_simplelist(conflist)
 	conf[0] = conf[0][1..-1]
 	case conf[0]
@@ -282,10 +307,12 @@ class TkCanvas<TkWindow
 
   def lower(tag, below=None)
     tk_send 'lower', tagid(tag), tagid(below)
+    self
   end
 
   def move(tag, x, y)
     tk_send 'move', tagid(tag), x, y
+    self
   end
 
   def postscript(keys)
@@ -294,21 +321,26 @@ class TkCanvas<TkWindow
 
   def raise(tag, above=None)
     tk_send 'raise', tagid(tag), tagid(above)
+    self
   end
 
   def scale(tag, x, y, xs, ys)
     tk_send 'scale', tagid(tag), x, y, xs, ys
+    self
   end
 
   def scan_mark(x, y)
     tk_send 'scan', 'mark', x, y
+    self
   end
   def scan_dragto(x, y)
     tk_send 'scan', 'dragto', x, y
+    self
   end
 
   def select(mode, *args)
-    tk_send 'select', mode, *args
+    r = tk_send('select', mode, *args)
+    (mode == 'item')? TkcItem.id2obj(self, r): self
   end
   def select_adjust(tagOrId, index)
     select('adjust', tagid(tagOrId), index)
@@ -327,7 +359,7 @@ class TkCanvas<TkWindow
   end
 
   def itemtype(tag)
-    TkcItem.type2class(tk_send 'type', tagid(tag))
+    TkcItem.type2class(tk_send('type', tagid(tag)))
   end
 end
 
@@ -337,6 +369,7 @@ module TkcTagAccess
 
   def addtag(tag)
     @c.addtag(tag, 'with', @id)
+    self
   end
 
   def bbox
@@ -345,6 +378,17 @@ module TkcTagAccess
 
   def bind(seq, cmd=Proc.new, args=nil)
     @c.itembind @id, seq, cmd, args
+    self
+  end
+
+  def bind_append(seq, cmd=Proc.new, args=nil)
+    @c.itembind_append @id, seq, cmd, args
+    self
+  end
+
+  def bind_remove(seq)
+    @c.itembind_remove @id, seq
+    self
   end
 
   def bindinfo(seq=nil)
@@ -357,6 +401,7 @@ module TkcTagAccess
 
   def configure(key, value=None)
     @c.itemconfigure @id, key, value
+    self
   end
 #  def configure(keys)
 #    @c.itemconfigure @id, keys
@@ -372,10 +417,12 @@ module TkcTagAccess
 
   def dchars(first, last=None)
     @c.dchars @id, first, last
+    self
   end
 
   def dtag(tag_to_del=None)
     @c.dtag @id, tag_to_del
+    self
   end
 
   def find
@@ -393,6 +440,7 @@ module TkcTagAccess
 
   def icursor(index)
     @c.icursor @id, index
+    self
   end
 
   def index(index)
@@ -401,39 +449,47 @@ module TkcTagAccess
 
   def insert(beforethis, string)
     @c.insert @id, beforethis, string
+    self
   end
 
   def lower(belowthis=None)
     @c.lower @id, belowthis
+    self
   end
 
   def move(xamount, yamount)
     @c.move @id, xamount, yamount
+    self
   end
 
   def raise(abovethis=None)
     @c.raise @id, abovethis
+    self
   end
 
   def scale(xorigin, yorigin, xscale, yscale)
     @c.scale @id, xorigin, yorigin, xscale, yscale
+    self
   end
 
   def select_adjust(index)
     @c.select('adjust', @id, index)
+    self
   end
   def select_from(index)
     @c.select('from', @id, index)
+    self
   end
   def select_to(index)
     @c.select('to', @id, index)
+    self
   end
 
   def itemtype
     @c.itemtype @id
   end
 
-  # Followings operators supports logical expressions of canvas tags
+  # Following operators support logical expressions of canvas tags
   # (for Tk8.3+).
   # If tag1.path is 't1' and tag2.path is 't2', then
   #      ltag = tag1 & tag2; ltag.path => "(t1)&&(t2)"
@@ -472,7 +528,10 @@ end
 class TkcTag<TkObject
   include TkcTagAccess
 
-  CTagID_TBL = {}
+  CTagID_TBL = TkCore::INTERP.create_table
+  Tk_CanvasTag_ID = ['ctag'.freeze, '00000'].freeze
+
+  TkCore::INTERP.init_ip_env{ CTagID_TBL.clear }
 
   def TkcTag.id2obj(canvas, id)
     cpath = canvas.path
@@ -480,64 +539,71 @@ class TkcTag<TkObject
     CTagID_TBL[cpath][id]? CTagID_TBL[cpath][id]: id
   end
 
-  Tk_CanvasTag_ID = ['ctag0000']
   def initialize(parent, mode=nil, *args)
     if not parent.kind_of?(TkCanvas)
       fail format("%s need to be TkCanvas", parent.inspect)
     end
     @c = parent
     @cpath = parent.path
-    @path = @id = Tk_CanvasTag_ID[0]
+    @path = @id = Tk_CanvasTag_ID.join
     CTagID_TBL[@cpath] = {} unless CTagID_TBL[@cpath]
     CTagID_TBL[@cpath][@id] = self
-    Tk_CanvasTag_ID[0] = Tk_CanvasTag_ID[0].succ
+    Tk_CanvasTag_ID[1].succ!
     if mode
       tk_call @c.path, "addtag", @id, mode, *args
     end
   end
   def id
-    return @id
+    @id
   end
 
   def delete
     @c.delete @id
-    CTagID_TBL[@path][@id] = nil if CTagID_TBL[@path]
+    CTagID_TBL[@cpath].delete(@id) if CTagID_TBL[@cpath]
+    self
   end
   alias remove  delete
   alias destroy delete
 
   def set_to_above(target)
     @c.addtag_above(@id, target)
+    self
   end
   alias above set_to_above
 
   def set_to_all
     @c.addtag_all(@id)
+    self
   end
   alias all set_to_all
 
   def set_to_below(target)
     @c.addtag_below(@id, target)
+    self
   end
   alias below set_to_below
 
   def set_to_closest(x, y, halo=None, start=None)
     @c.addtag_closest(@id, x, y, halo, start)
+    self
   end
   alias closest set_to_closest
 
   def set_to_enclosed(x1, y1, x2, y2)
     @c.addtag_enclosed(@id, x1, y1, x2, y2)
+    self
   end
   alias enclosed set_to_enclosed
 
   def set_to_overlapping(x1, y1, x2, y2)
     @c.addtag_overlapping(@id, x1, y1, x2, y2)
+    self
   end
   alias overlapping set_to_overlapping
 
   def set_to_withtag(target)
     @c.addtag_withtag(@id, target)
+    self
   end
   alias withtag set_to_withtag
 end
@@ -565,6 +631,7 @@ class TkcTagString<TkcTag
     end
   end
 end
+TkcNamedTag = TkcTagString
 
 class TkcTagAll<TkcTag
   def initialize(parent)
@@ -593,17 +660,17 @@ class TkcTagCurrent<TkcTag
 end
 
 class TkcGroup<TkcTag
-  Tk_cGroup_ID = ['tkcg00000']
+  Tk_cGroup_ID = ['tkcg'.freeze, '00000'].freeze
   def create_self(parent, *args)
     if not parent.kind_of?(TkCanvas)
       fail format("%s need to be TkCanvas", parent.inspect)
     end
     @c = parent
     @cpath = parent.path
-    @path = @id = Tk_cGroup_ID[0]
+    @path = @id = Tk_cGroup_ID.join
     CTagID_TBL[@cpath] = {} unless CTagID_TBL[@cpath]
     CTagID_TBL[@cpath][@id] = self
-    Tk_cGroup_ID[0] = Tk_cGroup_ID[0].succ
+    Tk_cGroup_ID[1].succ!
     add(*args) if args != []
   end
   
@@ -611,12 +678,14 @@ class TkcGroup<TkcTag
     for i in tags
       i.addtag @id
     end
+    self
   end
 
   def exclude(*tags)
     for i in tags
       i.delete @id
     end
+    self
   end
 end
 
@@ -624,7 +693,9 @@ class TkcItem<TkObject
   include TkcTagAccess
 
   CItemTypeToClass = {}
-  CItemID_TBL = {}
+  CItemID_TBL = TkCore::INTERP.create_table
+
+  TkCore::INTERP.init_ip_env{ CItemID_TBL.clear }
 
   def TkcItem.type2class(type)
     CItemTypeToClass[type]
@@ -643,9 +714,16 @@ class TkcItem<TkObject
     @parent = @c = parent
     @path = parent.path
     fontkeys = {}
+    if args.size == 1 && args[0].kind_of?(Hash)
+      args[0] = _symbolkey2str(args[0])
+      coords = args[0].delete('coords')
+      if not coords.kind_of?(Array)
+        fail "coords parameter must be given by an Array"
+      end
+      args[0,0] = coords.flatten
+    end
     if args[-1].kind_of? Hash
-      args = args.dup
-      keys = args.pop
+      keys = _symbolkey2str(args.pop)
       ['font', 'kanjifont', 'latinfont', 'asciifont'].each{|key|
 	fontkeys[key] = keys.delete(key) if keys.key?(key)
       }
@@ -672,12 +750,13 @@ class TkcItem<TkObject
   def create_self(*args); end
   private :create_self
   def id
-    return @id
+    @id
   end
 
   def delete
     @c.delete @id
-    CItemID_TBL[@path][@id] = nil if CItemID_TBL[@path]
+    CItemID_TBL[@path].delete(@id) if CItemID_TBL[@path]
+    self
   end
   alias remove  delete
   alias destroy delete
@@ -741,19 +820,24 @@ end
 class TkImage<TkObject
   include Tk
 
-  Tk_IMGTBL = {}
+  TkCommandNames = ['image'.freeze].freeze
+
+  Tk_IMGTBL = TkCore::INTERP.create_table
+  Tk_Image_ID = ['i'.freeze, '00000'].freeze
+
+  TkCore::INTERP.init_ip_env{ Tk_IMGTBL.clear }
 
-  Tk_Image_ID = ['i00000']
   def initialize(keys=nil)
-    @path = Tk_Image_ID[0]
-    Tk_Image_ID[0] = Tk_Image_ID[0].succ
+    @path = Tk_Image_ID.join
+    Tk_Image_ID[1].succ!
     tk_call 'image', 'create', @type, @path, *hash_kv(keys)
     Tk_IMGTBL[@path] = self
   end
 
   def delete
-    Tk_IMGTBL[@id] = nil if @id
+    Tk_IMGTBL.delete(@id) if @id
     tk_call('image', 'delete', @path)
+    self
   end
   def height
     number(tk_call('image', 'height', @path))
@@ -794,14 +878,15 @@ class TkPhotoImage<TkImage
 
   def blank
     tk_send 'blank'
+    self
   end
 
   def cget(option)
-    case option
+    case option.to_s
     when 'data', 'file'
       tk_send 'cget', option
     else
-      tk_tcl2ruby tk_send 'cget', option
+      tk_tcl2ruby tk_send('cget', option)
     end
   end
 
@@ -815,14 +900,16 @@ class TkPhotoImage<TkImage
     }.flatten
 
     tk_send 'copy', source, *args
+
+    self
   end
 
   def data(keys=nil)
-    tk_send 'data', *hash_kv(keys)
+    tk_send('data', *hash_kv(keys))
   end
 
   def get(x, y)
-    tk_send 'get', x, y
+    tk_send('get', x, y).split.collect{|n| n.to_i}
   end
 
   def put(data, *to)
@@ -831,6 +918,7 @@ class TkPhotoImage<TkImage
     else
       tk_send 'put', data, '-to', *to
     end
+    self
   end
 
   def read(file, *opts)
@@ -843,10 +931,21 @@ class TkPhotoImage<TkImage
     }.flatten
   
     tk_send 'read', file, *args
+
+    self
   end
 
   def redither
     tk_send 'redither'
+    self
+  end
+
+  def get_transparency(x, y)
+    bool(tk_send('transparency', 'get', x, y))
+  end
+  def set_transparency(x, y, st)
+    tk_send('transparency', 'set', x, y, st)
+    self
   end
 
   def write(file, *opts)
@@ -859,5 +958,7 @@ class TkPhotoImage<TkImage
     }.flatten
   
     tk_send 'write', file, *args
+
+    self
   end
 end
