@@ -3,10 +3,10 @@
   gc.c -
 
   $Author: matz $
-  $Date: 2002/12/02 18:19:03 $
+  $Date: 2003/08/02 06:45:57 $
   created at: Tue Oct  5 09:44:46 JST 1993
 
-  Copyright (C) 1993-2000 Yukihiro Matsumoto
+  Copyright (C) 1993-2003 Yukihiro Matsumoto
   Copyright (C) 2000  Network Applied Communication Laboratory, Inc.
   Copyright (C) 2000  Information-technology Promotion Agency, Japan
 
@@ -20,6 +20,21 @@
 #include "re.h"
 #include <stdio.h>
 #include <setjmp.h>
+#include <sys/types.h>
+
+#ifdef HAVE_SYS_TIME_H
+#include <sys/time.h>
+#endif
+
+#ifdef HAVE_SYS_RESOURCE_H
+#include <sys/resource.h>
+#endif
+
+#ifdef __ia64__
+#include <ucontext.h>
+#pragma weak __libc_ia64_register_backing_store_base
+extern unsigned long __libc_ia64_register_backing_store_base;
+#endif
 
 void re_free_registers _((struct re_registers*));
 void rb_io_fptr_finalize _((struct OpenFile*));
@@ -49,8 +64,6 @@ void *alloca ();
 # endif /* HAVE_ALLOCA_H */
 #endif /* __GNUC__ */
 
-static void run_final();
-
 #ifndef GC_MALLOC_LIMIT
 #if defined(MSDOS) || defined(__human68k__)
 #define GC_MALLOC_LIMIT 200000
@@ -59,23 +72,22 @@ static void run_final();
 #endif
 #endif
 
-static unsigned long malloc_memories = 0;
+static unsigned long malloc_increase = 0;
+static unsigned long malloc_limit = GC_MALLOC_LIMIT;
+static void run_final();
+static VALUE nomem_error;
 
-static void
-mem_error(mesg)
-    char *mesg;
+void
+rb_memerror()
 {
     static int recurse = 0;
 
-    if (rb_safe_level() >= 4) {
-	rb_raise(rb_eNoMemError, mesg);
+    if (recurse > 0 && rb_safe_level() < 4) {
+	fprintf(stderr, "[FATAL] failed to allocate memory\n");
+	exit(1);
     }
-    if (recurse == 0) {
-	recurse++;
-	rb_fatal(mesg);
-    }
-    fprintf(stderr, "[FATAL] failed to allocate memory\n");
-    exit(1);
+    recurse++;
+    rb_exc_raise(nomem_error);
 }
 
 void *
@@ -88,9 +100,9 @@ ruby_xmalloc(size)
 	rb_raise(rb_eNoMemError, "negative allocation size (or too big)");
     }
     if (size == 0) size = 1;
-    malloc_memories += size;
+    malloc_increase += size;
 
-    if (malloc_memories > GC_MALLOC_LIMIT) {
+    if (malloc_increase > malloc_limit) {
 	rb_gc();
     }
     RUBY_CRITICAL(mem = malloc(size));
@@ -98,10 +110,7 @@ ruby_xmalloc(size)
 	rb_gc();
 	RUBY_CRITICAL(mem = malloc(size));
 	if (!mem) {
-	    if (size >= 10 * 1024 * 1024) {
-		rb_raise(rb_eNoMemError, "tried to allocate too big memory");
-	    }
-	    mem_error("failed to allocate memory");
+	    rb_memerror();
 	}
     }
 
@@ -132,17 +141,14 @@ ruby_xrealloc(ptr, size)
     }
     if (!ptr) return xmalloc(size);
     if (size == 0) size = 1;
-    malloc_memories += size;
+    malloc_increase += size;
     RUBY_CRITICAL(mem = realloc(ptr, size));
     if (!mem) {
 	rb_gc();
 	RUBY_CRITICAL(mem = realloc(ptr, size));
 	if (!mem) {
-	    if (size >= 50 * 1024 * 1024) {
-		rb_raise(rb_eNoMemError, "tried to re-allocate too big memory");
-	    }
-	    mem_error("failed to allocate memory(realloc)");
-	}
+	    rb_memerror();
+        }
     }
 
     return mem;
@@ -162,8 +168,8 @@ static int during_gc;
 static int need_call_final = 0;
 static st_table *finalizer_table = 0;
 
-static VALUE
-gc_enable()
+VALUE
+rb_gc_enable()
 {
     int old = dont_gc;
 
@@ -171,8 +177,8 @@ gc_enable()
     return old;
 }
 
-static VALUE
-gc_disable()
+VALUE
+rb_gc_disable()
 {
     int old = dont_gc;
 
@@ -185,7 +191,7 @@ VALUE rb_mGC;
 static struct gc_list {
     VALUE *varptr;
     struct gc_list *next;
-} *Global_List = 0;
+} *global_List = 0;
 
 void
 rb_gc_register_address(addr)
@@ -194,19 +200,19 @@ rb_gc_register_address(addr)
     struct gc_list *tmp;
 
     tmp = ALLOC(struct gc_list);
-    tmp->next = Global_List;
+    tmp->next = global_List;
     tmp->varptr = addr;
-    Global_List = tmp;
+    global_List = tmp;
 }
 
 void
 rb_gc_unregister_address(addr)
     VALUE *addr;
 {
-    struct gc_list *tmp = Global_List;
+    struct gc_list *tmp = global_List;
 
     if (tmp->varptr == addr) {
-	Global_List = tmp->next;
+	global_List = tmp->next;
 	RUBY_CRITICAL(free(tmp));
 	return;
     }
@@ -232,7 +238,7 @@ rb_global_variable(var)
 typedef struct RVALUE {
     union {
 	struct {
-	    unsigned long flag;	/* always 0 for freed obj */
+	    unsigned long flags;	/* always 0 for freed obj */
 	    struct RVALUE *next;
 	} free;
 	struct RBasic  basic;
@@ -281,11 +287,11 @@ add_heap()
 	RUBY_CRITICAL(heaps = (heaps_used>0)?
 			(RVALUE**)realloc(heaps, heaps_length*sizeof(RVALUE*)):
 			(RVALUE**)malloc(heaps_length*sizeof(RVALUE*)));
-	if (heaps == 0) mem_error("heaps: can't alloc memory");
+	if (heaps == 0) rb_memerror();
 	RUBY_CRITICAL(heaps_limits = (heaps_used>0)?
 			(int*)realloc(heaps_limits, heaps_length*sizeof(int)):
 			(int*)malloc(heaps_length*sizeof(int)));
-	if (heaps_limits == 0) mem_error("heaps_limits: can't alloc memory");
+	if (heaps_limits == 0) rb_memerror();
     }
 
     for (;;) {
@@ -293,7 +299,7 @@ add_heap()
 	heaps_limits[heaps_used] = heap_slots;
 	if (p == 0) {
 	    if (heap_slots == HEAP_MIN_SLOTS) {
-		mem_error("add_heap: can't alloc memory");
+		rb_memerror();
 	    }
 	    heap_slots = HEAP_MIN_SLOTS;
 	    continue;
@@ -304,10 +310,10 @@ add_heap()
     if (lomem == 0 || lomem > p) lomem = p;
     if (himem < pend) himem = pend;
     heaps_used++;
-    heap_slots *= 2;
+    heap_slots *= 1.8;
 
     while (p < pend) {
-	p->as.free.flag = 0;
+	p->as.free.flags = 0;
 	p->as.free.next = freelist;
 	freelist = p;
 	p++;
@@ -347,6 +353,178 @@ rb_data_object_alloc(klass, datap, dmark, dfree)
 extern st_table *rb_class_tbl;
 VALUE *rb_gc_stack_start = 0;
 
+#if defined(DJGPP) || defined(_WIN32_WCE)
+static unsigned int STACK_LEVEL_MAX = 65535;
+#else
+#ifdef __human68k__
+extern unsigned int _stacksize;
+# define STACK_LEVEL_MAX (_stacksize - 4096)
+# undef HAVE_GETRLIMIT
+#else
+#ifdef HAVE_GETRLIMIT
+static unsigned int STACK_LEVEL_MAX = 655300;
+#else
+# define STACK_LEVEL_MAX 655300
+#endif
+#endif
+#endif
+
+#ifdef C_ALLOCA
+# define SET_STACK_END VALUE stack_end; alloca(0);
+# define STACK_END (&stack_end)
+#else
+# if defined(__GNUC__) && defined(USE_BUILTIN_FRAME_ADDRESS) && !defined(__ia64__)
+#  define  SET_STACK_END    VALUE *stack_end = __builtin_frame_address(0)
+# else
+#  define  SET_STACK_END    VALUE *stack_end = alloca(1)
+# endif
+# define STACK_END (stack_end)
+#endif
+#if defined(sparc) || defined(__sparc__)
+# define STACK_LENGTH  (rb_gc_stack_start - STACK_END + 0x80)
+#elif STACK_GROW_DIRECTION < 0
+# define STACK_LENGTH  (rb_gc_stack_start - STACK_END)
+#elif STACK_GROW_DIRECTION > 0
+# define STACK_LENGTH  (STACK_END - rb_gc_stack_start + 1)
+#else
+# define STACK_LENGTH  ((STACK_END < rb_gc_stack_start) ? rb_gc_stack_start - STACK_END\
+                                           : STACK_END - rb_gc_stack_start + 1)
+#endif
+#if STACK_GROW_DIRECTION > 0
+# define STACK_UPPER(x, a, b) a
+#elif STACK_GROW_DIRECTION < 0
+# define STACK_UPPER(x, a, b) b
+#else
+static int
+stack_growup_p(addr)
+    VALUE *addr;
+{
+    SET_STACK_END;
+
+    if (STACK_END > addr) return Qtrue;
+    return Qfalse;
+}
+# define STACK_UPPER(x, a, b) (stack_growup_p(x) ? a : b)
+#endif
+
+#define GC_WARTER_MARK 512
+
+#define CHECK_STACK(ret) do {\
+    SET_STACK_END;\
+    (ret) = (STACK_LENGTH > STACK_LEVEL_MAX + GC_WARTER_MARK);\
+} while (0)
+
+int
+ruby_stack_length(p)
+    VALUE **p;
+{
+    SET_STACK_END;
+    if (p) *p = STACK_UPPER(STACK_END, rb_gc_stack_start, STACK_END);
+    return STACK_LENGTH;
+}
+
+int
+ruby_stack_check()
+{
+    int ret;
+
+    CHECK_STACK(ret);
+    return ret;
+}
+
+#define MARK_STACK_MAX 1024
+static VALUE mark_stack[MARK_STACK_MAX];
+static VALUE *mark_stack_ptr;
+static int mark_stack_overflow;
+
+static void
+init_mark_stack()
+{
+    mark_stack_overflow = 0;
+    mark_stack_ptr = mark_stack;
+}
+
+#define MARK_STACK_EMPTY (mark_stack_ptr == mark_stack)
+            
+static st_table *source_filenames;
+
+char *
+rb_source_filename(f)
+    const char *f;
+{
+    char *name;
+
+    if (!st_lookup(source_filenames, (st_data_t)f, (st_data_t *)&name)) {
+	long len = strlen(f) + 1;
+	char *ptr = name = ALLOC_N(char, len + 1);
+	*ptr++ = 0;
+	MEMCPY(ptr, f, char, len);
+	st_add_direct(source_filenames, (st_data_t)ptr, (st_data_t)name);
+	return ptr;
+    }
+    return name + 1;
+}
+
+static void
+mark_source_filename(f)
+    char *f;
+{
+    if (f) {
+	f[-1] = 1;
+    }
+}
+
+static int
+sweep_source_filename(key, value)
+    char *key, *value;
+{
+    if (*value) {
+	*value = 0;
+	return ST_CONTINUE;
+    }
+    else {
+	free(value);
+	return ST_DELETE;
+    }
+}
+
+static void rb_gc_mark_children _((VALUE ptr));
+
+static void
+gc_mark_all()
+{
+    RVALUE *p, *pend;
+    int i;
+
+    init_mark_stack();
+    for (i = 0; i < heaps_used; i++) {
+	p = heaps[i]; pend = p + heaps_limits[i];
+	while (p < pend) {
+	    if ((p->as.basic.flags & FL_MARK) &&
+		(p->as.basic.flags != FL_MARK)) {
+		rb_gc_mark_children((VALUE)p);
+	    }
+	    p++;
+	}
+    }
+}
+
+static void
+gc_mark_rest()
+{
+    VALUE tmp_arry[MARK_STACK_MAX];
+    VALUE *p;
+
+    p = (mark_stack_ptr - mark_stack) + tmp_arry;
+    MEMCPY(tmp_arry, mark_stack, VALUE, MARK_STACK_MAX);
+
+    init_mark_stack();
+    while(p != tmp_arry){
+	p--;
+	rb_gc_mark_children(*p);
+    }
+}
+
 static inline int
 is_pointer_to_heap(ptr)
     void *ptr;
@@ -367,48 +545,6 @@ is_pointer_to_heap(ptr)
     return Qfalse;
 }
 
-static st_table *source_filenames;
-
-char *
-rb_source_filename(f)
-    const char *f;
-{
-    char *name;
-
-    if (!st_lookup(source_filenames, f, &name)) {
-	long len = strlen(f) + 1;
-	char *ptr = name = ALLOC_N(char, len + 1);
-	*ptr++ = 0;
-	MEMCPY(ptr, f, char, len);
-	st_add_direct(source_filenames, ptr, name);
-	return ptr;
-    }
-    return name + 1;
-}
-
-static void
-mark_source_filename(f)
-    char *f;
-{
-    if (f) {
-	f[-1] = 1;
-    }
-}
-
-static enum st_retval
-sweep_source_filename(key, value)
-    char *key, *value;
-{
-    if (*value) {
-	*value = 0;
-	return ST_CONTINUE;
-    }
-    else {
-	free(value);
-	return ST_DELETE;
-    }
-}
-
 static void
 mark_locations_array(x, n)
     register VALUE *x;
@@ -429,12 +565,7 @@ rb_gc_mark_locations(start, end)
     VALUE *tmp;
     long n;
 
-    if (start > end) {
-	tmp = start;
-	start = end;
-	end = tmp;
-    }
-    n = end - start + 1;
+    n = end - start;
     mark_locations_array(start,n);
 }
 
@@ -456,7 +587,7 @@ rb_mark_tbl(tbl)
 }
 
 static int
-mark_hashentry(key, value)
+mark_keyvalue(key, value)
     VALUE key;
     VALUE value;
 {
@@ -470,7 +601,7 @@ rb_mark_hash(tbl)
     st_table *tbl;
 {
     if (!tbl) return;
-    st_foreach(tbl, mark_hashentry, 0);
+    st_foreach(tbl, mark_keyvalue, 0);
 }
 
 void
@@ -486,17 +617,49 @@ void
 rb_gc_mark(ptr)
     VALUE ptr;
 {
-    register RVALUE *obj = RANY(ptr);
+    int ret;
+    register RVALUE *obj;
 
-  Top:
-    if (rb_special_const_p((VALUE)obj)) return; /* special const not marked */
+    obj = RANY(ptr);
+    if (rb_special_const_p(ptr)) return; /* special const not marked */
     if (obj->as.basic.flags == 0) return;       /* free cell */
-    if (obj->as.basic.flags & FL_MARK) return;  /* already marked */
-
+    if (obj->as.basic.flags & FL_MARK) return;  /* already marked */ 
     obj->as.basic.flags |= FL_MARK;
 
+    CHECK_STACK(ret);
+    if (ret) {
+	if (!mark_stack_overflow) {
+	    if (mark_stack_ptr - mark_stack < MARK_STACK_MAX) {
+		*mark_stack_ptr = ptr;
+		mark_stack_ptr++;		
+	    }
+	    else {
+		mark_stack_overflow = 1;
+	    }
+	}
+	return;
+    }
+    rb_gc_mark_children(ptr);
+}
+
+static void
+rb_gc_mark_children(ptr)
+    VALUE ptr;
+{
+    register RVALUE *obj = RANY(ptr);
+
+    goto marking;		/* skip */
+
+  again:
+    obj = RANY(ptr);
+    if (rb_special_const_p(ptr)) return; /* special const not marked */
+    if (obj->as.basic.flags == 0) return;       /* free cell */
+    if (obj->as.basic.flags & FL_MARK) return;  /* already marked */ 
+    obj->as.basic.flags |= FL_MARK;
+
+  marking:
     if (FL_TEST(obj, FL_EXIVAR)) {
-	rb_mark_generic_ivar((VALUE)obj);
+	rb_mark_generic_ivar(ptr);
     }
 
     switch (obj->as.basic.flags & T_MASK) {
@@ -516,13 +679,13 @@ rb_gc_mark(ptr)
 	  case NODE_MASGN:
 	  case NODE_RESCUE:
 	  case NODE_RESBODY:
+	  case NODE_CLASS:
 	    rb_gc_mark((VALUE)obj->as.node.u2.node);
 	    /* fall through */
 	  case NODE_BLOCK:	/* 1,3 */
 	  case NODE_ARRAY:
 	  case NODE_DSTR:
 	  case NODE_DXSTR:
-	  case NODE_EVSTR:
 	  case NODE_DREGX:
 	  case NODE_DREGX_ONCE:
 	  case NODE_FBODY:
@@ -536,8 +699,8 @@ rb_gc_mark(ptr)
 	  case NODE_FCALL:
 	  case NODE_DEFN:
 	  case NODE_NEWLINE:
-	    obj = RANY(obj->as.node.u3.node);
-	    goto Top;
+	    ptr = (VALUE)obj->as.node.u3.node;
+	    goto again;
 
 	  case NODE_WHILE:	/* 1,2 */
 	  case NODE_UNTIL:
@@ -553,6 +716,7 @@ rb_gc_mark(ptr)
 	  case NODE_MATCH3:
 	  case NODE_OP_ASGN_OR:
 	  case NODE_OP_ASGN_AND:
+	  case NODE_MODULE:
 	    rb_gc_mark((VALUE)obj->as.node.u1.node);
 	    /* fall through */
 	  case NODE_METHOD:	/* 2 */
@@ -562,14 +726,13 @@ rb_gc_mark(ptr)
 	  case NODE_DASGN:
 	  case NODE_DASGN_CURR:
 	  case NODE_IASGN:
-	  case NODE_CDECL:
 	  case NODE_CVDECL:
 	  case NODE_CVASGN:
-	  case NODE_MODULE:
 	  case NODE_COLON3:
 	  case NODE_OPT_N:
-	    obj = RANY(obj->as.node.u2.node);
-	    goto Top;
+	  case NODE_EVSTR:
+	    ptr = (VALUE)obj->as.node.u2.node;
+	    goto again;
 
 	  case NODE_HASH:	/* 1 */
 	  case NODE_LIT:
@@ -583,15 +746,17 @@ rb_gc_mark(ptr)
 	  case NODE_YIELD:
 	  case NODE_COLON2:
 	  case NODE_ARGS:
-	    obj = RANY(obj->as.node.u1.node);
-	    goto Top;
+	  case NODE_SPLAT:
+	  case NODE_SVALUE:
+	    ptr = (VALUE)obj->as.node.u1.node;
+	    goto again;
 
 	  case NODE_SCOPE:	/* 2,3 */
-	  case NODE_CLASS:
 	  case NODE_BLOCK_PASS:
+	  case NODE_CDECL:
 	    rb_gc_mark((VALUE)obj->as.node.u3.node);
-	    obj = RANY(obj->as.node.u2.node);
-	    goto Top;
+	    ptr = (VALUE)obj->as.node.u2.node;
+	    goto again;
 
 	  case NODE_ZARRAY:	/* - */
 	  case NODE_ZSUPER:
@@ -621,11 +786,11 @@ rb_gc_mark(ptr)
 	  case NODE_ALLOCA:
 	    mark_locations_array((VALUE*)obj->as.node.u1.value,
 				 obj->as.node.u3.cnt);
-	    obj = RANY(obj->as.node.u2.node);
-	    goto Top;
+	    ptr = (VALUE)obj->as.node.u2.node;
+	    goto again;
 #endif
 
-	  default:
+	  default:		/* unlisted NODE */
 	    if (is_pointer_to_heap(obj->as.node.u1.node)) {
 		rb_gc_mark((VALUE)obj->as.node.u1.node);
 	    }
@@ -633,8 +798,7 @@ rb_gc_mark(ptr)
 		rb_gc_mark((VALUE)obj->as.node.u2.node);
 	    }
 	    if (is_pointer_to_heap(obj->as.node.u3.node)) {
-		obj = RANY(obj->as.node.u3.node);
-		goto Top;
+		rb_gc_mark((VALUE)obj->as.node.u3.node);
 	    }
 	}
 	return;			/* no need to mark class. */
@@ -645,30 +809,36 @@ rb_gc_mark(ptr)
       case T_ICLASS:
       case T_CLASS:
       case T_MODULE:
-	rb_gc_mark(obj->as.klass.super);
 	rb_mark_tbl(obj->as.klass.m_tbl);
 	rb_mark_tbl(obj->as.klass.iv_tbl);
-	break;
+	ptr = obj->as.klass.super;
+	goto again;
 
       case T_ARRAY:
-	{
-	    int i, len = obj->as.array.len;
+	if (FL_TEST(obj, ELTS_SHARED)) {
+	    ptr = obj->as.array.aux.shared;
+	    goto again;
+	}
+	else {
+	    long i, len = obj->as.array.len;
 	    VALUE *ptr = obj->as.array.ptr;
 
-	    for (i=0; i < len; i++)
+	    for (i=0; i < len; i++) {
 		rb_gc_mark(*ptr++);
+	    }
 	}
 	break;
 
       case T_HASH:
 	rb_mark_hash(obj->as.hash.tbl);
-	rb_gc_mark(obj->as.hash.ifnone);
-	break;
+	ptr = obj->as.hash.ifnone;
+	goto again;
 
       case T_STRING:
-	if (obj->as.string.orig) {
-	    obj = RANY(obj->as.string.orig);
-	    goto Top;
+#define STR_ASSOC FL_USER3   /* copied from string.c */
+	if (FL_TEST(obj, ELTS_SHARED|STR_ASSOC)) {
+	    ptr = obj->as.string.aux.shared;
+	    goto again;
 	}
 	break;
 
@@ -689,42 +859,41 @@ rb_gc_mark(ptr)
 
       case T_MATCH:
 	if (obj->as.match.str) {
-	    obj = RANY(obj->as.match.str);
-	    goto Top;
+	    ptr = obj->as.match.str;
+	    goto again;
 	}
 	break;
 
       case T_VARMAP:
 	rb_gc_mark(obj->as.varmap.val);
-	obj = RANY(obj->as.varmap.next);
-	goto Top;
-	break;
+	ptr = (VALUE)obj->as.varmap.next;
+	goto again;
 
       case T_SCOPE:
-	if (obj->as.scope.local_vars && (obj->as.scope.flag & SCOPE_MALLOC)) {
+	if (obj->as.scope.local_vars && (obj->as.scope.flags & SCOPE_MALLOC)) {
 	    int n = obj->as.scope.local_tbl[0]+1;
 	    VALUE *vars = &obj->as.scope.local_vars[-1];
 
 	    while (n--) {
-		rb_gc_mark(*vars);
-		vars++;
+		rb_gc_mark(*vars++);
 	    }
 	}
 	break;
 
       case T_STRUCT:
 	{
-	    int i, len = obj->as.rstruct.len;
+	    long len = obj->as.rstruct.len;
 	    VALUE *ptr = obj->as.rstruct.ptr;
 
-	    for (i=0; i < len; i++)
+	    while (len--) {
 		rb_gc_mark(*ptr++);
+	    }
 	}
 	break;
 
       default:
 	rb_bug("rb_gc_mark(): unknown data type 0x%lx(0x%lx) %s",
-	       obj->as.basic.flags & T_MASK, (unsigned long)obj,
+	       obj->as.basic.flags & T_MASK, obj,
 	       is_pointer_to_heap(obj) ? "corrupted object" : "non object");
     }
 }
@@ -736,12 +905,13 @@ gc_sweep()
 {
     RVALUE *p, *pend, *final_list;
     int freed = 0;
-    int i, used = heaps_used;
+    int i, j;
+    unsigned long live = 0;
 
     if (ruby_in_compile && ruby_parser_stack_on_heap()) {
 	/* should not reclaim nodes during compilation
            if yacc's semantic stack is not allocated on machine stack */
-	for (i = 0; i < used; i++) {
+	for (i = 0; i < heaps_used; i++) {
 	    p = heaps[i]; pend = p + heaps_limits[i];
 	    while (p < pend) {
 		if (!(p->as.basic.flags&FL_MARK) && BUILTIN_TYPE(p) == T_NODE)
@@ -757,8 +927,10 @@ gc_sweep()
     freelist = 0;
     final_list = deferred_final_list;
     deferred_final_list = 0;
-    for (i = 0; i < used; i++) {
+    for (i = 0; i < heaps_used; i++) {
 	int n = 0;
+	RVALUE *free = freelist;
+	RVALUE *final = final_list;
 
 	p = heaps[i]; pend = p + heaps_limits[i];
 	while (p < pend) {
@@ -767,12 +939,12 @@ gc_sweep()
 		    obj_free((VALUE)p);
 		}
 		if (need_call_final && FL_TEST(p, FL_FINALIZE)) {
-		    p->as.free.flag = FL_MARK; /* remain marked */
+		    p->as.free.flags = FL_MARK; /* remain marked */
 		    p->as.free.next = final_list;
 		    final_list = p;
 		}
 		else {
-		    p->as.free.flag = 0;
+		    p->as.free.flags = 0;
 		    p->as.free.next = freelist;
 		    freelist = p;
 		}
@@ -784,11 +956,26 @@ gc_sweep()
 	    }
 	    else {
 		RBASIC(p)->flags &= ~FL_MARK;
+		live++;
 	    }
 	    p++;
 	}
-	freed += n;
+	if (n == heaps_limits[i] && freed + n > FREE_MIN) {
+	    RVALUE *pp;
+
+	    heaps_limits[i] = 0;
+	    for (pp = final_list; pp != final; pp = pp->as.free.next) {
+		p->as.free.flags |= FL_SINGLETON; /* freeing page mark */
+	    }
+	    freelist = free;	/* cancel this page from freelist */
+	}
+	else {
+	    freed += n;
+	}
     }
+    malloc_limit += (malloc_increase - malloc_limit) * (double)live / (live + freed);
+    if (malloc_limit < GC_MALLOC_LIMIT) malloc_limit = GC_MALLOC_LIMIT;
+    malloc_increase = 0;
     if (freed < FREE_MIN) {
 	add_heap();
     }
@@ -806,9 +993,24 @@ gc_sweep()
 	for (p = final_list; p; p = tmp) {
 	    tmp = p->as.free.next;
 	    run_final((VALUE)p);
-	    p->as.free.flag = 0;
-	    p->as.free.next = freelist;
-	    freelist = p;
+	    if (!FL_TEST(p, FL_SINGLETON)) { /* not freeing page */
+		p->as.free.flags = 0;
+		p->as.free.next = freelist;
+		freelist = p;
+	    }
+	}
+    }
+    for (i = j = 1; j < heaps_used; i++) {
+	if (heaps_limits[i] == 0) {
+	    free(heaps[i]);
+	    heaps_used--;
+	}
+	else {
+	    if (i != j) {
+		heaps[j] = heaps[i];
+		heaps_limits[j] = heaps_limits[i];
+	    }
+	    j++;
 	}
     }
 }
@@ -817,7 +1019,7 @@ void
 rb_gc_force_recycle(p)
     VALUE p;
 {
-    RANY(p)->as.free.flag = 0;
+    RANY(p)->as.free.flags = 0;
     RANY(p)->as.free.next = freelist;
     freelist = RANY(p);
 }
@@ -847,20 +1049,19 @@ obj_free(obj)
 	break;
       case T_MODULE:
       case T_CLASS:
-	rb_clear_cache();
+	rb_clear_cache_by_class((VALUE)obj);
 	st_free_table(RANY(obj)->as.klass.m_tbl);
 	if (RANY(obj)->as.object.iv_tbl) {
 	    st_free_table(RANY(obj)->as.object.iv_tbl);
 	}
 	break;
       case T_STRING:
-#define STR_NO_ORIG FL_USER2	/* copied from string.c */
-	if (!RANY(obj)->as.string.orig || FL_TEST(obj, STR_NO_ORIG)) {
+	if (RANY(obj)->as.string.ptr && !FL_TEST(obj, ELTS_SHARED)) {
 	    RUBY_CRITICAL(free(RANY(obj)->as.string.ptr));
 	}
 	break;
       case T_ARRAY:
-	if (RANY(obj)->as.array.ptr) {
+	if (RANY(obj)->as.array.ptr && !FL_TEST(obj, ELTS_SHARED)) {
 	    RUBY_CRITICAL(free(RANY(obj)->as.array.ptr));
 	}
 	break;
@@ -930,11 +1131,11 @@ obj_free(obj)
 
       case T_SCOPE:
 	if (RANY(obj)->as.scope.local_vars &&
-            RANY(obj)->as.scope.flag != SCOPE_ALLOCA) {
+            RANY(obj)->as.scope.flags != SCOPE_ALLOCA) {
 	    VALUE *vars = RANY(obj)->as.scope.local_vars-1;
 	    if (vars[0] == 0)
 		RUBY_CRITICAL(free(RANY(obj)->as.scope.local_tbl));
-	    if (RANY(obj)->as.scope.flag&SCOPE_MALLOC)
+	    if (RANY(obj)->as.scope.flags & SCOPE_MALLOC)
 		RUBY_CRITICAL(free(vars));
 	}
 	break;
@@ -946,7 +1147,7 @@ obj_free(obj)
 	break;
 
       default:
-	rb_bug("gc_sweep(): unknown data type %d",
+	rb_bug("gc_sweep(): unknown data type 0x%lx(%ld)", obj,
 	       RANY(obj)->as.basic.flags & T_MASK);
     }
 }
@@ -957,6 +1158,7 @@ rb_gc_mark_frame(frame)
 {
     mark_locations_array(frame->argv, frame->argc);
     rb_gc_mark(frame->cbase);
+    rb_gc_mark((VALUE)frame->node);
 }
 
 #ifdef __GNUC__
@@ -1003,32 +1205,19 @@ rb_gc()
     struct gc_list *list;
     struct FRAME * volatile frame; /* gcc 2.7.2.3 -O2 bug??  */
     jmp_buf save_regs_gc_mark;
-#ifdef C_ALLOCA
-    VALUE stack_end;
-    alloca(0);
-# define STACK_END (&stack_end)
-#else
-# if defined(__GNUC__) && (defined(__i386__) || defined(__mc68000__))
-    VALUE *stack_end = __builtin_frame_address(0);
-# else
-    VALUE *stack_end = alloca(1);
-# endif
-# define STACK_END (stack_end)
-#endif
+    SET_STACK_END;
 
     if (dont_gc || during_gc) {
-	if (!freelist || malloc_memories > GC_MALLOC_LIMIT) {
-	    malloc_memories = 0;
+	if (!freelist) {
 	    add_heap();
 	}
 	return;
     }
-
-    malloc_memories = 0;
-
     if (during_gc) return;
     during_gc++;
 
+    init_mark_stack();
+    
     /* mark frame stack */
     for (frame = ruby_frame; frame; frame = frame->prev) {
 	rb_gc_mark_frame(frame); 
@@ -1040,7 +1229,7 @@ rb_gc()
 	    }
 	}
     }
-    rb_gc_mark(ruby_class);
+    rb_gc_mark((VALUE)ruby_class);
     rb_gc_mark((VALUE)ruby_scope);
     rb_gc_mark((VALUE)ruby_dyna_vars);
     if (finalizer_table) {
@@ -1048,19 +1237,42 @@ rb_gc()
     }
 
     FLUSH_REGISTER_WINDOWS;
-    /* This assumes that all registers are saved into the jmp_buf */
+    /* This assumes that all registers are saved into the jmp_buf (and stack) */
     setjmp(save_regs_gc_mark);
     mark_locations_array((VALUE*)save_regs_gc_mark, sizeof(save_regs_gc_mark) / sizeof(VALUE *));
-    rb_gc_mark_locations(rb_gc_stack_start, (VALUE*)STACK_END);
-#if defined(__human68k__)
-    rb_gc_mark_locations((VALUE*)((char*)rb_gc_stack_start + 2),
-			 (VALUE*)((char*)STACK_END + 2));
+#if STACK_GROW_DIRECTION < 0
+    rb_gc_mark_locations((VALUE*)STACK_END, rb_gc_stack_start);
+#elif STACK_GROW_DIRECTION > 0
+    rb_gc_mark_locations(rb_gc_stack_start, (VALUE*)STACK_END + 1);
+#else
+    if ((VALUE*)STACK_END < rb_gc_stack_start)
+	rb_gc_mark_locations((VALUE*)STACK_END, rb_gc_stack_start);
+    else
+	rb_gc_mark_locations(rb_gc_stack_start, (VALUE*)STACK_END + 1);
+#endif
+#ifdef __ia64__
+    /* mark backing store (flushed register window on the stack) */
+    /* the basic idea from guile GC code                         */
+    {
+	ucontext_t ctx;
+	VALUE *top, *bot;
+	getcontext(&ctx);
+	mark_locations_array((VALUE*)&ctx.uc_mcontext,
+			     ((size_t)(sizeof(VALUE)-1 + sizeof ctx.uc_mcontext)/sizeof(VALUE)));
+	bot = (VALUE*)__libc_ia64_register_backing_store_base;
+	top = (VALUE*)ctx.uc_mcontext.sc_ar_bsp;
+	rb_gc_mark_locations(bot, top);
+    }
+#endif
+#if defined(__human68k__) || defined(__mc68000__)
+    rb_gc_mark_locations((VALUE*)((char*)STACK_END + 2),
+			 (VALUE*)((char*)rb_gc_stack_start + 2));
 #endif
     rb_gc_mark_threads();
 
     /* mark protected global variables */
-    for (list = Global_List; list; list = list->next) {
-	rb_gc_mark(*list->varptr);
+    for (list = global_List; list; list = list->next) {
+	rb_gc_mark_maybe(*list->varptr);
     }
     rb_mark_end_proc();
     rb_gc_mark_global_tbl();
@@ -1072,12 +1284,21 @@ rb_gc()
     rb_mark_generic_ivar_tbl();
 
     rb_gc_mark_parser();
-
+    
+    /* gc_mark objects whose marking are not completed*/
+    while (!MARK_STACK_EMPTY){
+	if (mark_stack_overflow){
+	    gc_mark_all();
+	}
+	else {
+	    gc_mark_rest();
+	}
+    }
     gc_sweep();
 }
 
-static VALUE
-gc_start()
+VALUE
+rb_gc_start()
 {
     rb_gc();
     return Qnil;
@@ -1092,8 +1313,35 @@ Init_stack(addr)
     rb_gc_stack_start = _SEND;
 #else
     if (!addr) addr = (VALUE *)&addr;
+    if (rb_gc_stack_start) {
+	if (STACK_UPPER(&addr,
+			rb_gc_stack_start > addr,
+			rb_gc_stack_start < ++addr))
+	    rb_gc_stack_start = addr;
+	return;
+    }
     rb_gc_stack_start = addr;
 #endif
+#ifdef HAVE_GETRLIMIT
+    {
+	struct rlimit rlim;
+
+	if (getrlimit(RLIMIT_STACK, &rlim) == 0) {
+	    double space = (double)rlim.rlim_cur*0.2;
+
+	    if (space > 1024*1024) space = 1024*1024;
+	    STACK_LEVEL_MAX = (rlim.rlim_cur - space) / sizeof(VALUE);
+	}
+    }
+#ifdef __ia64__
+    /* ruby crashes on IA64 if compiled with optimizer on */
+    /* when if STACK_LEVEL_MAX is greater than this magic number */
+    /* I know this is a kludge.  I suspect optimizer bug */
+#define IA64_MAGIC_STACK_LIMIT 49152
+    if (STACK_LEVEL_MAX > IA64_MAGIC_STACK_LIMIT)
+	STACK_LEVEL_MAX = IA64_MAGIC_STACK_LIMIT;
+#endif
+#endif
 }
 
 void
@@ -1179,6 +1427,7 @@ os_each_obj(argc, argv)
 {
     VALUE of;
 
+    rb_secure(4);
     if (rb_scan_args(argc, argv, "01", &of) == 0) {
 	return os_live_obj();
     }
@@ -1190,25 +1439,25 @@ os_each_obj(argc, argv)
 static VALUE finalizers;
 
 static VALUE
-add_final(os, proc)
-    VALUE os, proc;
+add_final(os, block)
+    VALUE os, block;
 {
     rb_warn("ObjectSpace::add_finalizer is deprecated; use define_finalizer");
-    if (!rb_obj_is_kind_of(proc, rb_cProc)) {
-	rb_raise(rb_eArgError, "wrong type argument %s (Proc required)",
-		 rb_class2name(CLASS_OF(proc)));
+    if (!rb_respond_to(block, rb_intern("call"))) {
+	rb_raise(rb_eArgError, "wrong type argument %s (should be callable)",
+		 rb_obj_classname(block));
     }
-    rb_ary_push(finalizers, proc);
-    return proc;
+    rb_ary_push(finalizers, block);
+    return block;
 }
 
 static VALUE
-rm_final(os, proc)
-    VALUE os, proc;
+rm_final(os, block)
+    VALUE os, block;
 {
     rb_warn("ObjectSpace::remove_finalizer is deprecated; use undefine_finalizer");
-    rb_ary_delete(finalizers, proc);
-    return proc;
+    rb_ary_delete(finalizers, block);
+    return block;
 }
 
 static VALUE
@@ -1222,7 +1471,7 @@ static VALUE
 call_final(os, obj)
     VALUE os, obj;
 {
-    rb_warn("ObjectSpace::call_final is deprecated; use define_finalizer");
+    rb_warn("ObjectSpace::call_finalizer is deprecated; use define_finalizer");
     need_call_final = 1;
     FL_SET(obj, FL_FINALIZE);
     return obj;
@@ -1232,10 +1481,8 @@ static VALUE
 undefine_final(os, obj)
     VALUE os, obj;
 {
-    VALUE table;
-
     if (finalizer_table) {
-	st_delete(finalizer_table, &obj, 0);
+	st_delete(finalizer_table, (st_data_t*)&obj, 0);
     }
     return obj;
 }
@@ -1246,15 +1493,15 @@ define_final(argc, argv, os)
     VALUE *argv;
     VALUE os;
 {
-    VALUE obj, proc, table;
+    VALUE obj, block, table;
 
-    rb_scan_args(argc, argv, "11", &obj, &proc);
+    rb_scan_args(argc, argv, "11", &obj, &block);
     if (argc == 1) {
-	proc = rb_f_lambda();
+	block = rb_block_proc();
     }
-    else if (!rb_obj_is_kind_of(proc, rb_cProc)) {
-	rb_raise(rb_eArgError, "wrong type argument %s (Proc required)",
-		 rb_class2name(CLASS_OF(proc)));
+    else if (!rb_respond_to(block, rb_intern("call"))) {
+	rb_raise(rb_eArgError, "wrong type argument %s (should be callable)",
+		 rb_obj_classname(block));
     }
     need_call_final = 1;
     FL_SET(obj, FL_FINALIZE);
@@ -1263,19 +1510,35 @@ define_final(argc, argv, os)
 	finalizer_table = st_init_numtable();
     }
     if (st_lookup(finalizer_table, obj, &table)) {
-	rb_ary_push(table, proc);
+	rb_ary_push(table, block);
     }
     else {
-	st_add_direct(finalizer_table, obj, rb_ary_new3(1, proc));
+	st_add_direct(finalizer_table, obj, rb_ary_new3(1, block));
+    }
+    return block;
+}
+
+void
+rb_gc_copy_finalizer(dest, obj)
+    VALUE dest, obj;
+{
+    VALUE table;
+
+    if (!finalizer_table) return;
+    if (!FL_TEST(obj, FL_FINALIZE)) return;
+    if (FL_TEST(dest, FL_FINALIZE)) {
+	rb_warn("copy_finalizer: descarding old finalizers");
+    }
+    if (st_lookup(finalizer_table, obj, &table)) {
+	st_insert(finalizer_table, dest, table);
     }
-    return proc;
 }
 
 static VALUE
 run_single_final(args)
     VALUE *args;
 {
-    rb_eval_cmd(args[0], args[1]);
+    rb_eval_cmd(args[0], args[1], 0);
     return Qnil;
 }
 
@@ -1283,20 +1546,23 @@ static void
 run_final(obj)
     VALUE obj;
 {
-    int i, status;
+    long i;
+    int status, critical_save = rb_thread_critical;
     VALUE args[2], table;
 
+    rb_thread_critical = Qtrue;
     args[1] = rb_ary_new3(1, rb_obj_id(obj)); /* make obj into id */
     for (i=0; i<RARRAY(finalizers)->len; i++) {
 	args[0] = RARRAY(finalizers)->ptr[i];
-	rb_protect(run_single_final, (VALUE)args, &status);
+	rb_protect((VALUE(*)_((VALUE)))run_single_final, (VALUE)args, &status);
     }
-    if (finalizer_table && st_delete(finalizer_table, &obj, &table)) {
+    if (finalizer_table && st_delete(finalizer_table, (st_data_t*)&obj, &table)) {
 	for (i=0; i<RARRAY(table)->len; i++) {
 	    args[0] = RARRAY(table)->ptr[i];
-	    rb_protect(run_single_final, (VALUE)args, &status);
+	    rb_protect((VALUE(*)_((VALUE)))run_single_final, (VALUE)args, &status);
 	}
     }
+    rb_thread_critical = critical_save;
 }
 
 void
@@ -1327,17 +1593,22 @@ rb_gc_call_finalizer_at_exit()
 	    }
 	}
     }
-    /* run data object's finaliers */
+    /* run data object's finalizers */
     for (i = 0; i < heaps_used; i++) {
 	p = heaps[i]; pend = p + heaps_limits[i];
 	while (p < pend) {
 	    if (BUILTIN_TYPE(p) == T_DATA &&
 		DATA_PTR(p) && RANY(p)->as.data.dfree) {
-		p->as.free.flag = 0;
-		(*RANY(p)->as.data.dfree)(DATA_PTR(p));
+		p->as.free.flags = 0;
+		if ((long)RANY(p)->as.data.dfree == -1) {
+		    RUBY_CRITICAL(free(DATA_PTR(p)));
+		}
+		else if (RANY(p)->as.data.dfree) {
+		    (*RANY(p)->as.data.dfree)(DATA_PTR(p));
+		}
 	    }
 	    else if (BUILTIN_TYPE(p) == T_FILE) {
-		p->as.free.flag = 0;
+		p->as.free.flags = 0;
 		rb_io_fptr_finalize(RANY(p)->as.file.fptr);
 	    }
 	    p++;
@@ -1377,14 +1648,14 @@ Init_GC()
     VALUE rb_mObSpace;
 
     rb_mGC = rb_define_module("GC");
-    rb_define_singleton_method(rb_mGC, "start", gc_start, 0);
-    rb_define_singleton_method(rb_mGC, "enable", gc_enable, 0);
-    rb_define_singleton_method(rb_mGC, "disable", gc_disable, 0);
-    rb_define_method(rb_mGC, "garbage_collect", gc_start, 0);
+    rb_define_singleton_method(rb_mGC, "start", rb_gc_start, 0);
+    rb_define_singleton_method(rb_mGC, "enable", rb_gc_enable, 0);
+    rb_define_singleton_method(rb_mGC, "disable", rb_gc_disable, 0);
+    rb_define_method(rb_mGC, "garbage_collect", rb_gc_start, 0);
 
     rb_mObSpace = rb_define_module("ObjectSpace");
     rb_define_module_function(rb_mObSpace, "each_object", os_each_obj, -1);
-    rb_define_module_function(rb_mObSpace, "garbage_collect", gc_start, 0);
+    rb_define_module_function(rb_mObSpace, "garbage_collect", rb_gc_start, 0);
     rb_define_module_function(rb_mObSpace, "add_finalizer", add_final, 1);
     rb_define_module_function(rb_mObSpace, "remove_finalizer", rm_final, 1);
     rb_define_module_function(rb_mObSpace, "finalizers", finals, 0);
@@ -1401,4 +1672,7 @@ Init_GC()
     finalizers = rb_ary_new();
 
     source_filenames = st_init_strtable();
+
+    nomem_error = rb_exc_new2(rb_eNoMemError, "failed to allocate memory");
+    rb_global_variable(&nomem_error);
 }
