@@ -2,11 +2,11 @@
 
   inits.c -
 
-  $Author: matz $
-  $Date: 2000/05/01 09:41:20 $
+  $Author: michal $
+  $Date: 2003/01/16 07:34:01 $
   created at: Tue Dec 28 16:01:58 JST 1993
 
-  Copyright (C) 1993-2000 Yukihiro Matsumoto
+  Copyright (C) 1993-2003 Yukihiro Matsumoto
 
 **********************************************************************/
 
