@@ -4,7 +4,7 @@
 
 net/telnet.rb - simple telnet client library
 
-Wakou Aoyama <wakou@fsinet.or.jp>
+Wakou Aoyama <wakou@ruby-lang.org>
 
 
 === MAKE NEW TELNET OBJECT
@@ -90,7 +90,7 @@ of cource, set sync=true or flush is necessary.
 
 Telnet#puts() adds "\n" to the last of "string".
 
-WARNING: Telnet#print() NOT adds "\n" to the last of "string", in the future.
+CAUTION: Telnet#print() NOT adds "\n" to the last of "string".
 
 If "Telnetmode" option is true, then escape IAC code ("\xFF"). If
 "Binmode" option is false, then convert "\n" to EOL(end of line) code.
@@ -237,7 +237,7 @@ module Net
     CR   = "\015"
     LF   = "\012"
     EOL  = CR + LF
-    REVISION = '$Id: telnet.rb,v 1.13.2.4 2001/09/07 22:15:33 wakou Exp $'
+    REVISION = '$Id: telnet.rb,v 1.20 2003/06/03 09:40:21 matz Exp $'
 
     def initialize(options)
       @options = options
@@ -313,10 +313,10 @@ module Net
 
         begin
           if @options["Timeout"] == false
-            @sock = TCPsocket.open(@options["Host"], @options["Port"])
+            @sock = TCPSocket.open(@options["Host"], @options["Port"])
           else
             timeout(@options["Timeout"]) do
-              @sock = TCPsocket.open(@options["Host"], @options["Port"])
+              @sock = TCPSocket.open(@options["Host"], @options["Port"])
             end
           end
         rescue TimeoutError
@@ -480,6 +480,14 @@ module Net
               buf = preprocess(c)
               rest = ''
             end
+         else
+           # Not Telnetmode.
+           #
+           # We cannot use preprocess() on this data, because that
+           # method makes some Telnetmode-specific assumptions.
+           buf = c
+           buf.gsub!(/#{EOL}/no, "\n") unless @options["Binmode"]
+           rest = ''
           end
           @log.print(buf) if @options.has_key?("Output_log")
           line += buf
@@ -504,7 +512,7 @@ module Net
       end
     end
 
-    def _print(string)
+    def print(string)
       string = string.gsub(/#{IAC}/no, IAC + IAC) if @options["Telnetmode"]
 
       if @options["Binmode"]
@@ -524,15 +532,7 @@ module Net
     end
 
     def puts(string)
-      self._print(string + "\n")
-    end
-
-    def print(string)
-      if $VERBOSE
-        $stderr.puts 'WARNING: Telnet#print("string") NOT adds "\n" to the last of "string", in the future.'
-        $stderr.puts '         cf. Telnet#puts().'
-      end
-      self.puts(string)
+      self.print(string + "\n")
     end
 
     def cmd(options)
