@@ -3,22 +3,62 @@
   numeric.c -
 
   $Author: matz $
-  $Date: 2001/11/01 05:09:51 $
+  $Date: 2003/08/04 05:28:50 $
   created at: Fri Aug 13 18:33:09 JST 1993
 
-  Copyright (C) 1993-2000 Yukihiro Matsumoto
+  Copyright (C) 1993-2003 Yukihiro Matsumoto
 
 **********************************************************************/
 
 #include "ruby.h"
+#include "env.h"
 #include <math.h>
 #include <stdio.h>
+
 #if defined(__FreeBSD__) && __FreeBSD__ < 4
 #include <floatingpoint.h>
 #endif
 
-static ID coerce;
-static ID to_i;
+#ifdef HAVE_FLOAT_H
+#include <float.h>
+#endif
+
+/* use IEEE 64bit values if not defined */
+#ifndef FLT_RADIX
+#define FLT_RADIX 2
+#endif
+#ifndef FLT_ROUNDS
+#define FLT_ROUNDS 1
+#endif
+#ifndef DBL_MIN
+#define DBL_MIN 2.2250738585072014e-308
+#endif
+#ifndef DBL_MAX
+#define DBL_MAX 1.7976931348623157e+308
+#endif
+#ifndef DBL_MIN_EXP
+#define DBL_MIN_EXP (-1021)
+#endif
+#ifndef DBL_MAX_EXP
+#define DBL_MAX_EXP 1024
+#endif
+#ifndef DBL_MIN_10_EXP
+#define DBL_MIN_10_EXP (-307)
+#endif
+#ifndef DBL_MAX_10_EXP
+#define DBL_MAX_10_EXP 308
+#endif
+#ifndef DBL_DIG
+#define DBL_DIG 15
+#endif
+#ifndef DBL_MANT_DIG
+#define DBL_MANT_DIG 53
+#endif
+#ifndef DBL_EPSILON
+#define DBL_EPSILON 2.2204460492503131e-16
+#endif
+
+static ID id_coerce, id_to_i, id_eq;
 
 VALUE rb_cNumeric;
 VALUE rb_cFloat;
@@ -47,53 +87,84 @@ static VALUE
 coerce_body(x)
     VALUE *x;
 {
-    return rb_funcall(x[1], coerce, 1, x[0]);
+    return rb_funcall(x[1], id_coerce, 1, x[0]);
 }
 
 static VALUE
 coerce_rescue(x)
     VALUE *x;
 {
+    volatile VALUE v = rb_inspect(x[1]);
+
     rb_raise(rb_eTypeError, "%s can't be coerced into %s",
 	     rb_special_const_p(x[1])?
-	     STR2CSTR(rb_inspect(x[1])):
-	     rb_class2name(CLASS_OF(x[1])),
-	     rb_class2name(CLASS_OF(x[0])));
+	     RSTRING(v)->ptr:
+	     rb_obj_classname(x[1]),
+	     rb_obj_classname(x[0]));
     return Qnil;		/* dummy */
 }
 
-static void
-do_coerce(x, y)
+static int
+do_coerce(x, y, err)
     VALUE *x, *y;
+    int err;
 {
     VALUE ary;
     VALUE a[2];
 
     a[0] = *x; a[1] = *y;
-    ary = rb_rescue2(coerce_body, (VALUE)a, coerce_rescue, (VALUE)a,
-		     rb_eStandardError, rb_eNameError, 0);
+
+    ary = rb_rescue(coerce_body, (VALUE)a, err?coerce_rescue:0, (VALUE)a);
     if (TYPE(ary) != T_ARRAY || RARRAY(ary)->len != 2) {
-	rb_raise(rb_eTypeError, "coerce must return [x, y]");
+	if (err) {
+	    rb_raise(rb_eTypeError, "coerce must return [x, y]");
+	}
+	return Qfalse;
     }
 
     *x = RARRAY(ary)->ptr[0];
     *y = RARRAY(ary)->ptr[1];
+    return Qtrue;
 }
 
 VALUE
 rb_num_coerce_bin(x, y)
     VALUE x, y;
 {
-    do_coerce(&x, &y);
-    return rb_funcall(x, rb_frame_last_func(), 1, y);
+    do_coerce(&x, &y, Qtrue);
+    return rb_funcall(x, ruby_frame->orig_func, 1, y);
+}
+
+VALUE
+rb_num_coerce_cmp(x, y)
+    VALUE x, y;
+{
+    if (do_coerce(&x, &y, Qfalse)) 
+	return rb_funcall(x, ruby_frame->orig_func, 1, y);
+    return Qnil;
+}
+
+VALUE
+rb_num_coerce_relop(x, y)
+    VALUE x, y;
+{
+    VALUE c, x0 = x, y0 = y;
+
+    if (!do_coerce(&x, &y, Qfalse) ||
+	NIL_P(c = rb_funcall(x, ruby_frame->orig_func, 1, y))) {
+	rb_cmperr(x0, y0);
+	return Qnil;		/* not reached */
+    }
+    return c;
 }
 
 static VALUE
-num_clone(x)
-    VALUE x;
+num_init_copy(x, y)
+    VALUE x, y;
 {
-    /* Numerics are immutable values, which need not to copy */
-    return x;
+    /* Numerics are immutable values, which should not be copied */
+    rb_raise(rb_eTypeError, "can't copy %s", rb_obj_classname(x));
+    return Qnil;		/* not reached */
 }
 
 static VALUE
@@ -110,27 +181,30 @@ num_uminus(num)
     VALUE zero;
 
     zero = INT2FIX(0);
-    do_coerce(&zero, &num);
+    do_coerce(&zero, &num, Qtrue);
 
     return rb_funcall(zero, '-', 1, num);
 }
 
+static VALUE
+num_quo(x, y)
+    VALUE x, y;
+{
+    return rb_funcall(x, '/', 1, y);
+}
+
+static VALUE
+num_div(x, y)
+    VALUE x, y;
+{
+    return rb_Integer(rb_funcall(x, '/', 1, y));
+}
+
 static VALUE
 num_divmod(x, y)
     VALUE x, y;
 {
-    VALUE div, mod;
-
-    div = rb_funcall(x, '/', 1, y);
-    if (TYPE(div) == T_FLOAT) {
-	double d = floor(RFLOAT(div)->value);
-
-	if (RFLOAT(div)->value > d) {
-	    div = rb_float_new(d);
-	}
-    }
-    mod = rb_funcall(x, '%', 1, y);
-    return rb_assoc_new(div, mod);
+    return rb_assoc_new(num_div(x, y), rb_funcall(x, '%', 1, y));
 }
 
 static VALUE
@@ -146,7 +220,7 @@ num_remainder(x, y)
 {
     VALUE z = rb_funcall(x, '%', 1, y);
 
-    if ((!RTEST(rb_equal(z, INT2FIX(0)))) &&
+    if ((!rb_equal(z, INT2FIX(0))) &&
 	((RTEST(rb_funcall(x, '<', 1, INT2FIX(0))) &&
 	  RTEST(rb_funcall(y, '>', 1, INT2FIX(0)))) ||
 	 (RTEST(rb_funcall(x, '>', 1, INT2FIX(0))) &&
@@ -177,7 +251,7 @@ static VALUE
 num_zero_p(num)
     VALUE num;
 {
-    if (RTEST(rb_equal(num, INT2FIX(0)))) {
+    if (rb_equal(num, INT2FIX(0))) {
 	return Qtrue;
     }
     return Qfalse;
@@ -193,6 +267,13 @@ num_nonzero_p(num)
     return num;
 }
 
+static VALUE
+num_to_int(num)
+    VALUE num;
+{
+    return rb_funcall(num, id_to_i, 0, 0);
+}
+
 VALUE
 rb_float_new(d)
     double d;
@@ -208,8 +289,8 @@ static VALUE
 flo_to_s(flt)
     VALUE flt;
 {
-    char buf[24];
-    char *fmt = "%.10g";
+    char buf[32];
+    char *fmt = "%.15g";
     double value = RFLOAT(flt)->value;
     double avalue, d1, d2;
 
@@ -217,7 +298,7 @@ flo_to_s(flt)
 	return rb_str_new2(value < 0 ? "-Infinity" : "Infinity");
     else if(isnan(value))
 	return rb_str_new2("NaN");
-
+    
     avalue = fabs(value);
     if (avalue == 0.0) {
 	fmt = "%.1f";
@@ -228,11 +309,12 @@ flo_to_s(flt)
 	d1 = modf(d1, &d2);
 	if (d1 == 0) fmt = "%.1e";
     }    
-    else if (avalue >= 1.0e10) {
+    else if (avalue >= 1.0e15) {
 	d1 = avalue;
 	while (d1 > 10.0) d1 /= 10.0;
 	d1 = modf(d1, &d2);
 	if (d1 == 0) fmt = "%.1e";
+	else fmt = "%.16e";
     }    
     else if ((d1 = modf(value, &d2)) == 0) {
 	fmt = "%.1f";
@@ -422,28 +504,44 @@ num_eql(x, y)
     return rb_equal(x, y);
 }
 
+static VALUE
+num_cmp(x, y)
+    VALUE x, y;
+{
+    if (x == y) return INT2FIX(0);
+    return Qnil;
+}
+
 static VALUE
 num_equal(x, y)
     VALUE x, y;
 {
-    return rb_equal(y, x);
+    if (x == y) return Qtrue;
+    return rb_funcall(y, id_eq, 1, x);
 }
 
 static VALUE
 flo_eq(x, y)
     VALUE x, y;
 {
+    double a, b;
+
     switch (TYPE(y)) {
       case T_FIXNUM:
-	if (RFLOAT(x)->value == FIX2LONG(y)) return Qtrue;
-	return Qfalse;
+	b = FIX2LONG(y);
+	break;
       case T_BIGNUM:
-	return (RFLOAT(x)->value == rb_big2dbl(y))?Qtrue:Qfalse;
+	b = rb_big2dbl(y);
+	break;
       case T_FLOAT:
-	return (RFLOAT(x)->value == RFLOAT(y)->value)?Qtrue:Qfalse;
+	b = RFLOAT(y)->value;
+	break;
       default:
 	return num_equal(x, y);
     }
+    a = RFLOAT(x)->value;
+    if (isnan(a) || isnan(b)) return Qfalse;
+    return (a == b)?Qtrue:Qfalse;
 }
 
 static VALUE
@@ -455,6 +553,7 @@ flo_hash(num)
     int i, hash;
 
     d = RFLOAT(num)->value;
+    if (d == 0) d = fabs(d);
     c = (char*)&d;
     for (hash=0, i=0; i<sizeof(double);i++) {
 	hash += c[i] * 971;
@@ -463,6 +562,17 @@ flo_hash(num)
     return INT2FIX(hash);
 }
 
+VALUE
+rb_dbl_cmp(a, b)
+    double a, b;
+{
+    if (isnan(a) || isnan(b)) return Qnil;
+    if (a == b) return INT2FIX(0);
+    if (a > b) return INT2FIX(1);
+    if (a < b) return INT2FIX(-1);
+    return Qnil;
+}
+
 static VALUE
 flo_cmp(x, y)
     VALUE x, y;
@@ -484,12 +594,9 @@ flo_cmp(x, y)
 	break;
 
       default:
-	return rb_num_coerce_bin(x, y);
+	return rb_num_coerce_cmp(x, y);
     }
-    if (a == b) return INT2FIX(0);
-    if (a > b) return INT2FIX(1);
-    if (a < b) return INT2FIX(-1);
-    rb_raise(rb_eFloatDomainError, "comparing NaN");
+    return rb_dbl_cmp(a, b);
 }
 
 static VALUE
@@ -513,8 +620,9 @@ flo_gt(x, y)
 	break;
 
       default:
-	return rb_num_coerce_bin(x, y);
+	return rb_num_coerce_relop(x, y);
     }
+    if (isnan(a) || isnan(b)) return Qfalse;
     return (a > b)?Qtrue:Qfalse;
 }
 
@@ -539,8 +647,9 @@ flo_ge(x, y)
 	break;
 
       default:
-	return rb_num_coerce_bin(x, y);
+	return rb_num_coerce_relop(x, y);
     }
+    if (isnan(a) || isnan(b)) return Qfalse;
     return (a >= b)?Qtrue:Qfalse;
 }
 
@@ -565,8 +674,9 @@ flo_lt(x, y)
 	break;
 
       default:
-	return rb_num_coerce_bin(x, y);
+	return rb_num_coerce_relop(x, y);
     }
+    if (isnan(a) || isnan(b)) return Qfalse;
     return (a < b)?Qtrue:Qfalse;
 }
 
@@ -591,8 +701,9 @@ flo_le(x, y)
 	break;
 
       default:
-	return rb_num_coerce_bin(x, y);
+	return rb_num_coerce_relop(x, y);
     }
+    if (isnan(a) || isnan(b)) return Qfalse;
     return (a <= b)?Qtrue:Qfalse;
 }
 
@@ -601,7 +712,11 @@ flo_eql(x, y)
     VALUE x, y;
 {
     if (TYPE(y) == T_FLOAT) {
-	if (RFLOAT(x)->value == RFLOAT(y)->value) return Qtrue;
+	double a = RFLOAT(x)->value;
+	double b = RFLOAT(y)->value;
+
+	if (isnan(a) || isnan(b)) return Qfalse;
+	if (a == b) return Qtrue;
     }
     return Qfalse;
 }
@@ -631,39 +746,45 @@ flo_zero_p(num)
     return Qfalse;
 }
 
-static VALUE flo_is_nan_p(num)
+static VALUE
+flo_is_nan_p(num)
      VALUE num;
 {     
+    double value = RFLOAT(num)->value;
 
-  double value = RFLOAT(num)->value;
-
-  return isnan(value) ? Qtrue : Qfalse;
+    return isnan(value) ? Qtrue : Qfalse;
 }
 
-static VALUE flo_is_infinite_p(num)
+static VALUE
+flo_is_infinite_p(num)
      VALUE num;
 {     
-  double value = RFLOAT(num)->value;
+    double value = RFLOAT(num)->value;
 
-  if (isinf(value)) {
-    return INT2FIX( value < 0 ? -1 : +1 );
-  }
+    if (isinf(value)) {
+	return INT2FIX( value < 0 ? -1 : 1 );
+    }
 
-  return Qnil;
+    return Qnil;
 }
 
-static VALUE flo_is_finite_p(num)
+static VALUE
+flo_is_finite_p(num)
      VALUE num;
 {     
-  double value = RFLOAT(num)->value;
+    double value = RFLOAT(num)->value;
 
-  if (isinf(value) || isnan(value))
-    return Qfalse;
-  
-  return Qtrue;
+#if HAVE_FINITE
+    if (!finite(value))
+	return Qfalse;
+#else
+    if (isinf(value) || isnan(value))
+	return Qfalse;
+#endif
+
+    return Qtrue;
 }
 
-
 static VALUE
 flo_floor(num)
     VALUE num;
@@ -675,7 +796,7 @@ flo_floor(num)
 	return rb_dbl2big(f);
     }
     val = f;
-    return INT2FIX(val);
+    return LONG2FIX(val);
 }
 
 static VALUE
@@ -689,7 +810,7 @@ flo_ceil(num)
 	return rb_dbl2big(f);
     }
     val = f;
-    return INT2FIX(val);
+    return LONG2FIX(val);
 }
 
 static VALUE
@@ -706,7 +827,7 @@ flo_round(num)
 	return rb_dbl2big(f);
     }
     val = f;
-    return INT2FIX(val);
+    return LONG2FIX(val);
 }
 
 static VALUE
@@ -723,7 +844,7 @@ flo_truncate(num)
 	return rb_dbl2big(f);
     }
     val = f;
-    return INT2FIX(val);
+    return LONG2FIX(val);
 }
 
 static VALUE
@@ -754,12 +875,91 @@ num_truncate(num)
     return flo_truncate(rb_Float(num));
 }
 
+static VALUE
+num_step(argc, argv, from)
+    int argc;
+    VALUE *argv;
+    VALUE from;
+{
+    VALUE to, step;
+
+    if (argc == 1) {
+	to = argv[0];
+	step = INT2FIX(1);
+    }
+    else {
+	if (argc == 2) {
+	    to = argv[0];
+	    step = argv[1];
+	}
+	else {
+	    rb_raise(rb_eArgError, "wrong number of arguments");
+	}
+	if (rb_equal(step, INT2FIX(0))) {
+	    rb_raise(rb_eArgError, "step cannot be 0");
+	}
+    }
+
+    if (FIXNUM_P(from) && FIXNUM_P(to) && FIXNUM_P(step)) {
+	long i, end, diff;
+
+	i = FIX2LONG(from);
+	end = FIX2LONG(to);
+	diff = FIX2LONG(step);
+
+	if (diff > 0) {
+	    while (i <= end) {
+		rb_yield(LONG2FIX(i));
+		i += diff;
+	    }
+	}
+	else {
+	    while (i >= end) {
+		rb_yield(LONG2FIX(i));
+		i += diff;
+	    }
+	}
+    }
+    else if (TYPE(from) == T_FLOAT || TYPE(to) == T_FLOAT || TYPE(step) == T_FLOAT) {
+	const double epsilon = DBL_EPSILON;
+	double beg = NUM2DBL(from);
+	double end = NUM2DBL(to);
+	double unit = NUM2DBL(step);
+	double n = (end - beg)/unit;
+	double err = (fabs(beg) + fabs(end) + fabs(end-beg)) / fabs(unit) * epsilon;
+	long i;
+
+	if (err>0.5) err=0.5;
+	n = floor(n + err) + 1;
+	for (i=0; i<n; i++) {
+	    rb_yield(rb_float_new(i*unit+beg));
+	}
+    }
+    else {
+	VALUE i = from;
+	ID cmp;
+
+	if (RTEST(rb_funcall(step, '>', 1, INT2FIX(0)))) {
+	    cmp = '>';
+	}
+	else {
+	    cmp = '<';
+	}
+	for (;;) {
+	    if (RTEST(rb_funcall(i, cmp, 1, to))) break;
+	    rb_yield(i);
+	    i = rb_funcall(i, '+', 1, step);
+	}
+    }
+    return from;
+}
+
 long
 rb_num2long(val)
     VALUE val;
 {
     if (NIL_P(val)) {
-	rb_raise(rb_eTypeError, "no implicit conversion from nil");
+	rb_raise(rb_eTypeError, "no implicit conversion from nil to integer");
     }
 
     if (FIXNUM_P(val)) return FIX2LONG(val);
@@ -782,15 +982,9 @@ rb_num2long(val)
       case T_BIGNUM:
 	return rb_big2long(val);
 
-      case T_STRING:
-	rb_raise(rb_eTypeError, "no implicit conversion from string");
-	return Qnil;		/* not reached */
-
-      case T_TRUE:
-      case T_FALSE:
-	rb_raise(rb_eTypeError, "no implicit conversion from boolean");
-	return Qnil;		/* not reached */
-
+      case T_SYMBOL:
+	rb_warning("treating Symbol as an integer");
+	/* fall through */
       default:
 	val = rb_to_int(val);
 	return NUM2LONG(val);
@@ -808,15 +1002,40 @@ rb_num2ulong(val)
 }
 
 #if SIZEOF_INT < SIZEOF_LONG
+static void
+check_int(num)
+    long num;
+{
+    const char *s;
+
+    if (num < INT_MIN) {
+	s = "small";
+    }
+    else if (num > INT_MAX) {
+	s = "big";
+    }
+    else {
+	return;
+    }
+    rb_raise(rb_eRangeError, "integer %ld too %s to convert to `int'", num, s);
+}
+
+static void
+check_uint(num)
+    unsigned long num;
+{
+    if (num > INT_MAX) {
+	rb_raise(rb_eRangeError, "integer %lu too big to convert to `int'", num);
+    }
+}
+
 int
 rb_num2int(val)
     VALUE val;
 {
     long num = rb_num2long(val);
 
-    if (num < INT_MIN || INT_MAX < num) {
-	rb_raise(rb_eRangeError, "integer %d too big to convert to `int'", num);
-    }
+    check_int(num);
     return (int)num;
 }
 
@@ -826,9 +1045,27 @@ rb_fix2int(val)
 {
     long num = FIXNUM_P(val)?FIX2LONG(val):rb_num2long(val);
 
-    if (num < INT_MIN || INT_MAX < num) {
-	rb_raise(rb_eRangeError, "integer %d too big to convert to `int'", num);
-    }
+    check_int(num);
+    return (int)num;
+}
+
+unsigned int
+rb_num2uint(val)
+    VALUE val;
+{
+    unsigned long num = rb_num2ulong(val);
+
+    check_uint(num);
+    return (int)num;
+}
+
+unsigned int
+rb_fix2uint(val)
+    VALUE val;
+{
+    unsigned long num = FIXNUM_P(val)?FIX2LONG(val):rb_num2ulong(val);
+
+    check_uint(num);
     return (int)num;
 }
 #else
@@ -857,10 +1094,67 @@ rb_num2fix(val)
 
     v = rb_num2long(val);
     if (!FIXABLE(v))
-	rb_raise(rb_eRangeError, "integer %d out of range of fixnum", v);
-    return INT2FIX(v);
+	rb_raise(rb_eRangeError, "integer %ld out of range of fixnum", v);
+    return LONG2FIX(v);
 }
 
+#if HAVE_LONG_LONG
+
+LONG_LONG
+rb_num2ll(val)
+    VALUE val;
+{
+    if (NIL_P(val)) {
+	rb_raise(rb_eTypeError, "no implicit conversion from nil");
+    }
+
+    if (FIXNUM_P(val)) return (LONG_LONG)FIX2LONG(val);
+
+    switch (TYPE(val)) {
+    case T_FLOAT:
+	if (RFLOAT(val)->value <= (double)LLONG_MAX
+	    && RFLOAT(val)->value >= (double)LLONG_MIN) {
+	    return (LONG_LONG)(RFLOAT(val)->value);
+	}
+	else {
+	    char buf[24];
+	    char *s;
+
+	    sprintf(buf, "%-.10g", RFLOAT(val)->value);
+	    if (s = strchr(buf, ' ')) *s = '\0';
+	    rb_raise(rb_eRangeError, "float %s out of range of long long", buf);
+	}
+
+    case T_BIGNUM:
+	return rb_big2ll(val);
+
+    case T_STRING:
+	rb_raise(rb_eTypeError, "no implicit conversion from string");
+	return Qnil;            /* not reached */
+
+    case T_TRUE:
+    case T_FALSE:
+	rb_raise(rb_eTypeError, "no implicit conversion from boolean");
+	return Qnil;		/* not reached */
+
+      default:
+	  val = rb_to_int(val);
+	  return NUM2LL(val);
+    }
+}
+
+unsigned LONG_LONG
+rb_num2ull(val)
+    VALUE val;
+{
+    if (TYPE(val) == T_BIGNUM) {
+	return rb_big2ull(val);
+    }
+    return (unsigned LONG_LONG)rb_num2ll(val);
+}
+
+#endif  /* HAVE_LONG_LONG */
+
 static VALUE
 int_to_i(num)
     VALUE num;
@@ -879,6 +1173,10 @@ static VALUE
 int_succ(num)
     VALUE num;
 {
+    if (FIXNUM_P(num)) {
+	long i = FIX2LONG(num) + 1;
+	return LONG2NUM(i);
+    }
     return rb_funcall(num, '+', 1, INT2FIX(1));
 }
 
@@ -890,7 +1188,7 @@ int_chr(num)
     long i = NUM2LONG(num);
 
     if (i < 0 || 0xff < i)
-	rb_raise(rb_eRangeError, "%d out of char range", i);
+	rb_raise(rb_eRangeError, "%ld out of char range", i);
     c = i;
     return rb_str_new(&c, 1);
 }
@@ -911,10 +1209,10 @@ rb_int_induced_from(klass, x)
     case T_BIGNUM:
        return x;
     case T_FLOAT:
-       return rb_funcall(x, rb_intern("to_i"), 0);
+       return rb_funcall(x, id_to_i, 0);
     default:
        rb_raise(rb_eTypeError, "failed to convert %s into Integer",
-                rb_class2name(CLASS_OF(x)));
+                rb_obj_classname(x));
     }
 }
 
@@ -930,7 +1228,7 @@ rb_flo_induced_from(klass, x)
        return x;
     default:
        rb_raise(rb_eTypeError, "failed to convert %s into Float",
-                rb_class2name(CLASS_OF(x)));
+                rb_obj_classname(x));
     }
 }
 
@@ -938,7 +1236,7 @@ static VALUE
 fix_uminus(num)
     VALUE num;
 {
-    return rb_int2inum(-FIX2LONG(num));
+    return LONG2NUM(-FIX2LONG(num));
 }
 
 VALUE
@@ -946,23 +1244,50 @@ rb_fix2str(x, base)
     VALUE x;
     int base;
 {
-    char fmt[4], buf[22];
+    extern const char ruby_digitmap[];
+    char buf[SIZEOF_LONG*CHAR_BIT + 2], *b = buf + sizeof buf;
+    long val = FIX2LONG(x);
+    int neg = 0;
 
-    fmt[0] = '%'; fmt[1] = 'l'; fmt[3] = '\0';
-    if (base == 10) fmt[2] = 'd';
-    else if (base == 16) fmt[2] = 'x';
-    else if (base == 8) fmt[2] = 'o';
-    else rb_fatal("fixnum cannot treat base %d", base);
+    if (base < 2 || 36 < base) {
+	rb_raise(rb_eArgError, "illegal radix %d", base);
+    }
+    if (val == 0) {
+	return rb_str_new2("0");
+    }
+    if (val < 0) {
+	val = -val;
+	neg = 1;
+    }
+    *--b = '\0';
+    do {
+	*--b = ruby_digitmap[(int)(val % base)];
+    } while (val /= base);
+    if (neg) {
+	*--b = '-';
+    }
 
-    snprintf(buf, 22, fmt, FIX2LONG(x));
-    return rb_str_new2(buf);
+    return rb_str_new2(b);
 }
 
 static VALUE
-fix_to_s(in)
-    VALUE in;
+fix_to_s(argc, argv, x)
+    int argc;
+    VALUE *argv;
+    VALUE x;
 {
-    return rb_fix2str(in, 10);
+    VALUE b;
+    int base;
+
+    rb_scan_args(argc, argv, "01", &b);
+    if (argc == 0) base = 10;
+    else base = NUM2INT(b);
+
+    if (base == 2) {
+	/* rb_fix2str() does not handle binary */
+	return rb_big2str(rb_int2big(FIX2INT(x)), 2);
+    }
+    return rb_fix2str(x, base);
 }
 
 static VALUE
@@ -976,7 +1301,7 @@ fix_plus(x, y)
 	a = FIX2LONG(x);
 	b = FIX2LONG(y);
 	c = a + b;
-	r = INT2FIX(c);
+	r = LONG2FIX(c);
 
 	if (FIX2LONG(r) != c) {
 	    r = rb_big_plus(rb_int2big(a), rb_int2big(b));
@@ -1000,7 +1325,7 @@ fix_minus(x, y)
 	a = FIX2LONG(x);
 	b = FIX2LONG(y);
 	c = a - b;
-	r = INT2FIX(c);
+	r = LONG2FIX(c);
 
 	if (FIX2LONG(r) != c) {
 	    r = rb_big_minus(rb_int2big(a), rb_int2big(b));
@@ -1026,7 +1351,7 @@ fix_mul(x, y)
 
 	b = FIX2LONG(y);
 	c = a * b;
-	r = INT2FIX(c);
+	r = LONG2FIX(c);
 
 	if (FIX2LONG(r) != c || c/a != b) {
 	    r = rb_big_mul(rb_int2big(a), rb_int2big(b));
@@ -1068,6 +1393,16 @@ fixdivmod(x, y, divp, modp)
     if (modp) *modp = mod;
 }
 
+static VALUE
+fix_quo(x, y)
+    VALUE x, y;
+{
+    if (FIXNUM_P(y)) {
+	return rb_float_new((double)FIX2LONG(x) / (double)FIX2LONG(y));
+    }
+    return rb_num_coerce_bin(x, y);
+}
+
 static VALUE
 fix_div(x, y)
     VALUE x, y;
@@ -1076,7 +1411,7 @@ fix_div(x, y)
 	long div;
 
 	fixdivmod(FIX2LONG(x), FIX2LONG(y), &div, 0);
-	return INT2NUM(div);
+	return LONG2NUM(div);
     }
     return rb_num_coerce_bin(x, y);
 }
@@ -1089,7 +1424,7 @@ fix_mod(x, y)
 	long mod;
 
 	fixdivmod(FIX2LONG(x), FIX2LONG(y), 0, &mod);
-	return INT2NUM(mod);
+	return LONG2NUM(mod);
     }
     return rb_num_coerce_bin(x, y);
 }
@@ -1103,7 +1438,7 @@ fix_divmod(x, y)
 
 	fixdivmod(FIX2LONG(x), FIX2LONG(y), &div, &mod);
 
-	return rb_assoc_new(INT2NUM(div), INT2NUM(mod));
+	return rb_assoc_new(LONG2NUM(div), LONG2NUM(mod));
     }
     return rb_num_coerce_bin(x, y);
 }
@@ -1132,7 +1467,7 @@ fix_equal(x, y)
     VALUE x, y;
 {
     if (FIXNUM_P(y)) {
-	return (x == y)?Qtrue:Qfalse;
+	return (FIX2LONG(x) == FIX2LONG(y))?Qtrue:Qfalse;
     }
     else {
 	return num_equal(x, y);
@@ -1151,7 +1486,7 @@ fix_cmp(x, y)
 	return INT2FIX(-1);
     }
     else {
-	return rb_num_coerce_bin(x, y);
+	return rb_num_coerce_cmp(x, y);
     }
 }
 
@@ -1166,7 +1501,7 @@ fix_gt(x, y)
 	return Qfalse;
     }
     else {
-	return rb_num_coerce_bin(x, y);
+	return rb_num_coerce_relop(x, y);
     }
 }
 
@@ -1181,7 +1516,7 @@ fix_ge(x, y)
 	return Qfalse;
     }
     else {
-	return rb_num_coerce_bin(x, y);
+	return rb_num_coerce_relop(x, y);
     }
 }
 
@@ -1196,7 +1531,7 @@ fix_lt(x, y)
 	return Qfalse;
     }
     else {
-	return rb_num_coerce_bin(x, y);
+	return rb_num_coerce_relop(x, y);
     }
 }
 
@@ -1211,7 +1546,7 @@ fix_le(x, y)
 	return Qfalse;
     }
     else {
-	return rb_num_coerce_bin(x, y);
+	return rb_num_coerce_relop(x, y);
     }
 }
 
@@ -1222,7 +1557,7 @@ fix_rev(num)
     long val = FIX2LONG(num);
 
     val = ~val;
-    return rb_int2inum(val);
+    return LONG2NUM(val);
 }
 
 static VALUE
@@ -1235,7 +1570,7 @@ fix_and(x, y)
 	return rb_big_and(y, x);
     }
     val = FIX2LONG(x) & NUM2LONG(y);
-    return rb_int2inum(val);
+    return LONG2NUM(val);
 }
 
 static VALUE
@@ -1248,7 +1583,7 @@ fix_or(x, y)
 	return rb_big_or(y, x);
     }
     val = FIX2LONG(x) | NUM2LONG(y);
-    return rb_int2inum(val);
+    return LONG2NUM(val);
 }
 
 static VALUE
@@ -1261,9 +1596,11 @@ fix_xor(x, y)
 	return rb_big_xor(y, x);
     }
     val = FIX2LONG(x) ^ NUM2LONG(y);
-    return rb_int2inum(val);
+    return LONG2NUM(val);
 }
 
+static VALUE fix_rshift _((VALUE, VALUE));
+
 static VALUE
 fix_lshift(x, y)
     VALUE x, y;
@@ -1272,12 +1609,14 @@ fix_lshift(x, y)
 
     val = NUM2LONG(x);
     width = NUM2LONG(y);
+    if (width < 0)
+	return fix_rshift(x, LONG2FIX(-width));
     if (width > (sizeof(VALUE)*CHAR_BIT-1)
 	|| ((unsigned long)val)>>(sizeof(VALUE)*CHAR_BIT-1-width) > 0) {
 	return rb_big_lshift(rb_int2big(val), y);
     }
     val = val << width;
-    return rb_int2inum(val);
+    return LONG2NUM(val);
 }
 
 static VALUE
@@ -1288,14 +1627,15 @@ fix_rshift(x, y)
 
     i = NUM2LONG(y);
     if (i < 0)
-	return fix_lshift(x, INT2FIX(-i));
+	return fix_lshift(x, LONG2FIX(-i));
     if (i == 0) return x;
+    val = FIX2LONG(x);
     if (i >= sizeof(long)*CHAR_BIT-1) {
-	if (i < 0) return INT2FIX(-1);
+	if (val < 0) return INT2FIX(-1);
 	return INT2FIX(0);
     }
-    val = RSHIFT(FIX2LONG(x), i);
-    return INT2FIX(val);
+    val = RSHIFT(val, i);
+    return LONG2FIX(val);
 }
 
 static VALUE
@@ -1303,23 +1643,26 @@ fix_aref(fix, idx)
     VALUE fix, idx;
 {
     long val = FIX2LONG(fix);
+    long i;
 
     if (TYPE(idx) == T_BIGNUM) {
-	if (!RBIGNUM(idx)->sign || val >= 0)
-	    return INT2FIX(0);
-	return INT2FIX(1);
-    }
-    else {
-	int i = NUM2INT(idx);
-
-	if (i < 0 || sizeof(VALUE)*CHAR_BIT-1 < i) {
-	    if (val < 0) return INT2FIX(1);
-	    return INT2FIX(0);
-	}
-	if (val & (1L<<i))
+	idx = rb_big_norm(idx);
+	if (!FIXNUM_P(idx)) {
+	    if (!RBIGNUM(idx)->sign || val >= 0)
+		return INT2FIX(0);
 	    return INT2FIX(1);
+	}
+    }
+    i = NUM2LONG(idx);
+
+    if (i < 0) return INT2FIX(0);
+    if (sizeof(VALUE)*CHAR_BIT-1 < i) {
+	if (val < 0) return INT2FIX(1);
 	return INT2FIX(0);
     }
+    if (val & (1L<<i))
+	return INT2FIX(1);
+    return INT2FIX(0);
 }
 
 static VALUE
@@ -1333,13 +1676,6 @@ fix_to_f(num)
     return rb_float_new(val);
 }
 
-static VALUE
-fix_type(fix)
-    VALUE fix;
-{
-    return rb_cFixnum;
-}
-
 static VALUE
 fix_abs(fix)
     VALUE fix;
@@ -1348,7 +1684,7 @@ fix_abs(fix)
 
     if (i < 0) i = -i;
 
-    return rb_int2inum(i);
+    return LONG2NUM(i);
 }
 
 static VALUE
@@ -1361,12 +1697,15 @@ fix_id2name(fix)
 }
 
 static VALUE
-fix_succ(fix)
+fix_to_sym(fix)
     VALUE fix;
 {
-    long i = FIX2LONG(fix) + 1;
+    ID id = FIX2UINT(fix);
 
-    return rb_int2inum(i);
+    if (rb_id2name(id)) {
+	return ID2SYM(id);
+    }
+    return Qnil;
 }
 
 static VALUE
@@ -1380,12 +1719,22 @@ static VALUE
 int_upto(from, to)
     VALUE from, to;
 {
-    VALUE i = from;
+    if (FIXNUM_P(from) && FIXNUM_P(to)) {
+	long i, end;
 
-    for (;;) {
-	if (RTEST(rb_funcall(i, '>', 1, to))) break;
-	rb_yield(i);
-	i = rb_funcall(i, '+', 1, INT2FIX(1));
+	end = FIX2LONG(to);
+	for (i = FIX2LONG(from); i <= end; i++) {
+	    rb_yield(LONG2FIX(i));
+	}
+    }
+    else {
+	VALUE i = from, c;
+
+	while (!(c = rb_funcall(i, '>', 1, to))) {
+	    rb_yield(i);
+	    i = rb_funcall(i, '+', 1, INT2FIX(1));
+	}
+	if (NIL_P(c)) rb_cmperr(i, to);
     }
     return from;
 }
@@ -1394,37 +1743,22 @@ static VALUE
 int_downto(from, to)
     VALUE from, to;
 {
-    VALUE i = from;
+    if (FIXNUM_P(from) && FIXNUM_P(to)) {
+	long i, end;
 
-    for (;;) {
-	if (RTEST(rb_funcall(i, '<', 1, to))) break;
-	rb_yield(i);
-	i = rb_funcall(i, '-', 1, INT2FIX(1));
-    }
-    return from;
-}
-
-static VALUE
-int_step(from, to, step)
-    VALUE from, to, step;
-{
-    VALUE i = from;
-    ID cmp;
-
-    if (NUM2INT(step) == 0) {
-	rb_raise(rb_eArgError, "step cannot be 0");
-    }
-
-    if (RTEST(rb_funcall(step, '>', 1, INT2FIX(0)))) {
-	cmp = '>';
+	end = FIX2LONG(to);
+	for (i=FIX2LONG(from); i >= end; i--) {
+	    rb_yield(LONG2FIX(i));
+	}
     }
     else {
-	cmp = '<';
-    }
-    for (;;) {
-	if (RTEST(rb_funcall(i, cmp, 1, to))) break;
-	rb_yield(i);
-	i = rb_funcall(i, '+', 1, step);
+	VALUE i = from, c;
+
+	while (!(c = rb_funcall(i, '<', 1, to))) {
+	    rb_yield(i);
+	    i = rb_funcall(i, '-', 1, INT2FIX(1));
+	}
+	if (NIL_P(c)) rb_cmperr(i, to);
     }
     return from;
 }
@@ -1433,94 +1767,23 @@ static VALUE
 int_dotimes(num)
     VALUE num;
 {
-    VALUE i = INT2FIX(0);
+    if (FIXNUM_P(num)) {
+	long i, end;
 
-    for (;;) {
-	if (!RTEST(rb_funcall(i, '<', 1, num))) break;
-	rb_yield(i);
-	i = rb_funcall(i, '+', 1, INT2FIX(1));
-    }
-    return num;
-}
-
-static VALUE
-fix_upto(from, to)
-    VALUE from, to;
-{
-    long i, end;
-
-    if (!FIXNUM_P(to)) return int_upto(from, to);
-    end = FIX2LONG(to);
-    for (i = FIX2LONG(from); i <= end; i++) {
-	rb_yield(INT2FIX(i));
-    }
-
-    return from;
-}
-
-VALUE
-rb_fix_upto(from, to)
-    VALUE from, to;
-{
-    return fix_upto(from, to);
-}
-
-static VALUE
-fix_downto(from, to)
-    VALUE from, to;
-{
-    long i, end;
-
-    if (!FIXNUM_P(to)) return int_downto(from, to);
-    end = FIX2LONG(to);
-    for (i=FIX2LONG(from); i >= end; i--) {
-	rb_yield(INT2FIX(i));
-    }
-
-    return from;
-}
-
-static VALUE
-fix_step(from, to, step)
-    VALUE from, to, step;
-{
-    long i, end, diff;
-
-    if (!FIXNUM_P(to) || !FIXNUM_P(step))
-	return int_step(from, to, step);
-
-    i = FIX2LONG(from);
-    end = FIX2LONG(to);
-    diff = FIX2LONG(step);
-
-    if (diff == 0) {
-	rb_raise(rb_eArgError, "step cannot be 0");
-    }
-    if (diff > 0) {
-	while (i <= end) {
-	    rb_yield(INT2FIX(i));
-	    i += diff;
+	end = FIX2LONG(num);
+	for (i=0; i<end; i++) {
+	    rb_yield(LONG2FIX(i));
 	}
     }
     else {
-	while (i >= end) {
-	    rb_yield(INT2FIX(i));
-	    i += diff;
+	VALUE i = INT2FIX(0);
+
+	for (;;) {
+	    if (!RTEST(rb_funcall(i, '<', 1, num))) break;
+	    rb_yield(i);
+	    i = rb_funcall(i, '+', 1, INT2FIX(1));
 	}
     }
-    return from;
-}
-
-static VALUE
-fix_dotimes(num)
-    VALUE num;
-{
-    long i, end;
-
-    end = FIX2LONG(num);
-    for (i=0; i<end; i++) {
-	rb_yield(INT2FIX(i));
-    }
     return num;
 }
 
@@ -1541,25 +1804,29 @@ Init_Numeric()
     /* allow divide by zero -- Inf */
     fpsetmask(fpgetmask() & ~(FP_X_DZ|FP_X_INV|FP_X_OFL));
 #endif
-    coerce = rb_intern("coerce");
-    to_i = rb_intern("to_i");
+    id_coerce = rb_intern("coerce");
+    id_to_i = rb_intern("to_i");
+    id_eq = rb_intern("==");
 
     rb_eZeroDivError = rb_define_class("ZeroDivisionError", rb_eStandardError);
     rb_eFloatDomainError = rb_define_class("FloatDomainError", rb_eRangeError);
     rb_cNumeric = rb_define_class("Numeric", rb_cObject);
 
     rb_include_module(rb_cNumeric, rb_mComparable);
+    rb_define_method(rb_cNumeric, "initialize_copy", num_init_copy, 1);
     rb_define_method(rb_cNumeric, "coerce", num_coerce, 1);
-    rb_define_method(rb_cNumeric, "clone", num_clone, 0);
 
     rb_define_method(rb_cNumeric, "+@", num_uplus, 0);
     rb_define_method(rb_cNumeric, "-@", num_uminus, 0);
-    rb_define_method(rb_cNumeric, "===", num_equal, 1);
+    rb_define_method(rb_cNumeric, "<=>", num_cmp, 1);
     rb_define_method(rb_cNumeric, "eql?", num_eql, 1);
+    rb_define_method(rb_cNumeric, "quo", num_quo, 1);
+    rb_define_method(rb_cNumeric, "div", num_div, 1);
     rb_define_method(rb_cNumeric, "divmod", num_divmod, 1);
     rb_define_method(rb_cNumeric, "modulo", num_modulo, 1);
     rb_define_method(rb_cNumeric, "remainder", num_remainder, 1);
     rb_define_method(rb_cNumeric, "abs", num_abs, 0);
+    rb_define_method(rb_cNumeric, "to_int", num_to_int, 0);
 
     rb_define_method(rb_cNumeric, "integer?", num_int_p, 0);
     rb_define_method(rb_cNumeric, "zero?", num_zero_p, 0);
@@ -1569,14 +1836,15 @@ Init_Numeric()
     rb_define_method(rb_cNumeric, "ceil", num_ceil, 0);
     rb_define_method(rb_cNumeric, "round", num_round, 0);
     rb_define_method(rb_cNumeric, "truncate", num_truncate, 0);
+    rb_define_method(rb_cNumeric, "step", num_step, -1);
 
     rb_cInteger = rb_define_class("Integer", rb_cNumeric);
+    rb_undef_alloc_func(rb_cInteger);
     rb_undef_method(CLASS_OF(rb_cInteger), "new");
 
     rb_define_method(rb_cInteger, "integer?", int_int_p, 0);
     rb_define_method(rb_cInteger, "upto", int_upto, 1);
     rb_define_method(rb_cInteger, "downto", int_downto, 1);
-    rb_define_method(rb_cInteger, "step", int_step, 2);
     rb_define_method(rb_cInteger, "times", int_dotimes, 0);
     rb_include_module(rb_cInteger, rb_mPrecision);
     rb_define_method(rb_cInteger, "succ", int_succ, 0);
@@ -1594,19 +1862,21 @@ Init_Numeric()
     rb_define_singleton_method(rb_cFixnum, "induced_from", rb_fix_induced_from, 1);
     rb_define_singleton_method(rb_cInteger, "induced_from", rb_int_induced_from, 1);
 
-    rb_define_method(rb_cFixnum, "to_s", fix_to_s, 0);
-    rb_define_method(rb_cFixnum, "type", fix_type, 0);
+    rb_define_method(rb_cFixnum, "to_s", fix_to_s, -1);
 
     rb_define_method(rb_cFixnum, "id2name", fix_id2name, 0);
+    rb_define_method(rb_cFixnum, "to_sym", fix_to_sym, 0);
 
     rb_define_method(rb_cFixnum, "-@", fix_uminus, 0);
     rb_define_method(rb_cFixnum, "+", fix_plus, 1);
     rb_define_method(rb_cFixnum, "-", fix_minus, 1);
     rb_define_method(rb_cFixnum, "*", fix_mul, 1);
     rb_define_method(rb_cFixnum, "/", fix_div, 1);
+    rb_define_method(rb_cFixnum, "div", fix_div, 1);
     rb_define_method(rb_cFixnum, "%", fix_mod, 1);
     rb_define_method(rb_cFixnum, "modulo", fix_mod, 1);
     rb_define_method(rb_cFixnum, "divmod", fix_divmod, 1);
+    rb_define_method(rb_cFixnum, "quo", fix_quo, 1);
     rb_define_method(rb_cFixnum, "**", fix_pow, 1);
 
     rb_define_method(rb_cFixnum, "abs", fix_abs, 0);
@@ -1628,24 +1898,29 @@ Init_Numeric()
     rb_define_method(rb_cFixnum, ">>", fix_rshift, 1);
 
     rb_define_method(rb_cFixnum, "to_f", fix_to_f, 0);
-
-    rb_define_method(rb_cFixnum, "succ", fix_succ, 0);
-    rb_define_method(rb_cFixnum, "next", fix_succ, 0);
     rb_define_method(rb_cFixnum, "size", fix_size, 0);
-
-    rb_define_method(rb_cFixnum, "upto", fix_upto, 1);
-    rb_define_method(rb_cFixnum, "downto", fix_downto, 1);
-    rb_define_method(rb_cFixnum, "step", fix_step, 2);
-    rb_define_method(rb_cFixnum, "times", fix_dotimes, 0);
     rb_define_method(rb_cFixnum, "zero?", fix_zero_p, 0);
 
     rb_cFloat  = rb_define_class("Float", rb_cNumeric);
 
+    rb_undef_alloc_func(rb_cFloat);
     rb_undef_method(CLASS_OF(rb_cFloat), "new");
 
     rb_define_singleton_method(rb_cFloat, "induced_from", rb_flo_induced_from, 1);
     rb_include_module(rb_cFloat, rb_mPrecision);
 
+    rb_define_const(rb_cFloat, "ROUNDS", INT2FIX(FLT_ROUNDS));
+    rb_define_const(rb_cFloat, "RADIX", INT2FIX(FLT_RADIX));
+    rb_define_const(rb_cFloat, "MANT_DIG", INT2FIX(DBL_MANT_DIG));
+    rb_define_const(rb_cFloat, "DIG", INT2FIX(DBL_DIG));
+    rb_define_const(rb_cFloat, "MIN_EXP", INT2FIX(DBL_MIN_EXP));
+    rb_define_const(rb_cFloat, "MAX_EXP", INT2FIX(DBL_MAX_EXP));
+    rb_define_const(rb_cFloat, "MIN_10_EXP", INT2FIX(DBL_MIN_10_EXP));
+    rb_define_const(rb_cFloat, "MAX_10_EXP", INT2FIX(DBL_MAX_10_EXP));
+    rb_define_const(rb_cFloat, "MIN", rb_float_new(DBL_MIN));
+    rb_define_const(rb_cFloat, "MAX", rb_float_new(DBL_MAX));
+    rb_define_const(rb_cFloat, "EPSILON", rb_float_new(DBL_EPSILON));
+
     rb_define_method(rb_cFloat, "to_s", flo_to_s, 0);
     rb_define_method(rb_cFloat, "coerce", flo_coerce, 1);
     rb_define_method(rb_cFloat, "-@", flo_uminus, 0);
@@ -1670,6 +1945,7 @@ Init_Numeric()
     rb_define_method(rb_cFloat, "zero?", flo_zero_p, 0);
 
     rb_define_method(rb_cFloat, "to_i", flo_truncate, 0);
+    rb_define_method(rb_cFloat, "to_int", flo_truncate, 0);
     rb_define_method(rb_cFloat, "floor", flo_floor, 0);
     rb_define_method(rb_cFloat, "ceil", flo_ceil, 0);
     rb_define_method(rb_cFloat, "round", flo_round, 0);
