@@ -3,10 +3,10 @@
   pack.c -
 
   $Author: matz $
-  $Date: 2001/11/19 04:58:03 $
+  $Date: 2003/07/09 22:28:36 $
   created at: Thu Feb 10 15:17:05 JST 1994
 
-  Copyright (C) 1993-2000 Yukihiro Matsumoto
+  Copyright (C) 1993-2003 Yukihiro Matsumoto
 
 **********************************************************************/
 
@@ -57,10 +57,10 @@ TOKEN_PASTE(swap,x)(z)			\
     unsigned char *s, *t;		\
     int i;				\
 					\
-    zp = (xtype *)malloc(sizeof(xtype));\
+    zp = malloc(sizeof(xtype));		\
     *zp = z;				\
-    s = (char *)zp;			\
-    t = (char *)malloc(sizeof(xtype));	\
+    s = (unsigned char*)zp;		\
+    t = malloc(sizeof(xtype));		\
     for (i=0; i<sizeof(xtype); i++) {	\
 	t[sizeof(xtype)-i-1] = s[i];	\
     }					\
@@ -309,20 +309,72 @@ typedef long I32;
 typedef unsigned long U32;
 #define NUM2I32(x) NUM2LONG(x)
 #define NUM2U32(x) NUM2ULONG(x)
-#elif SIZEOF_INT == SIZE32
+#else
 typedef int I32;
 typedef unsigned int U32;
-#define NUM2I32(x) NUM2INT(x)
-#define NUM2U32(x) NUM2UINT(x)
+# if SIZEOF_INT == SIZE32
+#  define NUM2I32(x) NUM2INT(x)
+#  define NUM2U32(x) NUM2UINT(x)
+# else
+
+#define I32_MAX 2147483647
+#define I32_MIN (-I32_MAX-1)
+
+static I32
+num2i32(x)
+    VALUE x;
+{
+    long num = NUM2LONG(x);
+
+    if (num < I32_MIN || I32_MAX < num) {
+	rb_raise(rb_eRangeError, "integer %ld too big to convert to `I32'", num);
+    }
+    return (I32)num;
+}
+    
+#define U32_MAX 4294967295
+
+static U32
+num2u32(x)
+    VALUE x;
+{
+    unsigned long num = NUM2ULONG(x);
+
+    if (U32_MAX < num) {
+	rb_raise(rb_eRangeError, "integer %ld too big to convert to `U32'", num);
+    }
+    return (U32)num;
+}
+
+#  define NUM2I32(x) num2i32(x)
+#  define NUM2U32(x) num2u32(x)
+# endif
 #endif
 
+#if SIZEOF_LONG == SIZE32 || SIZEOF_INT == SIZE32
+# define EXTEND32(x) ((I32)(x))
+#else
+/* invariant in modulo 1<<31 */
+# define EXTEND32(x) (I32)(((1<<31)-1-(x))^~(~0<<31))
+#endif
+#if SIZEOF_SHORT == SIZE16
+# define EXTEND16(x) (short)(x)
+#else
+# define EXTEND16(x) (short)(((1<<15)-1-(x))^~(~0<<15))
+#endif
+
+#ifdef HAVE_LONG_LONG
+# define QUAD_SIZE sizeof(LONG_LONG)
+#else
+# define QUAD_SIZE 8
+#endif
 static char *toofew = "too few arguments";
 
-static void encodes _((VALUE,char*,int,int));
-static void qpencode _((VALUE,VALUE,int));
+static void encodes _((VALUE,char*,long,int));
+static void qpencode _((VALUE,VALUE,long));
 
 static int uv_to_utf8 _((char*,unsigned long));
-static unsigned long utf8_to_uv _((char*,int*));
+static unsigned long utf8_to_uv _((char*,long*));
 
 static VALUE
 pack_pack(ary, fmt)
@@ -331,22 +383,23 @@ pack_pack(ary, fmt)
     static char *nul10 = "\0\0\0\0\0\0\0\0\0\0";
     static char *spc10 = "          ";
     char *p, *pend;
-    VALUE res, from;
+    VALUE res, from, associates = 0;
     char type;
-    int items, len, idx;
+    long items, len, idx, plen;
     char *ptr;
-    int plen;
 #ifdef NATINT_PACK
     int natint;		/* native integer */
 #endif
-    
-    p = rb_str2cstr(fmt, &plen);
-    pend = p + plen;
-    res = rb_str_new(0, 0);
+
+    StringValue(fmt);
+    p = RSTRING(fmt)->ptr;
+    pend = p + RSTRING(fmt)->len;
+    res = rb_str_buf_new(0);
 
     items = RARRAY(ary)->len;
     idx = 0;
 
+#define THISFROM RARRAY(ary)->ptr[idx]
 #define NEXTFROM (items-- > 0 ? RARRAY(ary)->ptr[idx++] : (rb_raise(rb_eArgError, toofew),0))
 
     while (p < pend) {
@@ -356,6 +409,12 @@ pack_pack(ary, fmt)
 #endif
 
 	if (ISSPACE(type)) continue;
+	if (type == '#') {
+	    while ((p < pend) && (*p != '\n')) {
+		p++;
+	    }
+	    continue;
+	}
         if (*p == '_' || *p == '!') {
 	    char *natstr = "sSiIlL";
 
@@ -390,7 +449,10 @@ pack_pack(ary, fmt)
 		plen = 0;
 	    }
 	    else {
-		ptr = rb_str2cstr(from, &plen);
+		StringValue(from);
+		ptr = RSTRING(from)->ptr;
+		plen = RSTRING(from)->len;
+		OBJ_INFECT(res, from);
 	    }
 
 	    if (p[-1] == '*')
@@ -401,22 +463,22 @@ pack_pack(ary, fmt)
 	      case 'A':
 	      case 'Z':
 		if (plen >= len)
-		    rb_str_cat(res, ptr, len);
+		    rb_str_buf_cat(res, ptr, len);
 		else {
-		    rb_str_cat(res, ptr, plen);
+		    rb_str_buf_cat(res, ptr, plen);
 		    len -= plen;
 		    while (len >= 10) {
-			rb_str_cat(res, (type == 'A')?spc10:nul10, 10);
+			rb_str_buf_cat(res, (type == 'A')?spc10:nul10, 10);
 			len -= 10;
 		    }
-		    rb_str_cat(res, (type == 'A')?spc10:nul10, len);
+		    rb_str_buf_cat(res, (type == 'A')?spc10:nul10, len);
 		}
 		break;
 
 	      case 'b':
 		{
 		    int byte = 0;
-		    int i, j = 0;
+		    long i, j = 0;
 
 		    if (len > plen) {
 			j = (len - plen + 1)/2;
@@ -429,7 +491,7 @@ pack_pack(ary, fmt)
 			    byte >>= 1;
 			else {
 			    char c = byte & 0xff;
-			    rb_str_cat(res, &c, 1);
+			    rb_str_buf_cat(res, &c, 1);
 			    byte = 0;
 			}
 		    }
@@ -437,18 +499,17 @@ pack_pack(ary, fmt)
 			char c;
 			byte >>= 7 - (len & 7);
 			c = byte & 0xff;
-			rb_str_cat(res, &c, 1);
+			rb_str_buf_cat(res, &c, 1);
 		    }
-		    len = RSTRING(res)->len;
-		    rb_str_resize(res, len+j);
-		    MEMZERO(RSTRING(res)->ptr+len, char, j);
+		    len = j;
+		    goto grow;
 		}
 		break;
 
 	      case 'B':
 		{
 		    int byte = 0;
-		    int i, j = 0;
+		    long i, j = 0;
 
 		    if (len > plen) {
 			j = (len - plen + 1)/2;
@@ -460,7 +521,7 @@ pack_pack(ary, fmt)
 			    byte <<= 1;
 			else {
 			    char c = byte & 0xff;
-			    rb_str_cat(res, &c, 1);
+			    rb_str_buf_cat(res, &c, 1);
 			    byte = 0;
 			}
 		    }
@@ -468,18 +529,17 @@ pack_pack(ary, fmt)
 			char c;
 			byte <<= 7 - (len & 7);
 			c = byte & 0xff;
-			rb_str_cat(res, &c, 1);
+			rb_str_buf_cat(res, &c, 1);
 		    }
-		    len = RSTRING(res)->len;
-		    rb_str_resize(res, len+j);
-		    MEMZERO(RSTRING(res)->ptr+len, char, j);
+		    len = j;
+		    goto grow;
 		}
 		break;
 
 	      case 'h':
 		{
 		    int byte = 0;
-		    int i, j = 0;
+		    long i, j = 0;
 
 		    if (len > plen) {
 			j = (len - plen + 1)/2;
@@ -494,24 +554,23 @@ pack_pack(ary, fmt)
 			    byte >>= 4;
 			else {
 			    char c = byte & 0xff;
-			    rb_str_cat(res, &c, 1);
+			    rb_str_buf_cat(res, &c, 1);
 			    byte = 0;
 			}
 		    }
 		    if (len & 1) {
 			char c = byte & 0xff;
-			rb_str_cat(res, &c, 1);
+			rb_str_buf_cat(res, &c, 1);
 		    }
-		    len = RSTRING(res)->len;
-		    rb_str_resize(res, len+j);
-		    MEMZERO(RSTRING(res)->ptr+len, char, j);
+		    len = j;
+		    goto grow;
 		}
 		break;
 
 	      case 'H':
 		{
 		    int byte = 0;
-		    int i, j = 0;
+		    long i, j = 0;
 
 		    if (len > plen) {
 			j = (len - plen + 1)/2;
@@ -526,17 +585,16 @@ pack_pack(ary, fmt)
 			    byte <<= 4;
 			else {
 			    char c = byte & 0xff;
-			    rb_str_cat(res, &c, 1);
+			    rb_str_buf_cat(res, &c, 1);
 			    byte = 0;
 			}
 		    }
 		    if (len & 1) {
 			char c = byte & 0xff;
-			rb_str_cat(res, &c, 1);
+			rb_str_buf_cat(res, &c, 1);
 		    }
-		    len = RSTRING(res)->len;
-		    rb_str_resize(res, len+j);
-		    MEMZERO(RSTRING(res)->ptr+len, char, j);
+		    len = j;
+		    goto grow;
 		}
 		break;
 	    }
@@ -552,7 +610,7 @@ pack_pack(ary, fmt)
 		else {
 		    c = NUM2INT(from);
 		}
-		rb_str_cat(res, &c, sizeof(char));
+		rb_str_buf_cat(res, &c, sizeof(char));
 	    }
 	    break;
 
@@ -566,7 +624,7 @@ pack_pack(ary, fmt)
 		else {
 		    s = NUM2INT(from);
 		}
-		rb_str_cat(res, OFF16(&s), NATINT_LEN(short,2));
+		rb_str_buf_cat(res, OFF16(&s), NATINT_LEN(short,2));
 	    }
 	    break;
 
@@ -580,7 +638,7 @@ pack_pack(ary, fmt)
 		else {
 		    i = NUM2UINT(from);
 		}
-		rb_str_cat(res, (char*)&i, sizeof(int));
+		rb_str_buf_cat(res, (char*)&i, sizeof(int));
 	    }
 	    break;
 
@@ -594,7 +652,19 @@ pack_pack(ary, fmt)
 		else {
 		    l = NATINT_U32(from);
 		}
-		rb_str_cat(res, OFF32(&l), NATINT_LEN(long,4));
+		rb_str_buf_cat(res, OFF32(&l), NATINT_LEN(long,4));
+	    }
+	    break;
+
+	  case 'q':
+	  case 'Q':
+	    while (len-- > 0) {
+		char tmp[QUAD_SIZE];
+
+		from = NEXTFROM;
+		if (NIL_P(from)) from = INT2FIX(0);
+		rb_quad_pack(tmp, from);
+		rb_str_buf_cat(res, (char*)&tmp, QUAD_SIZE);
 	    }
 	    break;
 
@@ -608,7 +678,7 @@ pack_pack(ary, fmt)
 		    s = NUM2INT(from);
 		}
 		s = htons(s);
-		rb_str_cat(res, OFF16B(&s), NATINT_LEN(short,2));
+		rb_str_buf_cat(res, OFF16B(&s), NATINT_LEN(short,2));
 	    }
 	    break;
 
@@ -622,7 +692,7 @@ pack_pack(ary, fmt)
 		    l = NATINT_U32(from);
 		}
 		l = htonl(l);
-		rb_str_cat(res, OFF32B(&l), NATINT_LEN(long,4));
+		rb_str_buf_cat(res, OFF32B(&l), NATINT_LEN(long,4));
 	    }
 	    break;
 
@@ -636,7 +706,7 @@ pack_pack(ary, fmt)
 		    s = NUM2INT(from);
 		}
 		s = htovs(s);
-		rb_str_cat(res, OFF16(&s), NATINT_LEN(short,2));
+		rb_str_buf_cat(res, OFF16(&s), NATINT_LEN(short,2));
 	    }
 	    break;
 
@@ -650,7 +720,7 @@ pack_pack(ary, fmt)
 		    l = NATINT_U32(from);
 		}
 		l = htovl(l);
-		rb_str_cat(res, OFF32(&l), NATINT_LEN(long,4));
+		rb_str_buf_cat(res, OFF32(&l), NATINT_LEN(long,4));
 	    }
 	    break;
 
@@ -660,17 +730,8 @@ pack_pack(ary, fmt)
 		float f;
 
 		from = NEXTFROM;
-		switch (TYPE(from)) {
-		  case T_FLOAT:
-		    f = RFLOAT(from)->value;
-		    break;
-		  case T_STRING:
-		    f = strtod(RSTRING(from)->ptr, 0);
-		  default:
-		    f = (float)NUM2INT(from);
-		    break;
-		}
-		rb_str_cat(res, (char*)&f, sizeof(float));
+		f = RFLOAT(rb_Float(from))->value;
+		rb_str_buf_cat(res, (char*)&f, sizeof(float));
 	    }
 	    break;
 
@@ -680,18 +741,9 @@ pack_pack(ary, fmt)
 		FLOAT_CONVWITH(ftmp);
 
 		from = NEXTFROM;
-		switch (TYPE(from)) {
-		  case T_FLOAT:
-		    f = RFLOAT(from)->value;
-		    break;
-		  case T_STRING:
-		    f = strtod(RSTRING(from)->ptr, 0);
-		  default:
-		    f = (float)NUM2INT(from);
-		    break;
-		}
+		f = RFLOAT(rb_Float(from))->value;
 		f = HTOVF(f,ftmp);
-		rb_str_cat(res, (char*)&f, sizeof(float));
+		rb_str_buf_cat(res, (char*)&f, sizeof(float));
 	    }
 	    break;
 
@@ -701,18 +753,9 @@ pack_pack(ary, fmt)
 		DOUBLE_CONVWITH(dtmp);
 
 		from = NEXTFROM;
-		switch (TYPE(from)) {
-		  case T_FLOAT:
-		    d = RFLOAT(from)->value;
-		    break;
-		  case T_STRING:
-		    d = strtod(RSTRING(from)->ptr, 0);
-		  default:
-		    d = (double)NUM2INT(from);
-		    break;
-		}
+		d = RFLOAT(rb_Float(from))->value;
 		d = HTOVD(d,dtmp);
-		rb_str_cat(res, (char*)&d, sizeof(double));
+		rb_str_buf_cat(res, (char*)&d, sizeof(double));
 	    }
 	    break;
 
@@ -722,17 +765,8 @@ pack_pack(ary, fmt)
 		double d;
 
 		from = NEXTFROM;
-		switch (TYPE(from)) {
-		  case T_FLOAT:
-		    d = RFLOAT(from)->value;
-		    break;
-		  case T_STRING:
-		    d = strtod(RSTRING(from)->ptr, 0);
-		  default:
-		    d = (double)NUM2INT(from);
-		    break;
-		}
-		rb_str_cat(res, (char*)&d, sizeof(double));
+		d = RFLOAT(rb_Float(from))->value;
+		rb_str_buf_cat(res, (char*)&d, sizeof(double));
 	    }
 	    break;
 
@@ -742,18 +776,9 @@ pack_pack(ary, fmt)
 		FLOAT_CONVWITH(ftmp);
 
 		from = NEXTFROM;
-		switch (TYPE(from)) {
-		  case T_FLOAT:
-		    f = RFLOAT(from)->value;
-		    break;
-		  case T_STRING:
-		    f = strtod(RSTRING(from)->ptr, 0);
-		  default:
-		    f = (float)NUM2INT(from);
-		    break;
-		}
+		f = RFLOAT(rb_Float(from))->value;
 		f = HTONF(f,ftmp);
-		rb_str_cat(res, (char*)&f, sizeof(float));
+		rb_str_buf_cat(res, (char*)&f, sizeof(float));
 	    }
 	    break;
 
@@ -763,36 +788,28 @@ pack_pack(ary, fmt)
 		DOUBLE_CONVWITH(dtmp);
 
 		from = NEXTFROM;
-		switch (TYPE(from)) {
-		  case T_FLOAT:
-		    d = RFLOAT(from)->value;
-		    break;
-		  case T_STRING:
-		    d = strtod(RSTRING(from)->ptr, 0);
-		  default:
-		    d = (double)NUM2INT(from);
-		    break;
-		}
+		d = RFLOAT(rb_Float(from))->value;
 		d = HTOND(d,dtmp);
-		rb_str_cat(res, (char*)&d, sizeof(double));
+		rb_str_buf_cat(res, (char*)&d, sizeof(double));
 	    }
 	    break;
 
 	  case 'x':
 	  grow:
 	    while (len >= 10) {
-		rb_str_cat(res, nul10, 10);
+		rb_str_buf_cat(res, nul10, 10);
 		len -= 10;
 	    }
-	    rb_str_cat(res, nul10, len);
+	    rb_str_buf_cat(res, nul10, len);
 	    break;
 
 	  case 'X':
 	  shrink:
-	    if (RSTRING(res)->len < len)
+	    plen = RSTRING(res)->len;
+	    if (plen < len)
 		rb_raise(rb_eArgError, "X outside of string");
-	    RSTRING(res)->len -= len;
-	    RSTRING(res)->ptr[RSTRING(res)->len] = '\0';
+	    RSTRING(res)->len = plen - len;
+	    RSTRING(res)->ptr[plen - len] = '\0';
 	    break;
 
 	  case '@':
@@ -818,20 +835,23 @@ pack_pack(ary, fmt)
 		    l = NUM2ULONG(from);
 		}
 		le = uv_to_utf8(buf, l);
-		rb_str_cat(res, (char*)buf, le);
+		rb_str_buf_cat(res, (char*)buf, le);
 	    }
 	    break;
 
 	  case 'u':
 	  case 'm':
-	    ptr = rb_str2cstr(NEXTFROM, &plen);
+	    from = NEXTFROM;
+	    StringValue(from);
+	    ptr = RSTRING(from)->ptr;
+	    plen = RSTRING(from)->len;
 
 	    if (len <= 2)
 		len = 45;
 	    else
 		len = len / 3 * 3;
 	    while (plen > 0) {
-		int todo;
+		long todo;
 
 		if (plen > len)
 		    todo = len;
@@ -851,18 +871,31 @@ pack_pack(ary, fmt)
 	    break;
 
 	  case 'P':
+	    from = THISFROM;
+	    if (!NIL_P(from)) {
+		StringValue(from);
+		if (RSTRING(from)->len < len) {
+		    rb_raise(rb_eArgError, "too short buffer for P(%ld for %ld)",
+			     RSTRING(from)->len, len);
+		}
+	    }
 	    len = 1;
 	    /* FALL THROUGH */
 	  case 'p':
 	    while (len-- > 0) {
 		char *t;
 		from = NEXTFROM;
-		if (NIL_P(from)) t = "";
-		else {
-		    t = STR2CSTR(from);
-		    rb_str_associate(res, from);
+		if (NIL_P(from)) {
+		    t = 0;
 		}
-		rb_str_cat(res, (char*)&t, sizeof(char*));
+		else {
+		    t = StringValuePtr(from);
+		}
+		if (!associates) {
+		    associates = rb_ary_new();
+		}
+		rb_ary_push(associates, from);
+		rb_str_buf_cat(res, (char*)&t, sizeof(char*));
 	    }
 	    break;
 
@@ -873,13 +906,12 @@ pack_pack(ary, fmt)
 		char c, *bufs, *bufe;
 
 		from = NEXTFROM;
-
 		if (TYPE(from) == T_BIGNUM) {
 		    VALUE big128 = rb_uint2big(128);
 		    while (TYPE(from) == T_BIGNUM) {
 			from = rb_big_divmod(from, big128);
 			c = NUM2INT(RARRAY(from)->ptr[1]) | 0x80; /* mod */
-			rb_str_cat(buf, &c, sizeof(char));
+			rb_str_buf_cat(buf, &c, sizeof(char));
 			from = RARRAY(from)->ptr[0]; /* div */
 		    }
 		}
@@ -891,7 +923,7 @@ pack_pack(ary, fmt)
 
 		while (ul) {
 		    c = ((ul & 0x7f) | 0x80);
-		    rb_str_cat(buf, &c, sizeof(char));
+		    rb_str_buf_cat(buf, &c, sizeof(char));
 		    ul >>=  7;
 		}
 
@@ -904,11 +936,11 @@ pack_pack(ary, fmt)
 			*bufs++ = *bufe;
 			*bufe-- = c;
 		    }
-		    rb_str_cat(res, RSTRING(buf)->ptr, RSTRING(buf)->len);
+		    rb_str_buf_cat(res, RSTRING(buf)->ptr, RSTRING(buf)->len);
 		}
 		else {
 		    c = 0;
-		    rb_str_cat(res, &c, sizeof(char));
+		    rb_str_buf_cat(res, &c, sizeof(char));
 		}
 	    }
 	    break;
@@ -918,6 +950,9 @@ pack_pack(ary, fmt)
 	}
     }
 
+    if (associates) {
+	rb_str_associate(res, associates);
+    }
     return res;
 }
 
@@ -930,11 +965,11 @@ static void
 encodes(str, s, len, type)
     VALUE str;
     char *s;
-    int len;
+    long len;
     int type;
 {
     char *buff = ALLOCA_N(char, len * 4 / 3 + 6);
-    int i = 0;
+    long i = 0;
     char *trans = type == 'u' ? uu_table : b64_table;
     int padding;
 
@@ -966,7 +1001,7 @@ encodes(str, s, len, type)
 	buff[i++] = padding;
     }
     buff[i++] = '\n';
-    rb_str_cat(str, buff, i);
+    rb_str_buf_cat(str, buff, i);
 }
 
 static char hex_table[] = "0123456789ABCDEF";
@@ -974,10 +1009,10 @@ static char hex_table[] = "0123456789ABCDEF";
 static void
 qpencode(str, from, len)
     VALUE str, from;
-    int len;
+    long len;
 {
     char buff[1024];
-    int i = 0, n = 0, prev = EOF;
+    long i = 0, n = 0, prev = EOF;
     unsigned char *s = (unsigned char*)RSTRING(from)->ptr;
     unsigned char *send = s + RSTRING(from)->len;
 
@@ -1012,7 +1047,7 @@ qpencode(str, from, len)
             prev = '\n';
         }
 	if (i > 1024 - 5) {
-	    rb_str_cat(str, buff, i);
+	    rb_str_buf_cat(str, buff, i);
 	    i = 0;
 	}
 	s++;
@@ -1022,7 +1057,7 @@ qpencode(str, from, len)
 	buff[i++] = '\n';
     }
     if (i > 0) {
-	rb_str_cat(str, buff, i);
+	rb_str_buf_cat(str, buff, i);
     }
 }
 
@@ -1045,31 +1080,40 @@ hex2num(c)
     }
 }
 
+#define PACK_LENGTH_ADJUST_SIZE(sz) do {	\
+    tmp = 0;					\
+    if (len > (send-s)/sz) {			\
+        if (!star) {				\
+	    tmp = len-(send-s)/sz;		\
+        }					\
+	len = (send-s)/sz;			\
+    }						\
+} while (0)
+
 #ifdef NATINT_PACK
 #define PACK_LENGTH_ADJUST(type,sz) do {	\
     int t__len = NATINT_LEN(type,(sz));		\
-    tmp = 0;					\
-    if (len > (send-s)/t__len) {		\
-        if (!star) {				\
-	    tmp = len-(send-s)/t__len;		\
-        }					\
-	len = (send-s)/t__len;			\
-    }						\
+    PACK_LENGTH_ADJUST_SIZE(t__len);		\
 } while (0)
 #else
-#define PACK_LENGTH_ADJUST(type,sz) do {	\
-    tmp = 0;					\
-    if (len > (send-s)/sizeof(type)) {		\
-        if (!star) {				\
-	    tmp = len - (send-s)/sizeof(type);	\
-        }					\
-	len = (send-s)/sizeof(type);		\
-    }						\
-} while (0)
+#define PACK_LENGTH_ADJUST(type,sz) 		\
+    PACK_LENGTH_ADJUST_SIZE(sizeof(type))
 #endif
 
-#define PACK_ITEM_ADJUST() while (tmp--) rb_ary_push(ary, Qnil);
+#define PACK_ITEM_ADJUST() while (tmp--) rb_ary_push(ary, Qnil)
 
+static VALUE
+infected_str_new(ptr, len, str)
+    const char *ptr;
+    long len;
+    VALUE str;
+{
+    VALUE s = rb_str_new(ptr, len);
+
+    OBJ_INFECT(s, str);
+    return s;
+}
+    
 static VALUE
 pack_unpack(str, fmt)
     VALUE str, fmt;
@@ -1079,23 +1123,34 @@ pack_unpack(str, fmt)
     char *p, *pend;
     VALUE ary;
     char type;
-    int len, tmp, star;
+    long len;
+    int tmp, star;
 #ifdef NATINT_PACK
     int natint;			/* native integer */
 #endif
 
-    s = rb_str2cstr(str, &len);
-    send = s + len;
-    p = rb_str2cstr(fmt, &len);
-    pend = p + len;
+    StringValue(str);
+    s = RSTRING(str)->ptr;
+    send = s + RSTRING(str)->len;
+    StringValue(fmt);
+    p = RSTRING(fmt)->ptr;
+    pend = p + RSTRING(fmt)->len;
 
     ary = rb_ary_new();
     while (p < pend) {
+	type = *p++;
 #ifdef NATINT_PACK
 	natint = 0;
 #endif
+
+	if (ISSPACE(type)) continue;
+	if (type == '#') {
+	    while ((p < pend) && (*p != '\n')) {
+		p++;
+	    }
+	    continue;
+	}
 	star = 0;
-	type = *p++;
 	if (*p == '_' || *p == '!') {
 	    char *natstr = "sSiIlL";
 
@@ -1131,14 +1186,14 @@ pack_unpack(str, fmt)
 	  case 'A':
 	    if (len > send - s) len = send - s;
 	    {
-		int end = len;
+		long end = len;
 		char *t = s + len - 1;
 
 		while (t >= s) {
 		    if (*t != ' ' && *t != '\0') break;
 		    t--; len--;
 		}
-		rb_ary_push(ary, rb_str_new(s, len));
+		rb_ary_push(ary, infected_str_new(s, len, str));
 		s += end;
 	    }
 	    break;
@@ -1146,21 +1201,21 @@ pack_unpack(str, fmt)
 	  case 'Z':
 	    if (len > send - s) len = send - s;
 	    {
-		int end = len;
+		long end = len;
 		char *t = s + len - 1;
 
 		while (t >= s) {
 		    if (*t) break;
 		    t--; len--;
 		}
-		rb_ary_push(ary, rb_str_new(s, len));
+		rb_ary_push(ary, infected_str_new(s, len, str));
 		s += end;
 	    }
 	    break;
 
 	  case 'a':
 	    if (len > send - s) len = send - s;
-	    rb_ary_push(ary, rb_str_new(s, len));
+	    rb_ary_push(ary, infected_str_new(s, len, str));
 	    s += len;
 	    break;
 
@@ -1169,7 +1224,8 @@ pack_unpack(str, fmt)
 	    {
 		VALUE bitstr;
 		char *t;
-		int bits, i;
+		int bits;
+		long i;
 
 		if (p[-1] == '*' || len > (send - s) * 8)
 		    len = (send - s) * 8;
@@ -1188,7 +1244,8 @@ pack_unpack(str, fmt)
 	    {
 		VALUE bitstr;
 		char *t;
-		int bits, i;
+		int bits;
+		long i;
 
 		if (p[-1] == '*' || len > (send - s) * 8)
 		    len = (send - s) * 8;
@@ -1207,7 +1264,8 @@ pack_unpack(str, fmt)
 	    {
 		VALUE bitstr;
 		char *t;
-		int bits, i;
+		int bits;
+		long i;
 
 		if (p[-1] == '*' || len > (send - s) * 2)
 		    len = (send - s) * 2;
@@ -1228,7 +1286,8 @@ pack_unpack(str, fmt)
 	    {
 		VALUE bitstr;
 		char *t;
-		int bits, i;
+		int bits;
+		long i;
 
 		if (p[-1] == '*' || len > (send - s) * 2)
 		    len = (send - s) * 2;
@@ -1269,6 +1328,9 @@ pack_unpack(str, fmt)
 	    while (len-- > 0) {
 		short tmp = 0;
 		memcpy(OFF16(&tmp), s, NATINT_LEN(short,2));
+#if SIZEOF_SHORT != SIZE16
+		if (!natint) tmp = EXTEND16(tmp);
+#endif
 		s += NATINT_LEN(short,2);
 		rb_ary_push(ary, INT2FIX(tmp));
 	    }
@@ -1292,7 +1354,7 @@ pack_unpack(str, fmt)
 		int tmp;
 		memcpy(&tmp, s, sizeof(int));
 		s += sizeof(int);
-		rb_ary_push(ary, rb_int2inum(tmp));
+		rb_ary_push(ary, INT2NUM(tmp));
 	    }
 	    PACK_ITEM_ADJUST();
 	    break;
@@ -1303,7 +1365,7 @@ pack_unpack(str, fmt)
 		unsigned int tmp;
 		memcpy(&tmp, s, sizeof(unsigned int));
 		s += sizeof(unsigned int);
-		rb_ary_push(ary, rb_uint2inum(tmp));
+		rb_ary_push(ary, UINT2NUM(tmp));
 	    }
 	    PACK_ITEM_ADJUST();
 	    break;
@@ -1313,8 +1375,11 @@ pack_unpack(str, fmt)
 	    while (len-- > 0) {
 		long tmp = 0;
 		memcpy(OFF32(&tmp), s, NATINT_LEN(long,4));
+#if SIZEOF_LONG != SIZE32
+		if (!natint) tmp = EXTEND32(tmp);
+#endif
 		s += NATINT_LEN(long,4);
-		rb_ary_push(ary, rb_int2inum(tmp));
+		rb_ary_push(ary, LONG2NUM(tmp));
 	    }
 	    PACK_ITEM_ADJUST();
 	    break;
@@ -1325,18 +1390,36 @@ pack_unpack(str, fmt)
 		unsigned long tmp = 0;
 		memcpy(OFF32(&tmp), s, NATINT_LEN(unsigned long,4));
 		s += NATINT_LEN(unsigned long,4);
-		rb_ary_push(ary, rb_uint2inum(tmp));
+		rb_ary_push(ary, ULONG2NUM(tmp));
 	    }
 	    PACK_ITEM_ADJUST();
 	    break;
 
+	  case 'q':
+	    PACK_LENGTH_ADJUST_SIZE(QUAD_SIZE);
+	    while (len-- > 0) {
+		char *tmp = (char*)s;
+		s += QUAD_SIZE;
+		rb_ary_push(ary, rb_quad_unpack(tmp, 1));
+	    }
+	    PACK_ITEM_ADJUST();
+	    break;
+	case 'Q':
+	    PACK_LENGTH_ADJUST_SIZE(QUAD_SIZE);
+	    while (len-- > 0) {
+		char *tmp = (char*)s;
+		s += QUAD_SIZE;
+		rb_ary_push(ary, rb_quad_unpack(tmp, 0));
+	    }
+	    break;
+
 	  case 'n':
 	    PACK_LENGTH_ADJUST(unsigned short,2);
 	    while (len-- > 0) {
 		unsigned short tmp = 0;
 		memcpy(OFF16B(&tmp), s, NATINT_LEN(unsigned short,2));
 		s += NATINT_LEN(unsigned short,2);
-		rb_ary_push(ary, rb_uint2inum(ntohs(tmp)));
+		rb_ary_push(ary, UINT2NUM(ntohs(tmp)));
 	    }
 	    PACK_ITEM_ADJUST();
 	    break;
@@ -1347,7 +1430,7 @@ pack_unpack(str, fmt)
 		unsigned long tmp = 0;
 		memcpy(OFF32B(&tmp), s, NATINT_LEN(unsigned long,4));
 		s += NATINT_LEN(unsigned long,4);
-		rb_ary_push(ary, rb_uint2inum(ntohl(tmp)));
+		rb_ary_push(ary, ULONG2NUM(ntohl(tmp)));
 	    }
 	    PACK_ITEM_ADJUST();
 	    break;
@@ -1358,7 +1441,7 @@ pack_unpack(str, fmt)
 		unsigned short tmp = 0;
 		memcpy(OFF16(&tmp), s, NATINT_LEN(unsigned short,2));
 		s += NATINT_LEN(unsigned short,2);
-		rb_ary_push(ary, rb_uint2inum(vtohs(tmp)));
+		rb_ary_push(ary, UINT2NUM(vtohs(tmp)));
 	    }
 	    PACK_ITEM_ADJUST();
 	    break;
@@ -1369,7 +1452,7 @@ pack_unpack(str, fmt)
 		unsigned long tmp = 0;
 		memcpy(OFF32(&tmp), s, NATINT_LEN(long,4));
 		s += NATINT_LEN(long,4);
-		rb_ary_push(ary, rb_uint2inum(vtohl(tmp)));
+		rb_ary_push(ary, ULONG2NUM(vtohl(tmp)));
 	    }
 	    PACK_ITEM_ADJUST();
 	    break;
@@ -1457,20 +1540,20 @@ pack_unpack(str, fmt)
 	  case 'U':
 	    if (len > send - s) len = send - s;
 	    while (len > 0 && s < send) {
-		int alen = len;
+		long alen = send - s;
 		unsigned long l;
 
 		l = utf8_to_uv(s, &alen);
-		s += alen; len -= alen;
-		rb_ary_push(ary, rb_uint2inum(l));
+		s += alen; len--;
+		rb_ary_push(ary, ULONG2NUM(l));
 	    }
 	    break;
 
 	  case 'u':
 	    {
-		VALUE str = rb_str_new(0, (send - s)*3/4);
-		char *ptr = RSTRING(str)->ptr;
-		int total = 0;
+		VALUE buf = infected_str_new(0, (send - s)*3/4, str);
+		char *ptr = RSTRING(buf)->ptr;
+		long total = 0;
 
 		while (s < send && *s > ' ' && *s < 'a') {
 		    long a,b,c,d;
@@ -1479,13 +1562,13 @@ pack_unpack(str, fmt)
 		    hunk[3] = '\0';
 		    len = (*s++ - ' ') & 077;
 		    total += len;
-		    if (total > RSTRING(str)->len) {
-			len -= total - RSTRING(str)->len;
-			total = RSTRING(str)->len;
+		    if (total > RSTRING(buf)->len) {
+			len -= total - RSTRING(buf)->len;
+			total = RSTRING(buf)->len;
 		    }
 
 		    while (len > 0) {
-			int mlen = len > 3 ? 3 : len;
+			long mlen = len > 3 ? 3 : len;
 
 			if (s < send && *s >= ' ')
 			    a = (*s++ - ' ') & 077;
@@ -1516,17 +1599,17 @@ pack_unpack(str, fmt)
 			s += 2;	/* possible checksum byte */
 		}
 		
-		RSTRING(str)->ptr[total] = '\0';
-		RSTRING(str)->len = total;
-		rb_ary_push(ary, str);
+		RSTRING(buf)->ptr[total] = '\0';
+		RSTRING(buf)->len = total;
+		rb_ary_push(ary, buf);
 	    }
 	    break;
 
 	  case 'm':
 	    {
-		VALUE str = rb_str_new(0, (send - s)*3/4);
-		char *ptr = RSTRING(str)->ptr;
-		int a,b,c,d;
+		VALUE buf = infected_str_new(0, (send - s)*3/4, str);
+		char *ptr = RSTRING(buf)->ptr;
+		int a = -1,b = -1,c = 0,d;
 		static int first = 1;
 		static int b64_xtable[256];
 
@@ -1541,7 +1624,7 @@ pack_unpack(str, fmt)
 			b64_xtable[(int)b64_table[i]] = i;
 		    }
 		}
-		for (;;) {
+		while (s < send) {
 		    while (s[0] == '\r' || s[0] == '\n') { s++; }
 		    if ((a = b64_xtable[(int)s[0]]) == -1) break;
 		    if ((b = b64_xtable[(int)s[1]]) == -1) break;
@@ -1552,23 +1635,24 @@ pack_unpack(str, fmt)
 		    *ptr++ = c << 6 | d;
 		    s += 4;
 		}
-		if (a != -1 && b != -1 && s[2] == '=') {
-		    *ptr++ = a << 2 | b >> 4;
-		}
-		if (a != -1 && b != -1 && c != -1 && s[3] == '=') {
-		    *ptr++ = a << 2 | b >> 4;
-		    *ptr++ = b << 4 | c >> 2;
+		if (a != -1 && b != -1) {
+		    if (s + 2 < send && s[2] == '=')
+			*ptr++ = a << 2 | b >> 4;
+		    if (c != -1 && s + 3 < send && s[3] == '=') {
+			*ptr++ = a << 2 | b >> 4;
+			*ptr++ = b << 4 | c >> 2;
+		    }
 		}
 		*ptr = '\0';
-		RSTRING(str)->len = ptr - RSTRING(str)->ptr;
-		rb_ary_push(ary, str);
+		RSTRING(buf)->len = ptr - RSTRING(buf)->ptr;
+		rb_ary_push(ary, buf);
 	    }
 	    break;
 
 	  case 'M':
 	    {
-		VALUE str = rb_str_new(0, send - s);
-		char *ptr = RSTRING(str)->ptr;
+		VALUE buf = infected_str_new(0, send - s, str);
+		char *ptr = RSTRING(buf)->ptr;
 		int c1, c2;
 
 		while (s < send) {
@@ -1587,12 +1671,14 @@ pack_unpack(str, fmt)
 		    s++;
 		}
 		*ptr = '\0';
-		RSTRING(str)->len = ptr - RSTRING(str)->ptr;
-		rb_ary_push(ary, str);
+		RSTRING(buf)->len = ptr - RSTRING(buf)->ptr;
+		rb_ary_push(ary, buf);
 	    }
 	    break;
 
 	  case '@':
+	    if (len > RSTRING(str)->len)
+		rb_raise(rb_eArgError, "@ outside of string");
 	    s = RSTRING(str)->ptr + len;
 	    break;
 
@@ -1611,12 +1697,37 @@ pack_unpack(str, fmt)
 	  case 'P':
 	    if (sizeof(char *) <= send - s) {
 		char *t;
-		VALUE str = rb_str_new(0, 0);
+		VALUE tmp;
+
 		memcpy(&t, s, sizeof(char *));
 		s += sizeof(char *);
-		if (t)
-		    rb_str_cat(str, t, len);
-		rb_ary_push(ary, str);
+
+		if (t) {
+		    VALUE a, *p, *pend;
+
+		    if (!(a = rb_str_associated(str))) {
+			rb_raise(rb_eArgError, "no associated pointer");
+		    }
+		    p = RARRAY(a)->ptr;
+		    pend = p + RARRAY(a)->len;
+		    while (p < pend) {
+			if (TYPE(*p) == T_STRING && RSTRING(*p)->ptr == t) {
+			    if (len > RSTRING(*p)->len) {
+				len = RSTRING(*p)->len;
+			    }
+			    break;
+			}
+			p++;
+		    }
+		    if (p == pend) {
+			rb_raise(rb_eArgError, "non associated pointer");
+		    }
+		    tmp = rb_tainted_str_new(t, len);
+		}
+		else {
+		    tmp = Qnil;
+		}
+		rb_ary_push(ary, tmp);
 	    }
 	    break;
 
@@ -1627,14 +1738,36 @@ pack_unpack(str, fmt)
 		if (send - s < sizeof(char *))
 		    break;
 		else {
+		    VALUE tmp;
 		    char *t;
-		    VALUE str = rb_str_new(0, 0);
+
 		    memcpy(&t, s, sizeof(char *));
 		    s += sizeof(char *);
+
 		    if (t) {
-			rb_str_cat2(str, t);
+			VALUE a, *p, *pend;
+
+			if (!(a = rb_str_associated(str))) {
+			    rb_raise(rb_eArgError, "no associated pointer");
+			}
+			p = RARRAY(a)->ptr;
+			pend = p + RARRAY(a)->len;
+			while (p < pend) {
+			    if (TYPE(*p) == T_STRING && RSTRING(*p)->ptr == t) {
+				break;
+			    }
+			    p++;
+			}
+			if (p == pend) {
+			    rb_raise(rb_eArgError, "non associated pointer");
+			}
+			tmp = rb_str_new2(t);
+			OBJ_INFECT(tmp, str);
 		    }
-		    rb_ary_push(ary, str);
+		    else {
+			tmp = Qnil;
+		    }
+		    rb_ary_push(ary, tmp);
 		}
 	    }
 	    break;
@@ -1648,7 +1781,7 @@ pack_unpack(str, fmt)
 		    ul <<= 7;
 		    ul |= (*s & 0x7f);
 		    if (!(*s++ & 0x80)) {
-			rb_ary_push(ary, rb_uint2inum(ul));
+			rb_ary_push(ary, ULONG2NUM(ul));
 			len--;
 			ul = 0;
 		    }
@@ -1724,49 +1857,68 @@ uv_to_utf8(buf, uv)
 	buf[5] = (uv&0x3f)|0x80;
 	return 6;
     }
-#if SIZEOF_LONG > 4
-    if (uv <= 0xfffffffff) {
-#endif
-	buf[0] = 0xfe;
-	buf[1] = ((uv>>30)&0x3f)|0x80;
-	buf[2] = ((uv>>24)&0x3f)|0x80;
-	buf[3] = ((uv>>18)&0x3f)|0x80;
-	buf[4] = ((uv>>12)&0x3f)|0x80;
-	buf[5] = ((uv>>6)&0x3f)|0x80;
-	buf[6] = (uv&0x3f)|0x80;
-	return 7;
-#if SIZEOF_LONG > 4
-    }
     rb_raise(rb_eArgError, "uv_to_utf8(); too big value");
-#endif
 }
 
+static const long utf8_limits[] = {
+    0x0,			/* 1 */
+    0x80,			/* 2 */
+    0x800,			/* 3 */
+    0x10000,			/* 4 */
+    0x200000,			/* 5 */
+    0x4000000,			/* 6 */
+    0x80000000,			/* 7 */
+};
+
 static unsigned long
 utf8_to_uv(p, lenp)
     char *p;
-    int *lenp;
+    long *lenp;
 {
-    int c = (*p++)&0xff;
-    unsigned long uv;
-    int n = 1;
+    int c = *p++ & 0xff;
+    unsigned long uv = c;
+    long n;
 
-    if (c < 0xc0) n = 1;
-    else if (c < 0xe0) n = 2;
-    else if (c < 0xf0) n = 3;
-    else if (c < 0xf8) n = 4;
-    else if (c < 0xfc) n = 5;
-    else if (c < 0xfe) n = 6;
-    else if (c == 0xfe) n = 7;
-    if (n > *lenp) return 0;
+    if (!(uv & 0x80)) {
+	*lenp = 1;
+        return uv;
+    }
+    if (!(uv & 0x40)) {
+	*lenp = 1;
+	rb_raise(rb_eArgError, "malformed UTF-8 character");
+    }
+
+    if      (!(uv & 0x20)) { n = 2; uv &= 0x1f; }
+    else if (!(uv & 0x10)) { n = 3; uv &= 0x0f; }
+    else if (!(uv & 0x08)) { n = 4; uv &= 0x07; }
+    else if (!(uv & 0x04)) { n = 5; uv &= 0x03; }
+    else if (!(uv & 0x02)) { n = 6; uv &= 0x01; }
+    else {
+	*lenp = 1;
+	rb_raise(rb_eArgError, "malformed UTF-8 character");
+    }
+    if (n > *lenp) {
+	rb_raise(rb_eArgError, "malformed UTF-8 character (expected %d bytes, given %d bytes)",
+		 n, *lenp);
+    }
     *lenp = n--;
-
-    uv = c;
     if (n != 0) {
-	uv &= (1<<(BYTEWIDTH-2-n)) - 1;
 	while (n--) {
-	    uv = uv << 6 | *p++ & ((1<<6)-1);
+	    c = *p++ & 0xff;
+	    if ((c & 0xc0) != 0x80) {
+		*lenp -= n + 1;
+		rb_raise(rb_eArgError, "malformed UTF-8 character");
+	    }
+	    else {
+		c &= 0x3f;
+		uv = uv << 6 | c;
+	    }
 	}
     }
+    n = *lenp - 1;
+    if (uv < utf8_limits[n]) {
+	rb_raise(rb_eArgError, "redundant UTF-8 sequence");
+    }
     return uv;
 }
 
