@@ -3,41 +3,24 @@
   time.c -
 
   $Author: matz $
-  $Date: 2002/07/31 05:17:28 $
+  $Date: 2003/07/24 18:33:40 $
   created at: Tue Dec 28 14:31:59 JST 1993
 
-  Copyright (C) 1993-2000 Yukihiro Matsumoto
+  Copyright (C) 1993-2003 Yukihiro Matsumoto
 
 **********************************************************************/
 
 #include "ruby.h"
 #include <sys/types.h>
-
 #include <time.h>
-#ifndef NT
-#ifdef HAVE_SYS_TIME_H
-# include <sys/time.h>
-#else
-#define time_t long
-struct timeval {
-        time_t tv_sec;		/* seconds */
-        time_t tv_usec;		/* and microseconds */
-};
-#endif
-#endif /* NT */
-
-#ifdef HAVE_SYS_TIMES_H
-#include <sys/times.h>
-#endif
 
 #ifdef HAVE_UNISTD_H
 #include <unistd.h>
 #endif
 
+#include <math.h>
+
 VALUE rb_cTime;
-#if defined(HAVE_TIMES) || defined(NT)
-static VALUE S_Tms;
-#endif
 
 struct time_object {
     struct timeval tv;
@@ -46,20 +29,28 @@ struct time_object {
     int tm_got;
 };
 
-#define GetTimeval(obj, tobj) {\
-    Data_Get_Struct(obj, struct time_object, tobj);\
+#define GetTimeval(obj, tobj) \
+    Data_Get_Struct(obj, struct time_object, tobj)
+
+static void time_free _((void *));
+
+static void
+time_free(tobj)
+    void *tobj;
+{
+    if (tobj) free(tobj);
 }
 
+static VALUE time_s_alloc _((VALUE));
 static VALUE
-time_s_now(klass)
+time_s_alloc(klass)
     VALUE klass;
 {
     VALUE obj;
     struct time_object *tobj;
 
-    obj = Data_Make_Struct(klass, struct time_object, 0, free, tobj);
+    obj = Data_Make_Struct(klass, struct time_object, 0, time_free, tobj);
     tobj->tm_got=0;
-
     if (gettimeofday(&tobj->tv, 0) < 0) {
 	rb_sys_fail("gettimeofday");
     }
@@ -68,37 +59,47 @@ time_s_now(klass)
 }
 
 static VALUE
-time_s_new(argc, argv, klass)
-    int argc;
-    VALUE *argv;
+time_s_now(klass)
     VALUE klass;
 {
-    VALUE obj = time_s_now(klass);
-
-    rb_obj_call_init(obj, argc, argv);
-    return obj;
+    return rb_obj_alloc(klass);
 }
 
+
+#define NDIV(x,y) (-(-((x)+1)/(y))-1)
+#define NMOD(x,y) ((y)-(-((x)+1)%(y))-1)
+
 static VALUE
 time_new_internal(klass, sec, usec)
     VALUE klass;
     time_t sec, usec;
 {
     VALUE obj;
+    time_t tmp;
     struct time_object *tobj;
 
-    if (usec >= 1000000) {	/* usec overflow */
-	sec += usec / 1000000;
+    if (usec >= 1000000) {	/* usec positive overflow */
+	tmp = sec + usec / 1000000;
 	usec %= 1000000;
+	if (sec > 0 && tmp < 0) {
+	    rb_raise(rb_eRangeError, "out of Time range");
+	}
+	sec = tmp;
     }
-    while (usec < 0) {		/* usec underflow */
-	sec--;
-	usec += 1000000;
+    if (usec < 0) {		/* usec negative overflow */
+	tmp = sec + NDIV(usec,1000000); /* negative div */
+	usec = NMOD(usec,1000000);      /* negative mod */
+	if (sec < 0 && tmp > 0) {
+	    rb_raise(rb_eRangeError, "out of Time range");
+	}
+	sec = tmp;
     }
+#ifndef NEGATIVE_TIME_T
     if (sec < 0 || (sec == 0 && usec < 0))
 	rb_raise(rb_eArgError, "time must be positive");
+#endif
 
-    obj = Data_Make_Struct(klass, struct time_object, 0, free, tobj);
+    obj = Data_Make_Struct(klass, struct time_object, 0, time_free, tobj);
     tobj->tm_got = 0;
     tobj->tv.tv_sec = sec;
     tobj->tv.tv_usec = usec;
@@ -119,33 +120,45 @@ time_timeval(time, interval)
     int interval;
 {
     struct timeval t;
+    char *tstr = interval ? "time interval" : "time";
+
+#ifndef NEGATIVE_TIME_T
+    interval = 1;
+#endif
 
     switch (TYPE(time)) {
       case T_FIXNUM:
 	t.tv_sec = FIX2LONG(time);
-	if (t.tv_sec < 0)
-	    rb_raise(rb_eArgError, "time must be positive");
+	if (interval && t.tv_sec < 0)
+	    rb_raise(rb_eArgError, "%s must be positive", tstr);
 	t.tv_usec = 0;
 	break;
 
       case T_FLOAT:
-	if (RFLOAT(time)->value < 0.0)
-	    rb_raise(rb_eArgError, "time must be positive");
-	t.tv_sec = (time_t)RFLOAT(time)->value;
-	t.tv_usec = (time_t)((RFLOAT(time)->value - (double)t.tv_sec)*1e6);
+	if (interval && RFLOAT(time)->value < 0.0)
+	    rb_raise(rb_eArgError, "%s must be positive", tstr);
+	else {
+	    double f, d;
+
+	    d = modf(RFLOAT(time)->value, &f);
+	    t.tv_sec = (time_t)f;
+	    if (f != t.tv_sec) {
+		rb_raise(rb_eRangeError, "%f out of Time range", RFLOAT(time)->value);
+	    }
+	    t.tv_usec = (time_t)(d*1e6);
+	}
 	break;
 
       case T_BIGNUM:
 	t.tv_sec = NUM2LONG(time);
-	if (t.tv_sec < 0)
-	    rb_raise(rb_eArgError, "time must be positive");
+	if (interval && t.tv_sec < 0)
+	    rb_raise(rb_eArgError, "%s must be positive", tstr);
 	t.tv_usec = 0;
 	break;
 
       default:
-	rb_raise(rb_eTypeError, "can't convert %s into Time%s",
-		 rb_class2name(CLASS_OF(time)),
-		 interval ? " interval" : "");
+	rb_raise(rb_eTypeError, "can't convert %s into %s",
+		 rb_obj_classname(time), tstr);
 	break;
     }
     return t;
@@ -165,7 +178,7 @@ rb_time_timeval(time)
     struct time_object *tobj;
     struct timeval t;
 
-    if (rb_obj_is_kind_of(time, rb_cTime)) {
+    if (TYPE(time) == T_DATA && RDATA(time)->dfree == time_free) {
 	GetTimeval(time, tobj);
 	t = tobj->tv;
 	return t;
@@ -183,14 +196,14 @@ time_s_at(argc, argv, klass)
     VALUE time, t;
 
     if (rb_scan_args(argc, argv, "11", &time, &t) == 2) {
-	tv.tv_sec = NUM2INT(time);
-	tv.tv_usec = NUM2INT(t);
+	tv.tv_sec = NUM2LONG(time);
+	tv.tv_usec = NUM2LONG(t);
     }
     else {
 	tv = rb_time_timeval(time);
     }
     t = time_new_internal(klass, tv.tv_sec, tv.tv_usec);
-    if (TYPE(time) == T_DATA) {
+    if (TYPE(time) == T_DATA && RDATA(time)->dfree == time_free) {
 	struct time_object *tobj, *tobj2;
 
 	GetTimeval(time, tobj);
@@ -210,7 +223,7 @@ obj2long(obj)
     VALUE obj;
 {
     if (TYPE(obj) == T_STRING) {
-	obj = rb_str2inum(obj, 10);
+	obj = rb_str_to_inum(obj, 10, Qfalse);
     }
 
     return NUM2LONG(obj);
@@ -223,10 +236,12 @@ time_arg(argc, argv, tm, usec)
     struct tm *tm;
     time_t *usec;
 {
-    VALUE v[7];
+    VALUE v[8];
     int i;
+    long year;
 
     MEMZERO(tm, struct tm, 1);
+    *usec = 0;
     if (argc == 10) {
 	v[0] = argv[5];
 	v[1] = argv[4];
@@ -234,40 +249,57 @@ time_arg(argc, argv, tm, usec)
 	v[3] = argv[2];
 	v[4] = argv[1];
 	v[5] = argv[0];
-	*usec = 0;
+	v[6] = Qnil;
 	tm->tm_isdst = RTEST(argv[8]) ? 1 : 0;
     }
     else {
-	rb_scan_args(argc, argv, "16", &v[0],&v[1],&v[2],&v[3],&v[4],&v[5],&v[6]);
-	*usec = NIL_P(v[6]) ? 0 : obj2long(v[6]);
+	rb_scan_args(argc, argv, "17", &v[0],&v[1],&v[2],&v[3],&v[4],&v[5],&v[6],&v[7]);
+	/* v[6] may be usec or zone (parsedate) */
+	/* v[7] is wday (parsedate; ignored) */
+	tm->tm_wday = -1;
 	tm->tm_isdst = -1;
     }
 
-    tm->tm_year = obj2long(v[0]);
-    if (0 <= tm->tm_year && tm->tm_year < 69) tm->tm_year += 100;
-    if (tm->tm_year >= 1900) tm->tm_year -= 1900;
+    year = obj2long(v[0]);
+
+    if (0 <= year && year < 39) {
+	year += 100;
+	rb_warning("2 digits year is used");
+    }
+    else if (69 <= year && year < 139) {
+	rb_warning("2 or 3 digits year is used");
+    }
+    else {
+	year -= 1900;
+    }
+
+    tm->tm_year = year;
+
     if (NIL_P(v[1])) {
 	tm->tm_mon = 0;
     }
-    else if (TYPE(v[1]) == T_STRING) {
-	tm->tm_mon = -1;
-	for (i=0; i<12; i++) {
-	    if (RSTRING(v[1])->len == 3 &&
-		strcasecmp(months[i], RSTRING(v[1])->ptr) == 0) {
-		tm->tm_mon = i;
-		break;
-	    }
-	}
-	if (tm->tm_mon == -1) {
-	    char c = RSTRING(v[1])->ptr[0];
-
-	    if ('0' <= c && c <= '9') {
-		tm->tm_mon = obj2long(v[1])-1;
-	    }
-	}
-    }
     else {
-	tm->tm_mon = obj2long(v[1]) - 1;
+	VALUE s = rb_check_string_type(v[1]);
+	if (!NIL_P(s)) {
+	    tm->tm_mon = -1;
+	    for (i=0; i<12; i++) {
+		if (RSTRING(s)->len == 3 &&
+		    strcasecmp(months[i], RSTRING(v[1])->ptr) == 0) {
+		    tm->tm_mon = i;
+		    break;
+		}
+	    }
+	    if (tm->tm_mon == -1) {
+		char c = RSTRING(s)->ptr[0];
+
+		if ('0' <= c && c <= '9') {
+		    tm->tm_mon = obj2long(s)-1;
+		}
+	    }
+	}
+	else {
+	    tm->tm_mon = obj2long(v[1])-1;
+	}
     }
     if (NIL_P(v[2])) {
 	tm->tm_mday = 1;
@@ -278,10 +310,22 @@ time_arg(argc, argv, tm, usec)
     tm->tm_hour = NIL_P(v[3])?0:obj2long(v[3]);
     tm->tm_min  = NIL_P(v[4])?0:obj2long(v[4]);
     tm->tm_sec  = NIL_P(v[5])?0:obj2long(v[5]);
+    if (!NIL_P(v[6])) {
+	if (argc == 8) {
+	    /* v[6] is timezone, but ignored */
+	}
+	else if (argc == 7) {
+	    *usec = obj2long(v[6]);
+	}
+    }
 
     /* value validation */
-    if (   tm->tm_year < 69
-	|| tm->tm_mon  < 0 || tm->tm_mon  > 11
+    if (
+           tm->tm_year != year ||
+#ifndef NEGATIVE_TIME_T
+           tm->tm_year < 69 ||
+#endif
+	   tm->tm_mon  < 0 || tm->tm_mon  > 11
 	|| tm->tm_mday < 1 || tm->tm_mday > 31
 	|| tm->tm_hour < 0 || tm->tm_hour > 23
 	|| tm->tm_min  < 0 || tm->tm_min  > 59
@@ -293,96 +337,248 @@ static VALUE time_gmtime _((VALUE));
 static VALUE time_localtime _((VALUE));
 static VALUE time_get_tm _((VALUE, int));
 
-static time_t
-make_time_t(tptr, utc_or_local)
-    struct tm *tptr;
-    int utc_or_local;
+static int
+tmcmp(a, b)
+    struct tm *a;
+    struct tm *b;
 {
-    struct timeval tv;
-    time_t oguess, guess;
-    struct tm *tm;
-    long t, diff;
+    if (a->tm_year != b->tm_year)
+        return a->tm_year < b->tm_year ? -1 : 1;
+    else if (a->tm_mon != b->tm_mon)
+        return a->tm_mon < b->tm_mon ? -1 : 1;
+    else if (a->tm_mday != b->tm_mday)
+        return a->tm_mday < b->tm_mday ? -1 : 1;
+    else if (a->tm_hour != b->tm_hour)
+        return a->tm_hour < b->tm_hour ? -1 : 1;
+    else if (a->tm_min != b->tm_min)
+        return a->tm_min < b->tm_min ? -1 : 1;
+    else if (a->tm_sec != b->tm_sec)
+        return a->tm_sec < b->tm_sec ? -1 : 1;
+    else
+        return 0;
+}
 
-    if (gettimeofday(&tv, 0) < 0) {
-	rb_sys_fail("gettimeofday");
-    }
-    guess = tv.tv_sec;
+static time_t
+search_time_t(tptr, utc_p)
+    struct tm *tptr;
+    int utc_p;
+{
+    time_t guess, guess_lo, guess_hi;
+    struct tm *tm, tm_lo, tm_hi;
+    int d, have_guess;
+    int find_dst;
 
-    tm = gmtime(&guess);
+    find_dst = 0 < tptr->tm_isdst;
+
+#ifdef NEGATIVE_TIME_T
+    guess_lo = 1L << (8 * sizeof(time_t) - 1);
+#else
+    guess_lo = 0;
+#endif
+    guess_hi = ((time_t)-1) < ((time_t)0) ?
+               (1UL << (8 * sizeof(time_t) - 1)) - 1 :
+	       ~(time_t)0;
+
+    tm = (utc_p ? gmtime : localtime)(&guess_lo);
     if (!tm) goto error;
-    t = tptr->tm_year;
-    if (t < 69) goto out_of_range;
-    while (diff = t - tm->tm_year) {
-	oguess = guess;
-	guess += diff * 363 * 24 * 3600;
-	if (diff > 0 && guess <= oguess) goto out_of_range;
-	tm = gmtime(&guess);
-	if (!tm) goto error;
-    }
-    t = tptr->tm_mon;
-    while (diff = t - tm->tm_mon) {
-	guess += diff * 27 * 24 * 3600;
-	tm = gmtime(&guess);
-	if (!tm) goto error;
-	if (tptr->tm_year != tm->tm_year) goto out_of_range;
-    }
-    guess += (tptr->tm_mday - tm->tm_mday) * 24 * 3600;
-    guess += (tptr->tm_hour - tm->tm_hour) * 3600;
-    guess += (tptr->tm_min - tm->tm_min) * 60;
-    guess += (tptr->tm_sec - tm->tm_sec);
-    if (guess < 0) goto out_of_range;
+    d = tmcmp(tptr, tm);
+    if (d < 0) goto out_of_range;
+    if (d == 0) return guess_lo;
+    tm_lo = *tm;
 
-    if (!utc_or_local) {	/* localtime zone adjust */
-	struct tm gt, lt;
-	long tzsec;
+    tm = (utc_p ? gmtime : localtime)(&guess_hi);
+    if (!tm) goto error;
+    d = tmcmp(tptr, tm);
+    if (d > 0) goto out_of_range;
+    if (d == 0) return guess_hi;
+    tm_hi = *tm;
 
-	t = 0;
-	tm = gmtime(&guess);
-	if (!tm) goto error;
-	gt = *tm;
-	tm = localtime(&guess);
-	if (!tm) goto error;
-	lt = *tm;
-	tzsec = (gt.tm_min-lt.tm_min)*60 + (gt.tm_hour-lt.tm_hour)*3600;
+    have_guess = 0;
 
-	if (lt.tm_year > gt.tm_year) {
-	    tzsec -= 24*3600;
-	}
-	else if(gt.tm_year > lt.tm_year) {
-	    tzsec += 24*3600;
-	}
-	else {
-	    tzsec += (gt.tm_yday - lt.tm_yday)*24*3600;
-	}
-	if (lt.tm_isdst) guess += 3600;
-	guess += tzsec;
-	if (guess < 0) {
-	    goto out_of_range;
-	}
-	tm = localtime(&guess);
-	if (!tm) goto error;
-	if (lt.tm_isdst != tm->tm_isdst || tptr->tm_hour != tm->tm_hour) {
-	    time_t tmp = guess - 3600;
-	    tm = localtime(&tmp);
-	    if (!tm) goto error;
-	    if (tptr->tm_hour == tm->tm_hour) {
-		guess = tmp;
-	    }
-	    else if (lt.tm_isdst == tm->tm_isdst) {
-		tmp = guess + 3600;
-		tm = localtime(&tmp);
-		if (!tm) goto error;
-		if (tptr->tm_hour == tm->tm_hour) {
-		    guess = tmp;
+    while (guess_lo + 1 < guess_hi) {
+      /* there is a gap between guess_lo and guess_hi. */
+      unsigned long range = 0;
+      if (!have_guess) {
+	int a, b;
+	/*
+	  Try precious guess by a linear interpolation at first.
+	  `a' and `b' is a coefficient of guess_lo and guess_hi as:
+
+	    guess = (guess_lo * a + guess_hi * b) / (a + b)
+
+	  However this causes overflow in most cases, following assignment
+	  is used instead:
+
+	    guess = guess_lo / d * a + (guess_lo % d) * a / d
+		  + guess_hi / d * b + (guess_hi % d) * b / d
+	      where d = a + b
+
+	  To avoid overflow in this assignment, `d' is restricted to less than
+	  sqrt(2**31).  By this restriction and other reasons, the guess is
+	  not accurate and some error is expected.  `range' approximates 
+	  the maximum error.
+
+	  When these parameters are not suitable, i.e. guess is not within
+	  guess_lo and guess_hi, simple guess by binary search is used.
+	*/
+	range = 366 * 24 * 60 * 60;
+	a = (tm_hi.tm_year - tptr->tm_year);
+	b = (tptr->tm_year - tm_lo.tm_year);
+	/* 46000 is selected as `some big number less than sqrt(2**31)'. */
+	if (a + b <= 46000 / 12) {
+	  range = 31 * 24 * 60 * 60;
+	  a *= 12;
+	  b *= 12;
+	  a += tm_hi.tm_mon - tptr->tm_mon;
+	  b += tptr->tm_mon - tm_lo.tm_mon;
+	  if (a + b <= 46000 / 31) {
+	    range = 24 * 60 * 60;
+	    a *= 31;
+	    b *= 31;
+	    a += tm_hi.tm_mday - tptr->tm_mday;
+	    b += tptr->tm_mday - tm_lo.tm_mday;
+	    if (a + b <= 46000 / 24) {
+	      range = 60 * 60;
+	      a *= 24;
+	      b *= 24;
+	      a += tm_hi.tm_hour - tptr->tm_hour;
+	      b += tptr->tm_hour - tm_lo.tm_hour;
+	      if (a + b <= 46000 / 60) {
+		range = 60;
+		a *= 60;
+		b *= 60;
+		a += tm_hi.tm_min - tptr->tm_min;
+		b += tptr->tm_min - tm_lo.tm_min;
+		if (a + b <= 46000 / 60) {
+		  range = 1;
+		  a *= 60;
+		  b *= 60;
+		  a += tm_hi.tm_sec - tptr->tm_sec;
+		  b += tptr->tm_sec - tm_lo.tm_sec;
 		}
+	      }
 	    }
+	  }
 	}
-	if (guess < 0) {
-	    goto out_of_range;
-	}
-    }
+	if (a <= 0) a = 1;
+	if (b <= 0) b = 1;
+	d = a + b;
+	/*
+	  Although `/' and `%' may produce unexpected result with negative
+	  argument, it doesn't cause serious problem because there is a
+	  fail safe.
+	*/
+	guess = guess_lo / d * a + (guess_lo % d) * a / d
+	      + guess_hi / d * b + (guess_hi % d) * b / d;
+	have_guess = 1;
+      }
 
-    return guess;
+      if (guess <= guess_lo || guess_hi <= guess) {
+	/* Precious guess is invalid. try binary search. */ 
+	guess = guess_lo / 2 + guess_hi / 2;
+	if (guess <= guess_lo)
+	  guess = guess_lo + 1;
+	else if (guess >= guess_hi)
+	  guess = guess_hi - 1;
+	range = 0;
+      }
+
+      tm = (utc_p ? gmtime : localtime)(&guess);
+      if (!tm) goto error;
+      have_guess = 0;
+
+      d = tmcmp(tptr, tm);
+      if (d < 0) {
+        guess_hi = guess;
+	tm_hi = *tm;
+	if (range) {
+	  guess = guess - range;
+	  range = 0;
+	  if (guess_lo < guess && guess < guess_hi)
+	    have_guess = 1;
+	}
+      }
+      else if (d > 0) {
+        guess_lo = guess;
+	tm_lo = *tm;
+	if (range) {
+	  guess = guess + range;
+	  range = 0;
+	  if (guess_lo < guess && guess < guess_hi)
+	    have_guess = 1;
+	}
+      }
+      else {
+	if (!utc_p) {
+	  /* If localtime is nonmonotonic, another result may exist. */
+	  time_t guess2;
+	  if (find_dst) {
+	    guess2 = guess - 2 * 60 * 60;
+	    tm = localtime(&guess2);
+	    if (tm) {
+	      if (tptr->tm_hour != (tm->tm_hour + 2) % 24 ||
+		  tptr->tm_min != tm->tm_min ||
+		  tptr->tm_sec != tm->tm_sec) {
+		guess2 -= (tm->tm_hour - tptr->tm_hour) * 60 * 60 +
+			  (tm->tm_min - tptr->tm_min) * 60 +
+			  (tm->tm_sec - tptr->tm_sec);
+		if (tptr->tm_mday != tm->tm_mday)
+		  guess2 += 24 * 60 * 60;
+		if (guess != guess2) {
+		  tm = localtime(&guess2);
+		  if (tmcmp(tptr, tm) == 0) {
+		    if (guess < guess2)
+		      return guess;
+		    else
+		      return guess2;
+		  }
+		}
+	      }
+	    }
+	  }
+	  else {
+	    guess2 = guess + 2 * 60 * 60;
+	    tm = localtime(&guess2);
+	    if (tm) {
+	      if ((tptr->tm_hour + 2) % 24 != tm->tm_hour ||
+		  tptr->tm_min != tm->tm_min ||
+		  tptr->tm_sec != tm->tm_sec) {
+		guess2 -= (tm->tm_hour - tptr->tm_hour) * 60 * 60 +
+			  (tm->tm_min - tptr->tm_min) * 60 +
+			  (tm->tm_sec - tptr->tm_sec);
+		if (tptr->tm_mday != tm->tm_mday)
+		  guess2 -= 24 * 60 * 60;
+		if (guess != guess2) {
+		  tm = localtime(&guess2);
+		  if (tmcmp(tptr, tm) == 0) {
+		    if (guess < guess2)
+		      return guess2;
+		    else
+		      return guess;
+		  }
+		}
+	      }
+	    }
+	  }
+	}
+	return guess;
+      }
+    }
+    /* Given argument has no corresponding time_t. Let's outerpolation. */
+    if (tm_lo.tm_year == tptr->tm_year && tm_lo.tm_mon == tptr->tm_mon) {
+      return guess_lo +
+        (tptr->tm_mday - tm_lo.tm_mday) * 24 * 60 * 60 +
+        (tptr->tm_hour - tm_lo.tm_hour) * 60 * 60 +
+        (tptr->tm_min - tm_lo.tm_min) * 60 +
+        (tptr->tm_sec - tm_lo.tm_sec);
+    }
+    else if (tm_hi.tm_year == tptr->tm_year && tm_hi.tm_mon == tptr->tm_mon) {
+      return guess_hi +
+        (tptr->tm_mday - tm_hi.tm_mday) * 24 * 60 * 60 +
+        (tptr->tm_hour - tm_hi.tm_hour) * 60 * 60 +
+        (tptr->tm_min - tm_hi.tm_min) * 60 +
+        (tptr->tm_sec - tm_hi.tm_sec);
+    }
 
   out_of_range:
     rb_raise(rb_eArgError, "time out of range");
@@ -392,6 +588,50 @@ make_time_t(tptr, utc_or_local)
     return 0;			/* not reached */
 }
 
+static time_t
+make_time_t(tptr, utc_p)
+    struct tm *tptr;
+    int utc_p;
+{
+    time_t t;
+    struct tm *tmp, buf;
+    buf = *tptr;
+    if (utc_p) {
+#if defined(HAVE_TIMEGM)
+	if ((t = timegm(&buf)) != -1)
+            return t;
+#ifdef NEGATIVE_TIME_T
+        if ((tmp = gmtime(&t)) &&
+            tptr->tm_year == tmp->tm_year &&
+            tptr->tm_mon == tmp->tm_mon &&
+            tptr->tm_mday == tmp->tm_mday &&
+            tptr->tm_hour == tmp->tm_hour &&
+            tptr->tm_min == tmp->tm_min &&
+            tptr->tm_sec == tmp->tm_sec)
+            return t;
+#endif
+#endif
+	return search_time_t(&buf, utc_p);
+    }
+    else {
+#if defined(HAVE_MKTIME)
+	if ((t = mktime(&buf)) != -1)
+            return t;
+#ifdef NEGATIVE_TIME_T
+        if ((tmp = localtime(&t)) &&
+            tptr->tm_year == tmp->tm_year &&
+            tptr->tm_mon == tmp->tm_mon &&
+            tptr->tm_mday == tmp->tm_mday &&
+            tptr->tm_hour == tmp->tm_hour &&
+            tptr->tm_min == tmp->tm_min &&
+            tptr->tm_sec == tmp->tm_sec)
+            return t;
+#endif
+#endif
+	return search_time_t(&buf, utc_p);
+    }
+}
+
 static VALUE
 time_utc_or_local(argc, argv, utc_p, klass)
     int argc;
@@ -434,7 +674,7 @@ time_to_i(time)
     struct time_object *tobj;
 
     GetTimeval(time, tobj);
-    return INT2NUM(tobj->tv.tv_sec);
+    return LONG2NUM(tobj->tv.tv_sec);
 }
 
 static VALUE
@@ -454,7 +694,7 @@ time_usec(time)
     struct time_object *tobj;
 
     GetTimeval(time, tobj);
-    return INT2NUM(tobj->tv.tv_usec);
+    return LONG2NUM(tobj->tv.tv_usec);
 }
 
 static VALUE
@@ -462,36 +702,9 @@ time_cmp(time1, time2)
     VALUE time1, time2;
 {
     struct time_object *tobj1, *tobj2;
-    long i;
 
     GetTimeval(time1, tobj1);
-    switch (TYPE(time2)) {
-      case T_FIXNUM:
-	i = FIX2LONG(time2);
-	if (tobj1->tv.tv_sec == i) {
-	    if (tobj1->tv.tv_usec == 0)
-		return INT2FIX(0);
-	    if (tobj1->tv.tv_usec > 0)
-		return INT2FIX(1);
-	    return INT2FIX(-1);
-	}
-	if (tobj1->tv.tv_sec > i) return INT2FIX(1);
-	return INT2FIX(-1);
-	
-      case T_FLOAT:
-	{
-	    double t;
-
-	    t = (double)tobj1->tv.tv_sec + (double)tobj1->tv.tv_usec*1e-6;
-	    if (t > RFLOAT(time2)->value)
-		return INT2FIX(1);
-	    if (t < RFLOAT(time2)->value)
-		return INT2FIX(-1);
-	    return INT2FIX(0);
-	}
-    }
-
-    if (rb_obj_is_kind_of(time2, rb_cTime)) {
+    if (TYPE(time2) == T_DATA && RDATA(time2)->dfree == time_free) {
 	GetTimeval(time2, tobj2);
 	if (tobj1->tv.tv_sec == tobj2->tv.tv_sec) {
 	    if (tobj1->tv.tv_usec == tobj2->tv.tv_usec) return INT2FIX(0);
@@ -501,24 +714,8 @@ time_cmp(time1, time2)
 	if (tobj1->tv.tv_sec > tobj2->tv.tv_sec) return INT2FIX(1);
 	return INT2FIX(-1);
     }
-    if (TYPE(time2) == T_BIGNUM) {
-	double a = (double)tobj1->tv.tv_sec+(double)tobj1->tv.tv_usec/1e6;
-	double b = rb_big2dbl(time2);
 
-	if (a == b) return INT2FIX(0);
-	if (a > b) return INT2FIX(1);
-	if (a < b) return INT2FIX(-1);
-    }
-    i = NUM2LONG(time2);
-    if (tobj1->tv.tv_sec == i) {
-	if (tobj1->tv.tv_usec == 0)
-	    return INT2FIX(0);
-	if (tobj1->tv.tv_usec > 0)
-	    return INT2FIX(1);
-	return INT2FIX(-1);
-    }
-    if (tobj1->tv.tv_sec > i) return INT2FIX(1);
-    return INT2FIX(-1);
+    return Qnil;
 }
 
 static VALUE
@@ -528,7 +725,7 @@ time_eql(time1, time2)
     struct time_object *tobj1, *tobj2;
 
     GetTimeval(time1, tobj1);
-    if (rb_obj_is_kind_of(time2, rb_cTime)) {
+    if (TYPE(time2) == T_DATA && RDATA(time2)->dfree == time_free) {
 	GetTimeval(time2, tobj2);
 	if (tobj1->tv.tv_sec == tobj2->tv.tv_sec) {
 	    if (tobj1->tv.tv_usec == tobj2->tv.tv_usec) return Qtrue;
@@ -557,33 +754,45 @@ time_hash(time)
 
     GetTimeval(time, tobj);
     hash = tobj->tv.tv_sec ^ tobj->tv.tv_usec;
-    return INT2FIX(hash);
-}
-
-static VALUE
-time_clone(time)
-    VALUE time;
-{
-    VALUE clone;
-    struct time_object *tobj, *tclone;
-
-    GetTimeval(time, tobj);
-    clone = Data_Make_Struct(0, struct time_object, 0, free, tclone);
-    CLONESETUP(clone, time);
-    MEMCPY(tclone, tobj, struct time_object, 1);
-
-    return clone;
+    return LONG2FIX(hash);
 }
 
 static void
 time_modify(time)
     VALUE time;
 {
-    if (OBJ_FROZEN(time)) rb_error_frozen("Time");
+    rb_check_frozen(time);
     if (!OBJ_TAINTED(time) && rb_safe_level() >= 4)
 	rb_raise(rb_eSecurityError, "Insecure: can't modify Time");
 }
 
+static VALUE
+time_init_copy(copy, time)
+    VALUE copy, time;
+{
+    struct time_object *tobj, *tcopy;
+
+    if (copy == time) return copy;
+    time_modify(copy);
+    if (TYPE(time) != T_DATA || RDATA(time)->dfree != time_free) {
+	rb_raise(rb_eTypeError, "wrong argument type");
+    }
+    GetTimeval(time, tobj);
+    GetTimeval(copy, tcopy);
+    MEMCPY(tcopy, tobj, struct time_object, 1);
+
+    return copy;
+}
+
+static VALUE
+time_dup(time)
+    VALUE time;
+{
+    VALUE dup = time_s_alloc(rb_cTime);
+    time_init_copy(dup, time);
+    return dup;
+}
+
 static VALUE
 time_localtime(time)
     VALUE time;
@@ -602,6 +811,8 @@ time_localtime(time)
     }
     t = tobj->tv.tv_sec;
     tm_tmp = localtime(&t);
+    if (!tm_tmp)
+	rb_raise(rb_eArgError, "localtime error");
     tobj->tm = *tm_tmp;
     tobj->tm_got = 1;
     tobj->gmt = 0;
@@ -626,12 +837,28 @@ time_gmtime(time)
     }
     t = tobj->tv.tv_sec;
     tm_tmp = gmtime(&t);
+    if (!tm_tmp)
+	rb_raise(rb_eArgError, "gmtime error");
     tobj->tm = *tm_tmp;
     tobj->tm_got = 1;
     tobj->gmt = 1;
     return time;
 }
 
+static VALUE
+time_getlocaltime(time)
+    VALUE time;
+{
+    return time_localtime(time_dup(time));
+}
+
+static VALUE
+time_getgmtime(time)
+    VALUE time;
+{
+    return time_gmtime(time_dup(time));
+}
+
 static VALUE
 time_get_tm(time, gmt)
     VALUE time;
@@ -685,18 +912,33 @@ time_plus(time1, time2)
 {
     struct time_object *tobj;
     time_t sec, usec;
-    double f;
+    double f, d, v;
 
     GetTimeval(time1, tobj);
 
-    if (rb_obj_is_kind_of(time2, rb_cTime)) {
+    if (TYPE(time2) == T_DATA && RDATA(time2)->dfree == time_free) {
 	rb_raise(rb_eTypeError, "time + time?");
     }
-    f = NUM2DBL(time2);
+    v = NUM2DBL(time2);
+    d = modf(v, &f);
     sec = (time_t)f;
-    usec = tobj->tv.tv_usec + (time_t)((f - (double)sec)*1e6);
-    sec = tobj->tv.tv_sec + sec;
+    if (f != (double)sec) {
+	rb_raise(rb_eRangeError, "time + %f out of Time range", v);
+    }
+#ifndef NEGATIVE_TIME_T
+    if (f < 0 && -f >= tobj->tv.tv_sec) {
+	rb_raise(rb_eArgError, "time must be positive");
+    }
+#endif
+    usec = tobj->tv.tv_usec + (time_t)(d*1e6);
+    sec = tobj->tv.tv_sec + (time_t)f;
 
+#ifdef NEGATIVE_TIME_T
+    if ((tobj->tv.tv_sec >= 0 && f >= 0 && sec < 0) ||
+	(tobj->tv.tv_sec <= 0 && f <= 0 && sec > 0)) {
+	rb_raise(rb_eRangeError, "time + %f out of Time range", v);
+    }
+#endif
     time2 = rb_time_new(sec, usec);
     if (tobj->gmt) {
 	GetTimeval(time2, tobj);
@@ -711,24 +953,38 @@ time_minus(time1, time2)
 {
     struct time_object *tobj;
     time_t sec, usec;
-    double f;
+    double f, d, v;
 
     GetTimeval(time1, tobj);
-    if (rb_obj_is_kind_of(time2, rb_cTime)) {
+    if (TYPE(time2) == T_DATA && RDATA(time2)->dfree == time_free) {
 	struct time_object *tobj2;
 
 	GetTimeval(time2, tobj2);
-	f = tobj->tv.tv_sec - tobj2->tv.tv_sec;
-	f += (tobj->tv.tv_usec - tobj2->tv.tv_usec)*1e-6;
+	f = (double)tobj->tv.tv_sec - (double)tobj2->tv.tv_sec;
+	f += ((double)tobj->tv.tv_usec - (double)tobj2->tv.tv_usec)*1e-6;
+	/* XXX: should check float overflow on 64bit time_t platforms */
 
 	return rb_float_new(f);
     }
-    else {
-	f = NUM2DBL(time2);
-	sec = (time_t)f;
-	usec = tobj->tv.tv_usec - (time_t)((f - (double)sec)*1e6);
-	sec = tobj->tv.tv_sec - sec;
+    v = NUM2DBL(time2);
+    d = modf(v, &f);
+    sec = (time_t)f;
+    if (f != (double)sec) {
+	rb_raise(rb_eRangeError, "time - %f out of Time range", v);
     }
+#ifndef NEGATIVE_TIME_T
+    if (f > 0 && f >= tobj->tv.tv_sec) {
+	rb_raise(rb_eArgError, "time must be positive");
+    }
+#endif
+    usec = tobj->tv.tv_usec - (time_t)(d*1e6);
+    sec = tobj->tv.tv_sec - (time_t)f;
+#ifdef NEGATIVE_TIME_T
+    if ((tobj->tv.tv_sec <= 0 && f >= 0 && sec > 0) ||
+	(tobj->tv.tv_sec >= 0 && f <= 0 && sec < 0)) {
+	rb_raise(rb_eRangeError, "time - %f out of Time range", v);
+    }
+#endif
 
     time2 = rb_time_new(sec, usec);
     if (tobj->gmt) {
@@ -813,7 +1069,7 @@ time_year(time)
     if (tobj->tm_got == 0) {
 	time_get_tm(time, tobj->gmt);
     }
-    return INT2FIX(tobj->tm.tm_year+1900);
+    return LONG2NUM((long)tobj->tm.tm_year+1900);
 }
 
 static VALUE
@@ -860,14 +1116,19 @@ time_zone(time)
     VALUE time;
 {
     struct time_object *tobj;
+#if !defined(HAVE_TM_ZONE) && (!defined(HAVE_TZNAME) || !defined(HAVE_DAYLIGHT))
     char buf[64];
     int len;
-
+#endif
+    
     GetTimeval(time, tobj);
     if (tobj->tm_got == 0) {
 	time_get_tm(time, tobj->gmt);
     }
 
+    if (tobj->gmt == 1) {
+	return rb_str_new2("UTC");
+    }
 #if defined(HAVE_TM_ZONE)
     return rb_str_new2(tobj->tm.tm_zone);
 #elif defined(HAVE_TZNAME) && defined(HAVE_DAYLIGHT)
@@ -886,36 +1147,36 @@ time_utc_offset(time)
 
     GetTimeval(time, tobj);
     if (tobj->tm_got == 0) {
-        time_get_tm(time, tobj->gmt);
+	time_get_tm(time, tobj->gmt);
     }
 
     if (tobj->gmt == 1) {
-        return INT2FIX(0);
+	return INT2FIX(0);
     }
     else {
 #if defined(HAVE_STRUCT_TM_TM_GMTOFF)
-        return INT2NUM(tobj->tm.tm_gmtoff);
+	return INT2NUM(tobj->tm.tm_gmtoff);
 #else
-        struct tm *u, *l;
-        time_t t;
-        int off;
-        l = &tobj->tm;
-        t = tobj->tv.tv_sec;
-        u = gmtime(&t);
-        if (!u)
-            rb_raise(rb_eArgError, "gmtime error");
-        if (l->tm_year != u->tm_year)
-            off = l->tm_year < u->tm_year ? -1 : 1;
-        else if (l->tm_mon != u->tm_mon)
-            off = l->tm_mon < u->tm_mon ? -1 : 1;
-        else if (l->tm_mday != u->tm_mday)
-            off = l->tm_mday < u->tm_mday ? -1 : 1;
-        else
-            off = 0;
-        off = off * 24 + l->tm_hour - u->tm_hour;
-        off = off * 60 + l->tm_min - u->tm_min;
-        off = off * 60 + l->tm_sec - u->tm_sec;
-        return INT2FIX(off);
+	struct tm *u, *l;
+	time_t t;
+	long off;
+	l = &tobj->tm;
+	t = tobj->tv.tv_sec;
+	u = gmtime(&t);
+	if (!u)
+	    rb_raise(rb_eArgError, "gmtime error");
+	if (l->tm_year != u->tm_year)
+	    off = l->tm_year < u->tm_year ? -1 : 1;
+	else if (l->tm_mon != u->tm_mon)
+	    off = l->tm_mon < u->tm_mon ? -1 : 1;
+	else if (l->tm_mday != u->tm_mday)
+	    off = l->tm_mday < u->tm_mday ? -1 : 1;
+	else
+	    off = 0;
+	off = off * 24 + l->tm_hour - u->tm_hour;
+	off = off * 60 + l->tm_min - u->tm_min;
+	off = off * 60 + l->tm_sec - u->tm_sec;
+	return LONG2FIX(off);
 #endif
     }
 }
@@ -936,7 +1197,7 @@ time_to_a(time)
 		    INT2FIX(tobj->tm.tm_hour),
 		    INT2FIX(tobj->tm.tm_mday),
 		    INT2FIX(tobj->tm.tm_mon+1),
-		    INT2FIX(tobj->tm.tm_year+1900),
+		    LONG2NUM((long)tobj->tm.tm_year+1900),
 		    INT2FIX(tobj->tm.tm_wday),
 		    INT2FIX(tobj->tm.tm_yday+1),
 		    tobj->tm.tm_isdst?Qtrue:Qfalse,
@@ -983,20 +1244,21 @@ time_strftime(time, format)
 {
     struct time_object *tobj;
     char buffer[SMALLBUF];
-    char *fmt;
-    char *buf = buffer;
-    int len;
+    char *fmt, *buf = buffer;
+    long len;
     VALUE str;
 
     GetTimeval(time, tobj);
     if (tobj->tm_got == 0) {
 	time_get_tm(time, tobj->gmt);
     }
-    fmt = rb_str2cstr(format, &len);
+    StringValue(format);
+    fmt = RSTRING(format)->ptr;
+    len = RSTRING(format)->len;
     if (len == 0) {
 	rb_warning("strftime called with empty format string");
     }
-    if (strlen(fmt) < len) {
+    else if (strlen(fmt) < len) {
 	/* Ruby string may contain \0's. */
 	char *p = fmt, *pe = fmt + len;
 
@@ -1014,7 +1276,9 @@ time_strftime(time, format)
 	}
 	return str;
     }
-    len = rb_strftime(&buf, RSTRING(format)->ptr, &tobj->tm);
+    else {
+	len = rb_strftime(&buf, RSTRING(format)->ptr, &tobj->tm);
+    }
     str = rb_str_new(buf, len);
     if (buf != buffer) free(buf);
     return str;
@@ -1024,25 +1288,8 @@ static VALUE
 time_s_times(obj)
     VALUE obj;
 {
-#if defined(HAVE_TIMES) && !defined(__CHECKER__)
-#ifndef HZ
-# ifdef CLK_TCK
-#   define HZ CLK_TCK
-# else
-#   define HZ 60
-# endif
-#endif /* HZ */
-    struct tms buf;
-
-    if (times(&buf) == -1) rb_sys_fail(0);
-    return rb_struct_new(S_Tms,
-			 rb_float_new((double)buf.tms_utime / HZ),
-			 rb_float_new((double)buf.tms_stime / HZ),
-			 rb_float_new((double)buf.tms_cutime / HZ),
-			 rb_float_new((double)buf.tms_cstime / HZ));
-#else
-    rb_notimplement();
-#endif
+    rb_warn("obsolete method Time::times; use Process::times");
+    return rb_proc_times(obj);
 }
 
 static VALUE
@@ -1065,6 +1312,9 @@ time_dump(argc, argv, time)
     t = tobj->tv.tv_sec;
     tm = gmtime(&t);
 
+    if ((tm->tm_year & 0x1ffff) != tm->tm_year)
+	rb_raise(rb_eArgError, "too big year to marshal");
+
     p = 0x1          << 31 | /*  1 */
 	tm->tm_year  << 14 | /* 17 */
 	tm->tm_mon   << 10 | /*  4 */
@@ -1096,8 +1346,9 @@ time_load(klass, str)
     struct tm tm;
     int i;
 
-    buf = rb_str2cstr(str, &i);
-    if (i != 8) {
+    StringValue(str);
+    buf = (unsigned char *)RSTRING(str)->ptr;
+    if (RSTRING(str)->len != 8) {
 	rb_raise(rb_eTypeError, "marshaled time format differ");
     }
 
@@ -1119,9 +1370,10 @@ time_load(klass, str)
     tm.tm_hour =  p        & 0x1f;
     tm.tm_min  = (s >> 26) & 0x3f;
     tm.tm_sec  = (s >> 20) & 0x3f;
+    tm.tm_isdst = 0;
 
-    sec = make_time_t(&tm, gmtime);
-    usec = (time_t) s & 0xfffff;
+    sec = make_time_t(&tm, Qtrue);
+    usec = (time_t)(s & 0xfffff);
 
     return time_new_internal(klass, sec, usec);
 }
@@ -1133,7 +1385,7 @@ Init_Time()
     rb_include_module(rb_cTime, rb_mComparable);
 
     rb_define_singleton_method(rb_cTime, "now", time_s_now, 0);
-    rb_define_singleton_method(rb_cTime, "new", time_s_new, -1);
+    rb_define_alloc_func(rb_cTime, time_s_alloc);
     rb_define_singleton_method(rb_cTime, "at", time_s_at, -1);
     rb_define_singleton_method(rb_cTime, "utc", time_s_mkutc, -1);
     rb_define_singleton_method(rb_cTime, "gm", time_s_mkutc, -1);
@@ -1147,11 +1399,15 @@ Init_Time()
     rb_define_method(rb_cTime, "<=>", time_cmp, 1);
     rb_define_method(rb_cTime, "eql?", time_eql, 1);
     rb_define_method(rb_cTime, "hash", time_hash, 0);
-    rb_define_method(rb_cTime, "clone", time_clone, 0);
+    rb_define_method(rb_cTime, "initialize_copy", time_init_copy, 1);
 
     rb_define_method(rb_cTime, "localtime", time_localtime, 0);
     rb_define_method(rb_cTime, "gmtime", time_gmtime, 0);
     rb_define_method(rb_cTime, "utc", time_gmtime, 0);
+    rb_define_method(rb_cTime, "getlocal", time_getlocaltime, 0);
+    rb_define_method(rb_cTime, "getgm", time_getgmtime, 0);
+    rb_define_method(rb_cTime, "getutc", time_getgmtime, 0);
+
     rb_define_method(rb_cTime, "ctime", time_asctime, 0);
     rb_define_method(rb_cTime, "asctime", time_asctime, 0);
     rb_define_method(rb_cTime, "to_s", time_to_s, 0);
@@ -1187,10 +1443,6 @@ Init_Time()
 
     rb_define_method(rb_cTime, "strftime", time_strftime, 1);
 
-#if defined(HAVE_TIMES) || defined(NT)
-    S_Tms = rb_struct_define("Tms", "utime", "stime", "cutime", "cstime", 0);
-#endif
-
     /* methods for marshaling */
     rb_define_method(rb_cTime, "_dump", time_dump, -1);
     rb_define_singleton_method(rb_cTime, "_load", time_load, 1);
